# -*- test-case-name: mv3d.test.test_playerclient -*-
# Copyright (C) 2006 Mortal Coil Games
# See LICENSE for details.
"""
Class for a Player client interface 
Handles controlling the remote PC, taking player input
and converting it to calls to the PC.
Handles connecting to a PC as well.

"""


import logging

from twisted.internet import defer
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.task import coiterate
from twisted.application.service import Service

from zope.interface import implements

from mv3d.util.conductor import ISimulationClient, IPlayerClient
from mv3d.net.client import ServiceLoc
from mv3d.client.ui.ogre3d import (
    OgreRenderer, RendererIterator
    )
from mv3d.client.ui.cegui import CEGUI
from mv3d.client.ui.cegui import MessageWindow
from mv3d.client.ui.connect import LoginWindow
from mv3d.client.ui.connect import SelectPCWindow



class PlayerError(Exception):
    """
    Raised when there's an issue with the player service
    """



class PlayerClient(Service):
    """
    You need to add a docstring here ! 
    """
    implements(IPlayerClient)

    statsready = False
    pc = None
    pcid = None
    pcname = None
    autoConnect = False

    def __init__(self):
        self.actions = {}
        self.bindings = {}
        self.loginservers = []
        self.renderer = OgreRenderer()
        self.ui = CEGUI(self)


    def configure(self, nm, cf):
        """
        Configure the interface with this config file
        """
        if cf.has_option(nm, "defaultLogin"):
            deflogin = cf.get(nm, "defaultLogin").split(":")
            self.login = unicode(deflogin[0])
            self.password = unicode(deflogin[1])
        if cf.has_option(nm, "defaultPC"):
            self.pcname = unicode(cf.get(nm, "defaultPC"))
        loginservers = [s.strip() for s in cf.get(nm, "servers").split(",")]
        for s in loginservers:
            self.loginservers.append(ServiceLoc(s))
        if cf.has_option(nm, "autoConnect"):
            self.autoConnect = cf.getboolean(nm, "autoConnect")
            self.acServer = self.loginservers[int(cf.get(nm, "ac-server"))]
        self.renderer.configure(cf)


    def startService(self):
        if not self.renderer.initialize():
            raise PlayerError("Can't initialize the renderer")
        self.ui.initialize(renderer=self.renderer)
        coiterate(RendererIterator(self))


    def stopService(self):
        self.ui.finalize()
        self.renderer.finalize()


    def getRenderer(self):
        return self.renderer


    def getUI(self):
        return self.ui


    def getPCs(self):
        """
        This function will return a deferred, that will
            eventually give you a list of oids that your account
            can use as Players.
        """
        d = self.parent.getOneService(self.loginservers)
        return d.addCallback(lambda i: i.getPCs())


    def connectPC(self, oid):
        """
        This function will return a deferred that will
        return 0 if the client doesn't have access to use
        oid as a PC. Otherwise, it'll return a PlayerManipulator
        object for that PC which we will store as self.pc
        """
        d = self.parent.getOneService(self.loginservers)
        self.pcid = oid
        d.addCallback(lambda i: i.connectPC(oid))
        def connected(m):
            pc, view = m
            self.pc = pc
            i = self.parent.getLocalService(ISimulationClient)
            i.views[view.from_connection] = view
            return m
        return d.addCallback(connected)


    def do(self, what, *a, **kw):
        """
        You need to add a docstring here ! 
        """
        if isinstance(self.pc, int):
            self.parent.log("%s failed to perform action %s -- you haven't"
                "logged in yet!" % (self.name, what), logging.ERROR)
            raise PlayerError("No PC yet!")
        if self.pc.broker.disconnected:
            self.parent.log("%s failed to perform action %s -- you are"
                "not connected." % (self.name, what), logging.ERROR)
            raise PlayerError("Disconnected")
        return self.pc.callRemote(what, *a, **kw)


    def createCharacter(self, realmId):
        """
        Tell the server that we would like to create a new player
        character
        """
        d = self.parent.getOneService(self.loginservers)
        d.addCallback(lambda i: i.createCharacter(realmId))
        def gotIt(c):
            d = c.callRemote("getUI")
            d.addCallback(lambda ui: defer.maybeDeferred(
                ui.construct, c))
            def done(ui):
                ui.initialize(self.ui)
                return ui.start()
            return d.addCallback(done)
        return d.addCallback(gotIt)


    def addAction(self, b):
        """
        You need to add a docstring here ! 
        """
        self.actions[b.getName()] = b
        if self.bindings.has_key(b.getName()):
            b.SetBindings(self.bindings[b.getName()])
        b.activate()


    def remAction(self, nm):
        """
        You need to add a docstring here ! 
        """
        self.actions[nm].deactivate()
        del self.actions[nm]


    def getActions(self):
        """
        You need to add a docstring here ! 
        """
        return self.actions


    def getAction(self, nm):
        """
        You need to add a docstring here ! 
        """
        return self.actions[nm]


    def addBinding(self, n, b):
        """
        You need to add a docstring here ! 
        """
        if self.bindings.has_key(n):
            self.bindings[n].append(b)
        else:
            self.bindings[n] = [b]


    def getBindingsFor(self, n):
        """
        You need to add a docstring here ! 
        """
        return self.bindings.get(n, [])


    def remBindingsFor(self, n):
        """
        You need to add a docstring here ! 
        """
        del self.bindings[n]


    def remBinding(self, n, b):
        """
        You need to add a docstring here ! 
        """
        self.bindings[n].remove(b)


    def updateLoginServers(self):
        """ 
        This will get a server connection
        and ask it for suggestions of other login
        servers that we can use.. Update the list
        kept by the client with this new info.
        """
        d = self.parent.getOneService(self.loginservers)
        d.addCallback(lambda i: i.getPeerServers())
        def gotServers(lst):
            for s in lst:
                if not s in self.loginservers:
                    self.loginservers.append(s)
        return d.addCallback(gotServers)


    def connectAndSelectPC(self):
        """
        You need to add a docstring here ! 
        """
        self.ui.createImageSet("Loading", "mv3dtitle.png")
        self.ui.setBackgroundImage("Loading")
        def tryIt():
            if self.autoConnect:
                rs = ServiceLoc(self.acServer)
                if not rs in self.loginservers:
                    self.loginservers.append(rs)
                d = self.parent.getService(rs)
                d.addCallback(self.connectedToService)
                d.addErrback(error)
                return d
            self.lw = LoginWindow()
            self.lw.initialize(self.ui)
            d = self.lw.start()
            d.addCallback(self.connectedToService)
            d.addErrback(error)
            return d
        def error(e):
            e.printBriefTraceback()
            self.autoConnect = False # no endless loop please
            self.parent.log("Error connecting: %s" % (e.getMessage()),
                logging.ERROR)
            return tryIt()
        return tryIt()


    def connectedToService(self, c):
        """
        You need to add a docstring here ! 
        """
        # c.broker.notifyOnDisconnect(self.disconnected)
        def error(e):
            e.printBriefTraceback()
            self.autoConnect = False # no endless loop please
            self.parent.log("Error selecting PC: %s" % (e.getMessage()),
                logging.ERROR)
            return self.connectAndSelectPC()
        if self.pcname is not None:
            # print "AUTOLUSER"
            d = self.getPCs()
            def gotPCs(pcs):
                d = self.parent.getOneService(self.loginservers)
                def gotPS(ps):
                    dd = []
                    pcifo = {}
                    for pc in pcs:
                        d = ps.getPCInfo(pc)
                        def gotIt(info, pc):
                            pcifo[pc] = info
                        d.addCallback(gotIt, pc)
                        dd.append(d)
                    d = defer.DeferredList(dd)
                    def gotInfo(ignored):
                        for pcid, info in pcifo.items():
                            if info[0].startswith(self.pcname):
                                d = self.connectPC(pcid)
                                d.addCallback(self.connectedPC)
                                # d.addErrback(error)
                                return d
                        raise PlayerError("Couldn't get default PC")
                    return d.addCallback(gotInfo)
                return d.addCallback(gotPS)
            return d.addCallback(gotPCs)
        self.pcw = SelectPCWindow()
        # print "starting pcw"
        self.pcw.initialize(self.ui)
        d = self.pcw.start()
        d.addCallback(self.connectedPC)
        # d.addErrback(error)
        return d


    def connectedPC(self, r):
        """
        You need to add a docstring here ! 
        """
        #now get the UI
        d = self.do("getUI")
        d.addCallback(lambda r: defer.maybeDeferred(r.construct))
        def gotUI(u):
            u.initialize(self.ui)
            u.start()
            self.ui.removeBackgroundImage()
        return d.addCallback(gotUI)


    def updateStat(self, nr, stat):
        """
        You need to add a docstring here ! 
        """
        if not self.statsready:
            return
        stat = str(stat)
        g = self.ui.getComponent("Status%d" % nr)
        g.text = stat


    def disconnected(self):
        """
        Called when our connection to the server has been cut off
        """
        self.parent.log("Connection to server lost.", logging.WARNING)
        # attempt a reconnect to one of our servers
        d = self.connectPC(self.pcid)
        def connected(r):
            """
            Yipee! We connected!
            """
            self.parent.log("Successfully reconnected to server.",
                logging.INFO)
        def notConnected(e):
            """
            Waah! We can't reconnect!
            """
            m = MessageWindow()
            m.initialize(self.ui, "You have been disconnected.")
            d = m.start()
            def ClickedOk(r):
                self.ConnectAndSelectPC()
            return d.addCallback(ClickedOk)
        d.addCallback(connected)
        return d


    def iterateRenderer(self, t):
        """
        You need to add a docstring here ! 
        """
        self.renderer.iterate(t)
        self.updateStat(3, "Triangles: %d" % self.renderer.renderWindow.triangleCount)
        self.updateStat(0, "IPS: %.2f FPS %.1f" % (self.parent.ips, self.renderer.getFPS()))
        # self.updateStat(1, "Updates: %d" % self.lastups)
        # if len(self.conman.connections):
            # self.updateStat(2, "Lag: %.1f ms" % (float(self.conman.connections[0].lastpingtime) * 1000))
        self.parent.getLocalService(ISimulationClient).iteratePC(t)


    @inlineCallbacks
    def getAssetServices(self):
        """
        Returns a list of asset services from a conneted server
        """
        s = yield self.parent.getOneService(self.loginservers)
        svcs = yield s.getAssetServices()
        returnValue(svcs)


    def isLocal(self):
        """
        We are local so return true
        """
        return True