# Copyright (C) 2006-2008 Mortal Coil Games
# See LICENSE for details.
# -*- test-case-name: mv3d.test.test_connectwin -*-
"""
This contains the class for the PC selection window
"""


from twisted.internet import defer, reactor
from twisted.python.failure import Failure, DefaultException

from ogre.gui import CEGUI as cegui

from mv3d.net.client import ServiceLoc
from mv3d.client.ui.cegui import UserInterface
from mv3d.client.ui.cegui import QuickMessage


class SelectPCWindow(UserInterface):
    """
    The PC selection window. This is the client side interface that
    allows the user to select the pc from their account that they
    want to log in as.
    """
    ui = None
    deferred = None
    displayed = False
    inerror = False
    created = False


    def initialize(self, ui):
        """
        Build the UI if it hasn't been done already. Otherwise
        get references to the components we need.
        """
        self.ui = ui
        if not self.created:
            self.pclist = ui.createComponent("WindowsLook/MultiColumnList", 
                "PCList", [0.45, 0.3], [0.5, 0.5], [0.01, 0.01], [1.0, 1.0], '')
            self.pclist.addColumn('ID', 0, cegui.UDim(0.2, 0))
            self.pclist.addColumn('Name', 1, cegui.UDim(0.8, 0))
            self.pclist.showVertScrollbar = True
            self.connect = ui.createComponent("WindowsLook/Button", 
                "Connect", [0.45, 0.85], [0.1, 0.05], [0.01, 0.01], [1.0, 1.0], 'Connect')
            self.create = ui.createComponent("WindowsLook/Button", 
                "NewPC", [0.85, 0.85], [0.1, 0.05], [0.01, 0.01], [1.0, 1.0], 'New PC')
            # this can only happen once per python instance
            self.__class__.created = True
        else:
            self.pclist = ui.getComponent("PCList")
            self.connect = ui.getComponent("Connect")
            self.create = ui.getComponent("NewPC")


    def start(self):
        """
        Start this window. We need to get the PC information from the
        server.
        """
        self.pclist.resetList()
        i = self.ui.client
        d = i.getPCs()
        def gotPCs(pcs):
            d = self.ui.client.parent.getOneService(
                self.ui.client.loginservers)
            def gotPS(ps):
                d = []
                self.pcs = []
                def gotInfo(inf, pc):
                    self.pcs.append(self.ui.addRowToListBox(
                        self.pclist, [str(pc), "%s in %s (%s)" % 
                        (inf[0], inf[1], inf[2])]))
                for pc in pcs:
                    d.append(ps.getPCInfo(pc).addCallback(gotInfo, pc))
                return defer.gatherResults(d).addCallback(self.startDone)
            return d.addCallback(gotPS)
        d.addCallback(gotPCs)
        return d.addErrback(self.error)


    def startDone(self, r):
        if self.deferred is None:
            self.deferred = defer.Deferred()
        self.ui.displayWindow(self.pclist)
        self.ui.displayWindow(self.connect)
        self.ui.displayWindow(self.create)
        self.displayed = True
        self.connectcb = self.ui.requestUICallback(
            "Connect", "EventClicked", self, "connectPressed")
        self.createcb = self.ui.requestUICallback(
            "NewPC", "EventClicked", self, "createPressed")
        return self.deferred


    def error(self, e):
        """
        Handle a generic error by popping up a window.
        """
        e.printBriefTraceback()
        self.inerror = 1
        d = QuickMessage(self.ui, e.getErrorMessage(), "Problem Connecting!")
        self.stop()
        return d.addCallback(lambda x: self.deferred.errback(e))


    def connectPressed(self, args):
        """
        UI Callback that gets run whenever the user hits the connect
        button.
        """
        if self.pclist.selectedCount == 0:
            return
        sel = self.pclist.getFirstSelectedItem().text.c_str()
        sel = sel.rstrip(")").lstrip("(").split(",")
        pcid = tuple([int(s) for s in sel])
        self.stop()
        d = self.ui.client.connectPC(pcid)
        d.addCallback(lambda r: self.deferred.callback(r))
        d.addErrback(self.error)


    def createPressed(self, *args):
        """
        UI Callback that gets run whenever the user hits the create
        button. 
        """
        self.stop()
        d = self.ui.client.createCharacter(0)
        def created(_):
            self.pclist.resetList()
            self.start()
        d.addCallback(created)
        d.addErrback(self.error)


    def stop(self):
        """
        Hide the windows and remove callbacks
        """
        if not self.displayed:
            return
        self.ui.hideWindow(self.pclist)
        self.ui.hideWindow(self.connect)
        self.ui.hideWindow(self.create)
        self.ui.stopUICallback(self.connectcb)
        self.ui.stopUICallback(self.createcb)
        self.displayed = False



class LoginWindow(UserInterface):
    """
    The login window allows the player to select a server and provide
    authentication details
    """
    ui = None
    window = None
    logincb = None
    deferred = None
    displayed = False
    timeout = None
    loaded = False


    def initialize(self, ui):
        """
        Load the ui or get a pointer to it if it has been loaded 
        """
        self.ui = ui
        if not self.loaded:
            ui.loadLayout("LoginWindow.layout", "")
            self.__class__.loaded = True
        self.window = ui.getComponent("ConnectWindow")
        ui.getComponent("ConnectWindow/Password").textMasked = True


    def finalize(self):
        """
        Kill all the callbacks and such
        """
        self.ui.stopUICallback(self.logincb)


    def start(self):
        """
        Display the window
        """
        self.ui.displayWindow(self.window)
        self.displayed = True
        self.fillServerBox()
        self.logincb = self.ui.requestUICallback("ConnectWindow/Connect",
            "EventClicked", self, "loginPressed")
        self.deferred = defer.Deferred()
        return self.deferred


    def stop(self):
        """
        Hide the window
        """
        if self.displayed:
            self.ui.hideWindow(self.window)
            self.ui.stopUICallback(self.logincb)
        self.displayed = False


    def fillServerBox(self):
        """
        Fill up the url listbox
        """
        box = self.ui.getComponent("ConnectWindow/Server")
        box.resetList()
        self.svrs = []
        for s in self.ui.client.loginservers:
            self.svrs.append(self.ui.addItemToListBox(box, s.toString()))
        box.text = self.ui.client.loginservers[0].toString()
        self.ui.getComponent(
            "ConnectWindow/Login").text = self.ui.client.login
        self.ui.getComponent(
            "ConnectWindow/Password").text = self.ui.client.password


    def loginPressed(self, args):
        """
        UI Callback that gets activated when the user clicks the 
        login button
        """
        host = self.ui.getComponent("ConnectWindow/Server").text.c_str()
        user = str(self.ui.getComponent("ConnectWindow/Login").text.c_str())
        passwd = str(self.ui.getComponent("ConnectWindow/Password"
            ).text.c_str())
        rs = ServiceLoc(host)
        rs.creds = [user, passwd]
        rs.name = u"UserEntered"
        self.rs = rs
        if not rs in self.ui.client.loginservers:
            self.ui.client.loginservers.append(rs)
        r = self.ui.client.parent.getService(rs)
        self.stop()
        self.timeout = reactor.callLater(15, self.timeOut)
        r.addCallback(self.loggedIn)
        r.addErrback(self.error)


    def timeOut(self):
        """
        This function is called when the connect attempt times out
        """
        self.timeout = None
        self.error(Failure(DefaultException("Timed out!")))


    def error(self, e):
        """
        If there is an error while connecting, this method will report
        it to the user and then errback our deferred
        """
        e.printBriefTraceback()
        if self.timeout:
            self.timeout.cancel()
            self.timeout = None
        d = QuickMessage(self.ui, e.getErrorMessage(), "Problem Connecting!")
        self.stop()
        d.addCallback(lambda _: self.deferred.errback(e))


    def loggedIn(self, con):
        """
        Called when logging in is completed. This will make sure everything
        is finalized        
        """
        if self.timeout:
            self.timeout.cancel()
            self.timeout = None
        self.stop()
        self.deferred.callback(con)
