# -*- test-case-name: mv3d.test.test_editor -*-
# Copyright (C) 2006-2008 Mortal Coil Games
# See LICENSE for details.
"""
"""

from twisted.internet.defer import inlineCallbacks

from mv3d.client.ui.layout import SingleWidget, Grid, TopLayout

class TopLevelEditor:
    """
    Defines the top level of the in game editor
    """
    visible = False
    editors = None
    window = None
    subWindows = None

    def initialize(self, ui, playerClient):
        """
        Initialize and create the ui if needed
        """
        self.ui = ui
        self.playerClient = playerClient
        self.subWindows = {}
        self.subButtons = {}
        self.editorButtons = {}
        d = self.playerClient.do("startEditor")
        def gotEditors(editors):
            self.editors = editors
            editorButtons = []
            for k in self.editors.keys():
                if len(self.editors[k].keys()):
                    editorButtons.append(SingleWidget(k,
                        "WindowsLook/Button", text=k))
                    self.editorButtons[k] = editorButtons[-1]
                    self.createSubwindow(k)
            if len(editorButtons):
                self.window = SingleWidget("TopLevelEditor",
                    "WindowsLook/Static", Grid([editorButtons]))
                self.window.build(ui, TopLayout(), (0,0), (1.0,0.05))
        return d.addCallback(gotEditors)


    def createSubwindow(self, nm):
        """
        Create a sub window for an editor
        """
        editor = self.editors[nm]
        numEditors = len(self.editors)
        offs = (1.0 / numEditors)
        pos = offs * self.editors.keys().index(nm)
        subButtons = []
        self.subButtons[nm] = dict()
        for k in editor.keys():
            subButtons.append([SingleWidget(k,
                "WindowsLook/Button", text=k)])
            self.subButtons[nm][k] = subButtons[-1][0]
        self.subWindows[nm] = SingleWidget("EditorSub %s" % nm,
            "WindowsLook/Static", Grid(subButtons))
        self.subWindows[nm].build(self.ui, TopLayout(),
            (pos, 0.05), (offs, 0.05 * len(editor)))


    def start(self):
        """
        Display the window
        """
        if self.window is None:
            return
        if not self.visible:
            self.window.show()
            for nm, b in self.subButtons.items():
                for k, s in b.items():
                    s.addCallback("EventClicked",
                        self.subButtonClicked, nm, k)
            for k, e in self.editorButtons.items():
                e.addCallback("EventClicked",
                    self.buttonClicked, k)
        self.visible = True


    def stop(self):
        """
        Hide the window
        """
        if self.window is None:
            return
        if self.visible:
            self.window.hide()
        self.visible = False


    def toggleVisibility(self):
        """
        Toggle between visible and hidden
        """
        if self.window is None:
            return
        if self.visible:
            self.stop()
        else:
            self.start()


    def buttonClicked(self, name):
        """
        Callback that is called when a top level button
        is clicked
        """
        self.subWindows[name].toggleVisibility()


    @inlineCallbacks
    def subButtonClicked(self, name, subName):
        """
        Callback that is called when a sub level button
        is clicked
        """
        editor = self.editors[name][subName]
        uiConstructor = yield editor.callRemote("getUI")
        ui = yield uiConstructor.construct()
        ui.initialize(self.ui, editor)
        ui.start()



class BoxCreatorView:
    """
    The client side view for the box creator modifier plugin
    """
    ui = None
    modifier = None

    def initialize(self, ui, modifier):
        """
        Initialize this with the modifier's remote reference
        """
        if self.ui is not None:
            return
        self.ui = ui
        self.modifier = modifier
        pos = [
            SingleWidget("Xlbl", "WindowsLook/StaticText",
                text="X:"),
            SingleWidget("X", "WindowsLook/Editbox", text="0.0"),
            SingleWidget("Ylbl", "WindowsLook/StaticText",
                text="Y:"),
            SingleWidget("Y", "WindowsLook/Editbox", text="50.0"),
            SingleWidget("Zlbl", "WindowsLook/StaticText",
                text="Z:"),
            SingleWidget("Z", "WindowsLook/Editbox", text="0.0")
            ]
        self.pos = [pos[1], pos[3], pos[5]]
        asset = [
            SingleWidget("Albl", "WindowsLook/StaticText",
                text="Asset:"),
            SingleWidget("Asset", "WindowsLook/Editbox", text="(0,17)")
            ]
        self.asset = asset[1]
        go = [
            SingleWidget("Create", "WindowsLook/Button",
                text="Create")
            ]
        self.window = SingleWidget("BoxCreatorWindow",
            "WindowsLook/FrameWindow", Grid([[], pos, asset, go]),
            text="Create Box")
        self.window.build(ui, TopLayout(), (0.2, 0.3),
            (0.3, 0.2))
        self.window.addCallback("EventCloseClicked",
            self.window.hide)
        go[0].addCallback("EventClicked",
            self.buttonClicked)


    def start(self):
        """
        Display the window
        """
        self.window.show()


    def stop(self):
        """
        Hide the window
        """
        self.window.hide()


    def buttonClicked(self):
        """
        Called when our create button is clicked
        """
        pos = [float(p.getText()) for p in self.pos]
        self.modifier.callRemote("createBox", tuple(pos))