# -*- test-case-name: mv3d.test.test_layout -*-
# Copyright (C) 2008 Mortal Coil Games
# See LICENSE for details.
"""
"""

from zope.interface import Interface, implements



class ILayout(Interface):
    """
    Defines a GUI layout specification
    """
    def build(ui, parent, pos, size):
        """
        Actually build the UI if it hasn't been done
        """


    def addChildWidget(child):
        """
        Add a child to this layout
        """


    def show():
        """
        Display this layout
        """


    def hide():
        """
        Hide this layout
        """


    def removeCallbacks():
        """
        Remove callbacks from the elements of the layout
        and any sub layouts.
        """



class TopLayout(object):
    """
    The top most layout
    """
    implements(ILayout)
    name = "Root"


    def build(self, ui, parent, pos, size):
        raise AttributeError("TopLayout is not buildable")


    def addChildWidget(self, child):
        """
        Add a top level child widget (no need to do anything.
        """



class CallbackHandler(object):
    def __init__(self, f, *a, **kw):
        self.function = f
        self.args = a
        self.keywords = kw


    def call(self, args):
        self.function(*self.args, **self.keywords)



class SingleWidget(object):
    """
    A layout for a single component
    """
    implements(ILayout)

    widget = None
    parent = None
    ui = None
    visible = False

    def __init__(self, name, widgetType, subLayout=None,
            **properties):
        self.name = name
        self.widgetType = widgetType
        self.subLayout = subLayout
        self.properties = properties
        self.callbacks = {}


    def build(self, ui, parent, pos, size):
        """
        Build this one widget if it hasn't been built already
        Then build the subLayout
        """
        self.ui = ui
        self.parent = parent
        myName = "/".join([parent.name, self.name])
        try:
            self.widget = ui.getComponent(myName)
        except RuntimeError:
            self.widget = ui.createComponent(self.widgetType,
                myName, pos, size, **self.properties)
            parent.addChildWidget(self.widget)
        if self.subLayout is not None:
            self.subLayout.build(ui, self, (0,0), (1.0,1.0))


    def addChildWidget(self, child):
        """
        Make child one of our widget's children
        """
        self.widget.addChildWindow(child)


    def show(self):
        """
        Show this widget
        """
        if not self.visible:
            self.ui.displayWindow(self.widget)
        self.visible = True


    def hide(self, removeCallbacks=True):
        """
        Hide this widget and remove callbacks
        on it and its children
        """
        if self.visible:
            self.ui.hideWindow(self.widget)
        if removeCallbacks:
            self.removeCallbacks()
        self.visible = False


    def toggleVisibility(self):
        """
        Show or hide
        """
        if self.visible:
            self.hide()
        else:
            self.show()


    def addCallback(self, event, f, *a, **kw):
        """
        Add a callback for an event happening to 
        our widget
        """
        cbh = CallbackHandler(f, *a, **kw)
        self.callbacks[event] = self.ui.requestUICallback(
            self.widget, event, cbh, "call")


    def remCallback(self, event):
        """
        Remove a callback for an event
        """
        self.ui.stopUICallback(self.callbacks[event])


    def removeCallbacks(self):
        """
        Remove all callbacks
        """
        for cb in self.callbacks.values():
            self.ui.stopUICallback(cb)
        self.callbacks = {}
        if self.subLayout is not None:
            self.subLayout.removeCallbacks()


    def getText(self):
        """
        Return the text of this widget
        """
        return self.widget.text.c_str()



class Grid(object):
    """
    A layout containing a grid of layouts
    """
    implements(ILayout)

    name = None
    parent = None

    def __init__(self, layouts):
        self.layouts = layouts


    def build(self, ui, parent, pos, size):
        """
        Build the grid
        """
        self.parent = parent
        self.name = parent.name
        width = 0
        for l in self.layouts:
            if len(l) > width:
                width = len(l)
        height = len(self.layouts)
        step = (size[0] / width, size[1] / height)
        spos = pos[0]
        for lrow in self.layouts:
            for layout in lrow:
                layout.build(ui, parent, pos, step)
                pos = (pos[0] + step[0], pos[1])
            pos = (spos, pos[1] + step[1])


    def addChildWidget(self, child):
        """
        Make child one of our parent's children
        """
        self.parent.addChildWidget(child)


    def show(self):
        """
        Show all the layouts
        """
        for row in self.layouts:
            for l in row:
                l.show()

    def hide(self):
        """
        Hide all the layouts
        """
        for row in self.layouts:
            for l in row:
                l.hide()


    def removeCallbacks(self):
        """
        Remove callbacks on all our widgets
        """
        for row in self.layouts:
            for l in row:
                l.removeCallbacks()