# -*- test-cast-name: mv3d.test.test_renderer -*-
# Copyright (C) 2006-2008 Mortal Coil Games
# See LICENSE for details.
"""

Code related to the ogre 3d engine
"""

import os
from time import time

from twisted.internet import defer

from ogre.renderer import OGRE as ogre
from ogre.io import OIS
from ogre.addons import ogreforests as forests
from ogre.addons import caelum

from zope.interface import Interface, implements

from mv3d.client.ui.image import Compositor
from mv3d.util import math3d as VectorMath


initialized = False


class FakeLog(ogre.LogListener):
    def __init__(self):
        ogre.LogListener.__init__(self)
        self.logfile = open("ogre.log", "a")
                 
    def messageLogged(self, message, level, debug, logName):
        self.logfile.write(message + "\n")
        self.logfile.flush()



class RendererIterator:
    """
    Just a simple class to iterate the renderer
    """
    def __init__(self, parent):
        self.parent = parent
        self.count = 0
        self.lastiter = time()


    def __iter__(self):
        """
        Just return ourselves because we are an iterator
        """
        return self


    def next(self):
        """
        You need to add a docstring here ! 
        """
        self.parent.iterateRenderer(0.02)
        # self.parent.frame += 1



class OgreRenderer(ogre.FrameListener, 
    ogre.WindowEventListener, OIS.KeyListener, OIS.MouseListener):
    """
    Manages the OgreRenderer
    """
    run = 1
    renderWindow = 0
    renderloop = 0
    camera = 0
    sceneManager = 0
    fps = 0
    config = None


    def __init__(self):
        ogre.FrameListener.__init__(self)
        ogre.WindowEventListener.__init__(self)
        self.lastframes = []
        self.animations = []
        self.nodes = []
        self.frameCallbacks = []


    def configure(self, cf):
        """
        Configure the renderer with this config file
        """
        self.config = cf


    def initialize(self, **kw):
        """
        initialize the ogre engine
        """
        global initialized
        if initialized:
            return False
        initialized = True
        displayconfig = kw.get("configwindow", self.config.getboolean("Graphics", "optionsWindow"))
        openwindow = kw.get("window", True)
        plugins = os.getcwd() + os.sep + "plugins.cfg"
        if self.config.has_option("Graphics", "plugins"):
            plugins = self.config.get("Graphics", "plugins")
        #cleaner output
        self.lman = ogre.LogManager()
        self.currentLog = ogre.LogManager.getSingletonPtr().createLog(
            "fake.log", True, False, True)
        self.fakelog = FakeLog()
        self.currentLog.addListener ( self.fakelog)
        self.root = ogre.Root(plugins)
        self.root.setFrameSmoothingPeriod (5.0)
        config = ogre.ConfigFile()
        config.load('resources.cfg' ) #, '', False )
        seci = config.getSectionIterator()
        while (seci.hasMoreElements()):
            SectionName = seci.peekNextKey()
            Section = seci.getNext()
            for item in Section:
                ogre.ResourceGroupManager.getSingleton().\
                    addResourceLocation(item.value, item.key, SectionName)
        if not displayconfig:
            # this has changed, see test_visual for working code
            renList = self.root.getAvailableRenderers()
            renderer = self.config.get("Graphics", "system")
            dims = self.config.get("Graphics", "resolution")
            width = int(dims.split("x")[0])
            height = int(dims.split("x")[1])
            fullscreen = self.config.getboolean("Graphics", "fullScreen")
            for r in renList:
                # print r.getName(), renderer, r.getName()==renderer
                if r.getName() == renderer:
                    self.root.setRenderSystem(r)
                    self.root.initialise( False )
                    if openwindow:
                        self.renderWindow = self.root.createRenderWindow( "MV3D", 
                            width, height, False )
                    break
            else:
                raise Exception("Couldn't find a valid render system")
        else:
            carryOn = self.root.showConfigDialog()
            if carryOn:
                if openwindow:
                    self.renderWindow = self.root.initialise(True, "Multiverse3D")
                else:
                    self.renderWindow = self.root.initialise(True)
            else:
                return 0
        # self.sceneManager = self.root.createSceneManager(ogre.ST_GENERIC, "ExampleSMInstance")
        self.sceneManager = self.root.createSceneManager("TerrainSceneManager")
        self.sceneManager.ambientLight = ogre.ColourValue(0.0, 0.0, 0.0)
        if self.config.has_option("Graphics", "shadowType"):
            c = self.config.get("Graphics", "shadowType")
            if c.lower() == "additive":
                self.sceneManager.shadowTechnique = ogre.SHADOWTYPE_STENCIL_ADDITIVE
            if c.lower() == "modulative":
                self.sceneManager.shadowTechnique = ogre.SHADOWTYPE_STENCIL_MODULATIVE
        if openwindow:
            # create Camera
            self.camera = self.sceneManager.createCamera('PlayerCam')
            self.camera.setPosition (0, 0, 500)
            self.camera.lookAt((0, 0, -300))
            # self.camera.setFarClipDistance(0);
            self.camera.setNearClipDistance(0.01);
            # self.camera.nearClipDistance = 5
            #self.camera.polygonMode = ogre.PM_WIREFRAME
            # Create viewports
            self.viewport = self.renderWindow.addViewport(self.camera)
            self.viewport.backgroundColour = ogre.ColourValue(0, 0, 0)
            ogre.TextureManager.getSingleton().defaultNumMipmaps = 5
            ogre.ResourceGroupManager.getSingleton().initialiseAllResourceGroups()
            self.root.addFrameListener(self)
            # Frame Listener setup input
            windowHnd = self.renderWindow.getCustomAttributeInt("WINDOW")
            self.imsys = OIS.createPythonInputSystem( [("WINDOW", str(windowHnd))] )
            self.Keyboard = self.imsys.createInputObjectKeyboard( OIS.OISKeyboard, True )
            self.Mouse = self.imsys.createInputObjectMouse( OIS.OISMouse, True )
            ss = self.getScreenDimensions()
            self.Mouse.getMouseState().width = ss[0]
            self.Mouse.getMouseState().height = ss[1]
            ogre.WindowEventUtilities.addWindowEventListener(self.renderWindow, self);
        self.windoweventutils = ogre.WindowEventUtilities()
        return 1


    def finalize(self):
        """
        You need to add a docstring here ! 
        """
        if( self.imsys ):
            self.imsys.destroyInputObjectMouse( self.Mouse )
            self.imsys.destroyInputObjectKeyboard( self.Keyboard )
            OIS.InputManager.destroyInputSystem(self.imsys)
            self.imsys=None
        # for n in list(self.nodes):
            # n.finalize(1)
        self.root.shutdown()


    def frameStarted(self, frameEvent):
        """
        You need to add a docstring here ! 
        """
        for anim in self.animations:
            anim.addTime( frameEvent.timeSinceLastFrame )
            if not anim.enabled:
                self.animations.remove(anim)
        return True


    def frameEnded(self, frameEvent):
        """
        You need to add a docstring here ! 
        """
        for f in self.frameCallbacks:
            f[0](frameEvent.timeSinceLastFrame, *f[1], **f[2])
        return True


    def addFrameCallback(self, f, *a, **kw):
        """
        Add a function to be called every frame
        """
        self.frameCallbacks.append((f, a, kw))


    def remFrameCallback(self, f, *a, **kw):
        """
        Remove a frame callback
        """
        self.frameCallbacks.remove((f, a, kw))


    def iterate(self, tm):
        """
        You need to add a docstring here ! 
        """
        if self.run:
            #print "Hi"
            self.windoweventutils.messagePump()
            self.root.renderOneFrame()
            self.Keyboard.capture()
            self.Mouse.capture()
            #OIS.PlatformManager.getSingleton().messagePump(self.renderWindow)
            self.lastframes.append(time())
            if len(self.lastframes)>100:
                self.lastframes=self.lastframes[1:]
            fo=(time()-self.lastframes[0])/len(self.lastframes)
            if fo!=0:
                self.fps=1/fo


    def createEntity(self, name, file):
        """
        You need to add a docstring here ! 
        """
        if not self.renderWindow:
            return 0
        return self.sceneManager.createEntity(name, file)


    def createParticles(self, name, file):
        """
        You need to add a docstring here ! 
        """
        if not self.renderWindow:
            return 0
        return self.sceneManager.createParticleSystem(name, file)


    def getFPS(self):
        """
        You need to add a docstring here ! 
        """
        return self.renderWindow.lastFPS


    def addAnimation(self, a):
        """
        You need to add a docstring here ! 
        """
        if not a in self.animations:
            self.animations.append(a)


    def remAnimation(self, a):
        """
        Remove an animation
        """
        if a in self.animations:
            self.animations.remove(a)


    def loadImage(self, file, group="General"):
        """
        You need to add a docstring here ! 
        """
        t=ogre.TextureManager.getSingleton().getByName(file)
        if not t:
            t=ogre.TextureManager.getSingleton().load(file, group)
        return t


    def getScreenDimensions(self):
        """
        You need to add a docstring here ! 
        """
        return (self.renderWindow.width,self.renderWindow.height)

    
    def addNode(self, node):
        """
        Add a node to our list
        """
        self.nodes.append(node)
        
        
    def remNode(self, node):
        """
        Remove a node from our list
        """
        if node in self.nodes:
            self.nodes.remove(node)
            return 1
        return 0



class OgreNode:
    """
    You need to add a docstring here ! 
    """
    def __init__(self, renderer, parent=0):
        self.renderer = renderer
        if parent != 0:
            self.parent = parent
        else:
            self.parent = self.renderer.sceneManager.rootSceneNode
        self.node = 0
        self.basename = ""
        self.objects = {}


    def initialize(self, basename):
        """
        You need to add a docstring here ! 
        """
        self.basename = basename
        self.node = self.parent.createChildSceneNode(self.basename)
        self.renderer.addNode(self)


    def finalize(self, killchildren=0):
        """
        You need to add a docstring here ! 
        """
        if self.node == 0:
            return
        self.node.detachAllObjects()
        # print "finalize", self.basename
        try:
            self.parent.removeChild(self.basename)
        except:
            print "Couldn't destroy node!"
        self.renderer.sceneManager.destroySceneNode(self.basename)
        if killchildren != 0:
            for n in self.objects.values():
                n.finalize()
        self.renderer.remNode(self)
        self.node = 0
        # print "done"


    def getNode(self):
        """
        You need to add a docstring here ! 
        """
        return self.node


    def addObject(self, obj):
        """
        You need to add a docstring here ! 
        """
        if not self.node:
            return
        self.node.attachObject(obj.getObject())
        self.objects[obj.getName()] = obj


    def remObject(self, nam):
        """
        You need to add a docstring here ! 
        """
        if not self.node:
            return 0
        if self.objects.has_key(nam):
            obj=self.objects[nam]
            self.node.detachObject(obj.getObject())
            del self.objects[nam]
            return 1
        return 0


    def setPosition(self, p):
        """
        You need to add a docstring here ! 
        """
        if not self.node:
            return
        self.node.setPosition(ogre.Vector3(p[0], p[1], p[2]))


    def setRotation(self, r):
        """
        You need to add a docstring here ! 
        """
        if not self.node:
            return
        self.node.setOrientation(ogre.Quaternion(r[0], r[1], r[2], r[3]))


    def setScale(self, s):
        """
        You need to add a docstring here ! 
        """
        if not self.node:
            return
        self.node.setScale(ogre.Vector3(s[0], s[1], s[2]))


    def getPosition(self):
        """
        You need to add a docstring here ! 
        """
        if not self.node:
            return (0, 0, 0)
        return self.node.position


    def getRotation(self):
        """
        You need to add a docstring here ! 
        """
        if not self.node:
            return (0, 0, 0, 0)
        return (self.orientation.w, self.orientation.x, self.orientation.y, 
            self.orientation.z)


    def getScale(self):
        """
        You need to add a docstring here ! 
        """
        if not self.node:
            return (0, 0, 0)
        return self.node.scale


    def setParent(self, parent):
        """
        Change the parent of this node
        """
        p = self.node.getParent()
        p.removeChild(self.node)
        parent.getNode().addChild(self.node)


class IOgreObject(Interface):
    """
    Defines the interface to an Ogre Object
    """
    def create(*a, **kw):
        """
        Create the object
        """


    def destroy():
        """
        Destroy the object
        """


    def getObject():
        """
        Get the underlying object
        """


    def setObject(o):
        """
        Set the underlying object
        """



class OgrePagedGeometry:
    """
    An Ogre paged geometry instance
    """
    implements(IOgreObject)

    object = None
    paged = None

    def __init__(self, renderer, name=""):
        self.name = name
        self.renderer = renderer


    def create(self, bounds, pageSize=100, 
            batchPageLod=(100, 30), imposterPageLod=(4000, 50)):
        """
        Create the paged geometry instance
        """
        self.paged = forests.PagedGeometry()
        self.paged.setCamera(self.renderer.camera)
        self.paged.setPageSize(pageSize)
        self.paged.addDetailLevelBatchPage(*batchPageLod)
        self.paged.addDetailLevelImpostorPage(*imposterPageLod)
        self.object = forests.TreeLoader3D(self.paged,
            ogre.FloatRect(*bounds))
        self.paged.setPageLoader(self.object)
        self.renderer.addFrameCallback(self.update)


    def addGeometry(self, asset, position, yRot=0.0, scale=1.0):
        """
        Add a new piece of geometry from the specified asset
        at the given position with yRot as the rotation
        """
        entity = self.renderer.sceneManager.getEntity(asset.name)
        self.object.addTree(entity, ogre.Vector3(*position),
            ogre.Degree(yRot), scale)
        self.paged.update()


    def destroy(self):
        """
        Destroy the geometry (not sure this'll work)
        """
        self.renderer.remFrameCallback(self.update)
        self.paged = None
        self.object = None


    def update(self, tm):
        """
        Update the paged geometry
        """
        # print tm
        self.paged.update()



class OgreGrass:
    """
    An Ogre grass instance
    """
    implements(IOgreObject)

    object = None
    paged = None
    heightFunctions = []
    heightFunction = None

    def __init__(self, renderer, name=""):
        self.name = name
        self.renderer = renderer
        self.layers = []


    def create(self, bounds, heightFunc, pageSize=100, maxDist=100,
            *a, **kw):
        """
        Create the paged geometry instance
        """
        self.heightFunction = (bounds, heightFunc, a, kw)
        self.paged = forests.PagedGeometry()
        self.paged.setBounds(ogre.FloatRect(*bounds))
        self.paged.setCamera(self.renderer.camera)
        self.paged.setPageSize(pageSize)
        self.paged.addDetailLevelGrassPage(maxDist)
        self.object = forests.GrassLoader(self.paged)
        self.paged.setPageLoader(self.object)
        self.object.setHeightFunction(OgreGrass, "getHeight")
        self.heightFunctions.append(self.heightFunction)
        self.renderer.addFrameCallback(self.update)


    def addLayer(self, asset, bounds=None, animated=True,
            swaySpeed=1.5, swayLength=0.08, swayDistrib=10.0,
            density=0.01, colorMapAsset=None, densityMapAsset=None,
            minSize=(4.0, 2.0), maxSize=(8.0, 5.0), heightRange=None):
        """
        Add a layer of grass
        """
        grassLayer = self.object.addLayer(asset.name)
        grassLayer.setDensity(density)       
        grassLayer.setMinimumSize(*minSize)
        grassLayer.setMaximumSize(*maxSize)
        if bounds is not None:
            grassLayer.setMapBounds(ogre.FloatRect(*bounds))
        grassLayer.setAnimationEnabled(animated)
        grassLayer.setSwaySpeed(swaySpeed)
        grassLayer.setSwayLength(swayLength)
        grassLayer.setSwayDistribution(swayDistrib)      
        grassLayer.setFadeTechnique(forests.FADETECH_ALPHA)
        if colorMapAsset is not None:
            grassLayer.setColorMap(colorMapAsset.getFile())
            grassLayer.setColorMapFilter(forests.MAPFILTER_BILINEAR)
        if densityMapAsset is not None:
            grassLayer.setDensityMap(densityMapAsset.getFile())
            grassLayer.setDensityMapFilter(forests.MAPFILTER_NONE)
        if heightRange is not None:
            grassLayer.setHeightRange(*heightRange)
        self.layers.append(grassLayer)


    def destroy(self):
        """
        Destroy the grass
        """
        self.renderer.remFrameCallback(self.update)
        self.heightFunctions.remove(self.heightFunction)
        #enable when we upgrade to newer version of paged geom
        # for l in self.layers:
            # self.object.deleteLayer(l)
        self.paged = None
        self.object = None
        self.layers = []


    def update(self, tm):
        """
        Update the grass
        """
        try:
            self.paged.update()
            self.object.updateAnimation()
        except:
            import traceback
            traceback.print_exc()


    @classmethod
    def getHeight(cls, x, z, data=None):
        """
        Determine if x, z is in our list of bounds. If so,
        return the results of the height getter function for
        that coord
        """
        for bounds, f, a, kw in cls.heightFunctions:
            if x >= bounds[0] and x <= bounds[2]:
                if z >= bounds[1] and z <= bounds[3]:
                    return f((x, z), *a, **kw)
        return 0.0



class CaelumSky:
    """
    This implements a Caelum based sky
    """
    implements(IOgreObject)


    caelumSystem = None

    # Components = dict(
        # all=caelum.CaelumSystem.CAELUM_COMPONENT_SKY_COLOUR_MODEL |
        # caelum.CaelumSystem.CAELUM_COMPONENT_SUN |
        # caelum.CaelumSystem.CAELUM_COMPONENT_SOLAR_SYSTEM_MODEL |
        # caelum.CaelumSystem.CAELUM_COMPONENT_SKY_DOME |
        # caelum.CaelumSystem.CAELUM_COMPONENT_STARFIELD |
        # caelum.CaelumSystem.CAELUM_COMPONENT_CLOUDS |
        # caelum.CaelumSystem.CAELUM_COMPONENT_GROUND_FOG | 0,

        # skyColor=caelum.CaelumSystem.CAELUM_COMPONENT_SKY_COLOUR_MODEL,
        # sun=caelum.CaelumSystem.CAELUM_COMPONENT_SUN,
        # solarSystem=caelum.CaelumSystem.CAELUM_COMPONENT_SOLAR_SYSTEM_MODEL,
        # skyDome=caelum.CaelumSystem.CAELUM_COMPONENT_SKY_DOME,
        # starField=caelum.CaelumSystem.CAELUM_COMPONENT_STARFIELD,
        # clouds=caelum.CaelumSystem.CAELUM_COMPONENT_CLOUDS,
        # fog=caelum.CaelumSystem.CAELUM_COMPONENT_GROUND_FOG,
        # )
        

    def __init__(self, renderer):
        self.renderer = renderer


    def create(self, components=("all",)):
        """
        Create the caelum sky system
        """
        if self.caelumSystem is not None:
            return
        r = self.renderer
        fadeColour = (0.8, 0.8, 0.9)
        r.sceneManager.setFog( ogre.FOG_LINEAR, fadeColour, 0.001, 1000, 2500)
        self.caelumSystem = caelum.CaelumSystem(r.root, r.sceneManager)
        
        ## Setup sun position model.
        self.spm = caelum.SimpleSunPositionModel (ogre.Degree (13))
        ## IMHO EarthSunPositionModel is broken, so it default to sspm.
        ##caelum.SunPositionModel *spm = new caelum.EarthSunPositionModel (ogre.Radian (ogre.Degree (37)), 0.5)
        
        ## Setup sun options
        self.caelumSystem.getSun ().setAmbientMultiplier (ogre.ColourValue(0.5, 0.5, 0.5, 100.0))
        self.caelumSystem.getSun ().setDiffuseMultiplier (ogre.ColourValue(3, 3, 2.7, 100.0))
        ## For green terrain:
        #self.caelumSystem.getSun ().setDiffuseMultiplier (ogre.ColourValue(0.1, 3, 0.1))
        self.caelumSystem.getSun ().setSpecularMultiplier (ogre.ColourValue(0.1, 0.1, 0.1))
        self.caelumSystem.getSun ().setSunPositionModel (self.spm)
        self.caelumSystem.getSun ().setManageAmbientLight (True)

        sun = self.caelumSystem.getSun()
        
        ## Basic fogging setup.
        ## This is a hack until proper fogging is implemented.
        # self.caelumSystem.getSkyColourModel().setFogDensityMultiplier(0.0015)
        # self.caelumSystem.setManageFog(True)
        
        ## Setup cloud options

        self.caelumSystem.getClouds ().setCloudSpeed(ogre.Vector2(-0.00001, 0.00001))
        clouds = self.caelumSystem.getClouds()
        self.caelumSystem.getClouds().setCloudMassOffset(ogre.Vector2(0.5, 0.2))
        self.caelumSystem.getClouds().setCloudCover(0.5)
        self.caelumSystem.getClouds().setCloudBlendTime(7200 * 24)
        self.caelumSystem.getClouds().setCloudMassBlend(0.9)
        self.caelumSystem.getClouds().update(self.caelumSystem.getUniversalClock().getCurrentTime(), 
            ogre.Vector3.UNIT_Y, ogre.ColourValue(0.5, 0.5, 0.5), 
            ogre.ColourValue(3, 3, 2.7))
        # self.caelumSystem.getClouds().setAnimating(False)
        self.caelumSystem.getUniversalClock ().setTimeScale (256)
        ogre.ControllerManager.getSingleton().setTimeFactor (256)
        # print dir(self.caelumSystem.getClouds())
        
        
        ## Setup starfield options
        self.caelumSystem.getStarfield ().setInclination (ogre.Degree (13))
        
        ## Set some time parameters
        self.caelumSystem.getUniversalClock ().setTimeScale (16)
        self.caelumSystem.getUniversalClock ().setCurrentTime (28000) ## Jan 1st, 5am


        ## Register all to the render window
        r.renderWindow.addListener(self.caelumSystem)

        if 0: # new version of caelum
            capabilities = ogre.Root.getSingleton().getRenderSystem().capabilities
            if not capabilities.hasCapability(ogre.RSC_VERTEX_PROGRAM) or\
                        not capabilities.hasCapability(ogre.RSC_FRAGMENT_PROGRAM):
                skyDome = False
                clouds = False
            else:
                ogre.CompositorManager.getSingleton().addCompositor(self.renderer.renderWindow.getViewport(0),"Bloom")
            comps = 0
            for comp in components:
                comps |= self.Components[comp]
            # print comps
            componentMask = caelum.CaelumSystem.CaelumComponent(comps)
            self.caelumSystem = caelum.CaelumSystem(self.renderer.root, self.renderer.sceneManager, componentMask )
                    
            ## KNOWN BUG: The horizon is pure white if setManageFog is False.
            ## I blame it on the transparent skydome.
            self.caelumSystem.setManageSceneFog(True)
            self.caelumSystem.setSceneFogDensityMultiplier(0.0015)
            
            ## Setup sun options
            if self.caelumSystem.getSun() is not None:
                self.caelumSystem.getSun ().setAmbientMultiplier (
                    ogre.ColourValue(0.5, 0.5, 0.5))
                self.caelumSystem.getSun ().setDiffuseMultiplier (
                    ogre.ColourValue(3, 3, 2.7))
                ## For green terrain:
                # self.caelumSystem.getSun ().setDiffuseMultiplier (
                # ogre.ColourValue(0.1, 3, 0.1))
                self.caelumSystem.getSun ().setSpecularMultiplier (
                    ogre.ColourValue(5, 5, 5))
                self.caelumSystem.getSun ().setManageAmbientLight (True)

            ## Setup fog options.
            if self.caelumSystem.getGroundFog() is not None:
               self.caelumSystem.getGroundFog().findFogPassesByName()

            ## Setup cloud options.
            ## Tweak these settings to make the demo look pretty.
            if self.caelumSystem.getClouds() is not None:
                self.caelumSystem.getClouds().setCloudSpeed(
                    ogre.Vector2(0.000005, -0.000009))
                self.caelumSystem.getClouds().setCloudBlendTime(3600 * 24)
                self.caelumSystem.getClouds().setCloudCover(0.4)

            ## Setup starfield options
            if self.caelumSystem.getStarfield() is not None:
               self.caelumSystem.getStarfield().setInclination (ogre.Degree (13))

            ## Set time acceleration.
            self.caelumSystem.getUniversalClock().setTimeScale (32)
            ## Winter dawn in the southern hemisphere, looking north
            self.caelumSystem.getUniversalClock().setGregorianDateTime (
                2008, 7, 4, 20, 33, 0)

            if self.caelumSystem.getSolarSystemModel() is not None:
                ## Sidney
                self.caelumSystem.getSolarSystemModel().setObserverLongitude (
                    ogre.Degree(151 + 12.0 / 60))        
                self.caelumSystem.getSolarSystemModel().setObserverLatitude (
                    ogre.Degree(-33 + 52.0 / 60))
                ## Beyond the southern polar circle, no sunrise
                ##mCaelumSystem.getSolarSystemModel ().setObserverLatitude (ogre.Degree(-70))
                ## Beyond the northern polar circle, no sunset
                ##mCaelumSystem.getSolarSystemModel ().setObserverLatitude (ogre.Degree(70))
            
            ## Register all to the render window
            # self.renderer.renderWindow.addListener(self.caelumSystem)


    def destroy(self):
        """
        Destroy the sky system
        """
        self.caelumSystem.shutdown(False)
        self.caelumSystem = None


    def getObject(self):
        """
        Return our object
        """
        return self.caelumSystem


    def setObject(self, o):
        """
        Set our object
        """
        if self.caelumSystem is not None:
            self.destroy()
        self.caelumSystem = o


    def setCloudCover(self, percent):
        """
        Set the amount of clouds
        """
        self.caelumSystem.getClouds().setCloudCover(percent)


    def setDate(self, yr, month, day, hr, min, sec):
        """
        Set the date
        """
        self.caelumSystem.getUniversalClock().setGregorianDateTime(
            yr, month, day, hr, min, sec)


    def setSpeed(self, tm):
        """
        Set the time dilation. (Probably real time / tm)
        """
        self.caelumSystem.getUniversalClock().setTimeScale (tm)



class OgreObject:
    """
    You need to add a docstring here ! 
    """
    def __init__(self, renderer, name=""):
        self.name = name
        self.renderer = renderer
        self.object = 0
        self.asset = 0
        self.animations = {}


    def setName(self, n):
        """
        You need to add a docstring here ! 
        """
        self.name = n


    def getName(self):
        """
        You need to add a docstring here ! 
        """
        return self.name


    def createLight(self, params):
        """
        You need to add a docstring here ! 
        """
        self.object = self.renderer.sceneManager.createLight(self.name)
        self.modifyLight(params)


    def modifyLight(self, params):
        """
        You need to add a docstring here ! 
        """
        if not isinstance(self.object, ogre.Light):
            return
        light = self.object
        lt = params.get("type", "point")
        if lt =="point":
            light.type = ogre.Light.LT_POINT
        if lt =="directional":
            light.type = ogre.Light.LT_DIRECTIONAL
        if lt =="spot":
            light.type = ogre.Light.LT_SPOTLIGHT
        light.setDiffuseColour(params.get("diffuse", ogre.ColourValue(1.0, 1.0, 1.0)))
        light.setSpecularColour(params.get("specular", ogre.ColourValue(1.0, 1.0, 1.0)))
        light.setPosition(params.get("position", ogre.Vector3(0, 0, 0)))
        if params.has_key("direction"):
            light.setDirection(params.get("direction"))
        if params.has_key("spotlight params"):
            p = params["spotlight params"]
            inner = ogre.Radian(p.get("inner", 0.2))
            outer = ogre.Radian(p.get("outer", 0.3))
            light.setSpotlightRange(inner, outer, p.get("falloff", 1.0))


    def createFromMesh(self, name, mesh):
        """
        Create from a local mesh file
        """
        self.object = self.renderer.sceneManager.createEntity(name, mesh)


    def createFromAsset(self, asset):
        """
        You need to add a docstring here ! 
        """
        d = asset.create(self.name)
        if isinstance(d, defer.Deferred):
            d.addCallback(self.createFromAsset_Created)
            return d
        return self.createFromAsset_Created(d)


    def createFromAsset_Created(self, o):
        """
        You need to add a docstring here ! 
        """
        if o ==0:
            return 0
        self.object = o
        return 1


    def createManual(self, name, **kw):
        """
        Create a manual mesh.. Keywords:
            * verts: an array of vertices (3 tuple)
            norms: an array of vertex normals (3 tuple) (no need to normalize)
            colors: an array of vertex colors (3 tuple)
            uvs: an array of vertex uv points (3 tuple)
            tris: an array of triangles (3 tuple of vertex indexes)
            * material: an asset that has a getMaterialName() function
            identityprojection: set to true to use identity projection
            identityview: set to true to use identityview
            renderoperation: set to various render options
                (Something like ogre.RenderOperation.OT_LINE_LIST/OT_LINE_STRIP/OT_POINT_LIST/OT_TRIANGLE_FAN/LIST/STRIP)
            recreate: set this to recreate the object instead of making a new one
        """
        recreate = kw.get("recreate")
        # import traceback
        if not recreate:
            msh = self.renderer.sceneManager.createManualObject(name)
        else:
            if self.object:
                raise Exception("Whattha!")
                # print "WATGTHA ----------------", self.object.name
                # traceback.print_stack()
                msh = self.object
                msh.clear()
            else:
                    # print "NAAAME-----", name
                # try:
                    # print "Has entite", self.renderer.sceneManager.hasEntity(name)
                    if self.renderer.sceneManager.hasManualObject(name):
                        msh = self.renderer.sceneManager.getManualObject(name)
                        msh.clear()
                    else:
                        msh = self.renderer.sceneManager.createManualObject(name)
                # except:
                    # msh = self.renderer.sceneManager.createManualObject(name)
        self.object = msh
        self.object.setUseIdentityProjection(kw.get("identityprojection", False))
        self.object.setUseIdentityView(kw.get("identityview", False))
        verts = kw["verts"]
        norms = kw.get("norms", [])
        color = kw.get("colors", [])
        uv = kw.get("uvs", None)
        tris = kw.get("tris", [])
        material = kw["material"]
        if kw.get("renderoperation") != None:
            msh.begin(material.getFilename(), kw["renderoperation"])
        else:
            msh.begin(material.getMaterialName())
        for x in range(len(verts)):
            msh.position(*verts[x])
            if len(norms):
                n = VectorMath.normalizeVector(norms[x])
                #print n
                msh.normal(n[0], n[1], n[2])
            if len(color):
                msh.colour(*color[x])
            if len(uv):
                msh.textureCoord(uv[x][0], uv[x][1])
        for x in range(len(tris)):
            msh.triangle(*tris[x])
        msh.end()
        self.object = msh
        # print "done normo"


    def createFromMeshGen(self, name, meshGen, materialAsset, lodfactor=1.0):
        """
        Create a mesh from a meshGen
        """
        assert self.object == 0
        d = meshGen.genMesh(lodfactor=lodfactor, justTris=False)
        def madeMesh(mesh):
            # print "Generating mesh!", len(mesh.verts)
            # print mesh.verts, mesh.tris, mesh.uvs
            self.createManual(name, material=materialAsset, verts=mesh.verts,
                tris=mesh.tris, norms=mesh.norms, uvs=mesh.uvs, recreate=True)
        d.addCallback(madeMesh)
        return d


    def createBillboardSet(self, name, **kw):
        """
        Create a billboardset. Keywords:
            * material: an asset that has a getMaterialName() function
            billboards: an array of hashes each hash should have:
            * position: the position for the billboard
            color: the color for the billboard (r,g,b,a)
            rotation: a raidian value for the rotation
            dimensions: (w,h) width and height
        """
        self.object = self.renderer.sceneManager.createBillboardSet(name)
        self.object.materialName = kw["material"]
        for bbd in kw.get("billboards", []):
            c = bbd.get("color", (1.0, 1.0, 1.0, 1.0))
            cc = ogre.ColourValue()
            cc.r = c[0]
            cc.g = c[1]
            cc.b = c[2]
            cc.a = c[3]
            bb = self.object.createBillboard(bbd["position"], cc)
            if bbd.has_key("rotation"):
                bb.rotation = ogre.Radian(bbd["rotation"])
            if bbd.has_key("dimensions"):
                bb.setDimensions(bbd["dimensions"][0], bbd["dimensions"][1])


    def createTextBillboard(self, name, **kw):
        """
        Creates a billboardset with an image that has the specified
        text on it.
        * text - the text to use
        font - the name of the font (arial.ttf)
        fontsize - the size of the font (15)
        group - group to place the material in
        size - the size in 3d (w, h)
        """
        c = Compositor()
        im = c.createText(kw.get("text"), kw.get("font", "arial.ttf"), kw.get("fontsize", 15)).convert("Ogre", name = name + "Texture")
        m = c.createNewMaterialFor(name + "Material", im, kw.get("group", ""))
        m.getTechnique(0).setLightingEnabled(False)
        m.getTechnique(0).getPass(0).setSceneBlending(ogre.SBT_TRANSPARENT_ALPHA)
        m.getTechnique(0).getPass(0).setDepthCheckEnabled(False)
        self.createBillboardSet(name, material = name + "Material", billboards = [
            {"position":(0, 0, 0), "dimensions":kw.get("size", (5, 1))}])
        #self.createBillboardSet(name, material = "Examples/OgreLogo", billboards = [
        #    {"position":(0, 40, 0), "dimensions":kw.get("size", (5, 1))}])


    def createWireBoundingBox(self, name, **kw):
        """
        Creates a wire bounding box. Parameters:
            *material - asset that gives the name of the material to use
            min - (x,y,z) of the minimum of the box
            max - (x,y,z) of the maximum of the box
            obj - an OgreObject to model the box after
        """
        if kw.has_key("obj"):
            obj = kw["obj"]
            bb = obj.object.getBoundingBox()
        else:
            bb = ogre.AxisAlignedBox(kw.get("min"), kw.get("max"))
        mn = bb.getMinimum()
        mx = bb.getMaximum()
        verts = [mn, (mn[0], mn[1], mx[2]), 
            (mn[0], mn[1], mx[2]), (mx[0], mn[1], mx[2]), 
            (mx[0], mn[1], mx[2]), (mx[0], mn[1], mn[2]), 
            (mx[0], mn[1], mn[2]), mn, 
            (mn[0], mx[1], mn[2]), (mn[0], mx[1], mx[2]), 
            (mn[0], mx[1], mx[2]), mx, 
            mx, (mx[0], mx[1], mn[2]), 
            (mx[0], mx[1], mn[2]), (mx[0], mn[1], mn[2]), 
            mn, (mn[0], mx[1], mn[2]), 
            (mn[0], mx[1], mx[2]), (mn[0], mn[1], mx[2]), 
            (mx[0], mn[1], mx[2]), mx]
        # print bb.getMinimum(), bb.getMaximum()
        class StupidMaterial:
            """
            You need to add a docstring here ! 
            """
            def __init__(self, m):
                self.matname = m
            def getFilename(self):
                """
                You need to add a docstring here ! 
                """
                return self.matname
        return self.createManual(name, verts = verts, material = StupidMaterial(kw["material"]), 
            renderoperation = ogre.RenderOperation.OT_LINE_LIST)


    def createPlane(self, name, material, distance, normal):
        """
        Creates a plane.
        """
        self.plane = ogre.MovablePlane("MoveablePlane "+name)
        self.plane.d = distance
        self.plane.normal = ogre.Vector3.UNIT_Y
        ogre.MeshManager.getSingleton().createPlane("Plane "+name, 
            ogre.ResourceGroupManager.DEFAULT_RESOURCE_GROUP_NAME, 
            self.plane, 10000.0, 10000.0, 
            1, 1, True, 1, 1.0, 1.0, ogre.Vector3.UNIT_Z,
            ogre.HardwareBuffer.HBU_STATIC_WRITE_ONLY, 
            ogre.HardwareBuffer.HBU_STATIC_WRITE_ONLY, 
            True,True
            )
        self.object = self.renderer.sceneManager.createEntity("Plane "+name, "Plane "+name)


    def setMaterial(self, name):
        """
        Set the material to use on the object
        Only works on some objects
        """
        self.object.setMaterialName(name)


    def castShadows(self, t):
        """
        You need to add a docstring here ! 
        """
        if hasattr(self.object, "castShadows"):
            self.object.setcastShadows(t)


    def finalize(self):
        """
        You need to add a docstring here ! 
        """
        self.destroy()


    def destroy(self):
        """
        You need to add a docstring here ! 
        """
        # print "Destroy object"
        if self.object ==0:
            # print "object==0"
            return
        for a in self.animations.values():
            self.renderer.remAnimation(a)
        n = self.object.__class__.__name__
        if hasattr(self.renderer.sceneManager, "destroy%s" % n):
            getattr(self.renderer.sceneManager, "destroy%s" % n)(self.object)
        else:
            print "No function to destroy", n
        self.object = 0
        # print "Done"


    def getObject(self):
        """
        You need to add a docstring here ! 
        """
        return self.object


    def setObject(self, o):
        """
        You need to add a docstring here ! 
        """
        self.object = o


    def startAnimation(self, name, loop=True, weight=100):
        """
        You need to add a docstring here ! 
        """
        if not isinstance(self.object, ogre.Entity):
            return
        if not self.object.hasSkeleton() and not self.object.hasVertexAnimation():
            return
        if not self.object.getAllAnimationStates().hasAnimationState(name):
            print "Unknown animation", name
            return
        if not self.animations.has_key(name):
            self.animations[name] = self.object.getAnimationState(name)
        if self.animations[name].getEnabled():
            if loop:
                return
        if not loop:
            self.animations[name].setTimePosition(0)
        self.animations[name].setLoop(loop)
        self.animations[name].setWeight(weight)
        self.animations[name].setEnabled(True)
        self.renderer.addAnimation(self.animations[name])
        # print "Start Animation", self.animations[name].animationName


    def stopAnimation(self, name):
        """
        You need to add a docstring here ! 
        """
        if not isinstance(self.object, ogre.Entity):
            return
        if not self.object.hasSkeleton() and not self.object.hasVertexAnimation():
            return
        if not self.object.getAllAnimationStates().hasAnimationState(name):
            print "we don't know how to", name
            return
        if not self.animations.has_key(name):
            self.animations[name] = self.object.getAnimationState(name)
        self.animations[name].setEnabled(False)
        self.renderer.remAnimation(self.animations[name])
        #print "Stop Animation", self.animations[name].animationName


    def setMaxDistance(self, dist):
        """
        You need to add a docstring here ! 
        """
        self.object.setRenderingDistance(dist)
