# Copyright (C) 2007-2008 Mortal Coil Games
# See LICENSE for details.
"""
This file contains definitions for the OgreAL related sound system
"""

from twisted.internet import defer

from ogre.sound import OgreAL


class Sound:
    """
    This class represents a single sound
    """
    def attach(self, what):
        """
        Attaches this sound to something
        """


    def detach(self, what):
        """
        Detaches this sound from something
        """


    def play(self):
        """
        Start the sound playing
        """


    def isPlaying(self):
        """
        Return true if the sound is playing
        """


    def stop(self):
        """
        Stop the sound from playing
        """


    def setPitch(self, pitch):
        """
        Set the pitch to play the sound at
        """


    def getPitch(self):
        """
        Returns the pitch the sound is playing at
        """


    def setGain(self, gain):
        """
        Set the volume of the sound
        """


    def getGain(self):
        """
        Get the volume of the sound
        """


    def setLoop(self, loop):
        """
        Make this sound loop if loop is true
        """


    def isLooping(self):
        """
        Returns true if the sound will loop when played
        """


class Listener:
    """
    This defines a class for something that hears sounds
    """
    def attach(self, what):
        """
        Attaches this sound to something
        """


    def detach(self, what):
        """
        Detaches this sound from something
        """


class SoundManager:
    """
    The overall manager of sounds
    """
    def createSound(self, name, assetid):
        """
        Acquires the asset if needed and then creates a sound based on it
        Returns a deferred.
        """


    def getSound(self, name):
        """
        Returns the named sound
        """


    def destroySound(self, name):
        """
        Kills the named sound
        """


    def createListener(self):
        """
        Creates the listener
        """


    def getListener(self):
        """
        Returns the current listener
        """


    def iterate(self, time):
        """
        Advance the sound simulation by time
        """


class OgreALSound(Sound):
    """
    This class defines a sound based on OgreAL
    """
    def __init__(self, name, sound):
        self.sound = sound
        self.name = name


    def attach(self, what):
        """
        Attaches this sound to something. In our case, an OgreNode
        """
        what.getNode().attachObject(self.sound)


    def detach(self, what):
        """
        Detaches this sound from something. In our case, an OgreNode
        """
        what.getNode().detachObject(self.name)


    def play(self):
        """
        Start the sound playing
        """
        self.sound.play()


    def isPlaying(self):
        """
        Return true if the sound is playing
        """
        return self.sound.isPlaying()


    def stop(self):
        """
        Stop the sound from playing
        """
        self.sound.stop()


    def setPitch(self, pitch):
        """
        Set the pitch to play the sound at
        """
        self.sound.setPitch(pitch)


    def getPitch(self):
        """
        Returns the pitch the sound is playing at
        """
        return self.sound.getPitch()


    def setGain(self, gain):
        """
        Set the volume of the sound
        """
        self.sound.setGain(gain)


    def getGain(self):
        """
        Get the volume of the sound
        """
        return self.sound.getGain()


    def setLoop(self, loop):
        """
        Make this sound loop if loop is true
        """
        self.sound.setLoop(loop)


    def isLooping(self):
        """
        Returns true if the sound will loop when played
        """
        return self.sound.isLooping()


class OgreALListener(Listener):
    """
    This class defines a listener based on OgreAL
    """
    def __init__(self, listener):
        self.listener = listener


    def attach(self, what):
        """
        Attaches this sound to something
        """
        what.getNode().attachObject(self.listener)


    def detach(self, what):
        """
        Detaches this sound from something
        """
        what.getNode().detachObject(self.listener)



class OgreALSoundManager(SoundManager):
    """
    This class defines a sound manager based on OgreAL
    """
    def __init__(self, server):
        self.manager = OgreAL.SoundManager()
        self.server = server
        self.listener = None

    def createSound(self, name, assetid):
        """
        Acquires the asset if needed and then creates a sound based on it
        Returns a deferred.
        """
        d = self.server.getInterface("Asset Client").acquireAsset(assetid)

        def gotAsset(asset, name):
            return OgreALSound(name, self.manager.createSound(name, asset.getFilename()))

        if isinstance(d, defer.Deferred):
            d.addCallback(gotAsset, name)
            return d
        rd = defer.Deferred()
        rd.callback(gotAsset(d, name))
        return rd

    def getSound(self, name):
        """
        Returns the named sound
        """
        return OgreALSound(name, self.manager.getSound(name))


    def destroySound(self, name):
        """
        Kills the named sound
        """
        self.manager.destroySound(name)


    def createListener(self):
        """
        Creates the listener
        """
        self.listener = OgreALListener(self.manager.createListener())
        return self.listener


    def getListener(self):
        """
        Returns the current listener
        """
        return self.listener

