# Copyright (C) 2006-2008 Mortal Coil Games
# See LICENSE for details.
# -*- test-case-name: mv3d.test.test_playerui -*-
"""
An interface to handle the movement of the player
"""

from twisted.internet import defer

from mv3d.util.binding import Action
from mv3d.util import math3d as VectorMath
from mv3d.util.conductor import ISimulationClient
from mv3d.client.ui.cegui import UserInterface
from mv3d.client.ui.connect import LoginWindow
from mv3d.client.ui.editor import TopLevelEditor
from mv3d.reporter.guilog import LogToGUI


class SimpleCharacterGenerator(UserInterface):
    """
    This simple char gen just asks for a name, and that's it.
    """
    created = False
    deferred = None
    ccb = None
    cgwindow = None
    nametext = None
    createbutton = None
    models = None

    def __init__(self, cgi):
        self.cgi = cgi
        self.modellbis = []


    def initialize(self, ui):
        self.ui = ui
        if not self.created:
            self.cgwindow = ui.createComponent("WindowsLook/FrameWindow", 
                "CG/Window", [0.2, 0.6], [0.6, 0.25], [0.01, 0.01], [1.0, 1.0],
                '')
            namelabel = ui.createComponent("WindowsLook/StaticText", 
                "CG/Name", [0.05, 0.2], [0.2, 0.15], [0.01, 0.01], [1.0, 1.0],
                'Name:')
            self.nametext = ui.createComponent("WindowsLook/Editbox", 
                "CG/NameTxt", [0.25, 0.2], [0.5, 0.15], [0.01, 0.01], [1.0, 1.0],
                '')
            modellabel = ui.createComponent("WindowsLook/StaticText",
                "CG/ModelLbl", [0.05, 0.4], [0.2, 0.15], [0.01, 0.01], [1.0, 1.0],
                'Model:')
            self.model = ui.createComponent("WindowsLook/Combobox",
                "CG/Model", [0.25, 0.4], [0.5, 0.6], [0.01, 0.01], [1.0, 1.0],
                '')
            self.createbutton = ui.createComponent("WindowsLook/Button", 
                "CG/Create", [0.8, 0.82], [0.2, 0.15], [0.01, 0.01], [1.0, 1.0],
                'Create')
            self.cgwindow.addChildWindow(namelabel)
            self.cgwindow.addChildWindow(self.nametext)
            self.cgwindow.addChildWindow(modellabel)
            self.cgwindow.addChildWindow(self.model)
            self.cgwindow.addChildWindow(self.createbutton)
            SimpleCharacterGenerator.created = True
        else:
            self.cgwindow = ui.getComponent("CG/Window")
            self.nametext = ui.getComponent("CG/NameTxt")
            self.model = ui.getComponent("CG/Model")
            self.createbutton = ui.getComponent("CG/Create")


    def start(self):
        """
        Display the window and generate the deferred that is called when
        done
        """
        self.deferred = defer.Deferred()
        self.model.resetList()
        # self.modellbis = []
        self.ui.displayWindow(self.cgwindow)
        self.ccb = self.ui.requestUICallback(self.createbutton, "EventClicked",
            self, "create")
        def gotModels(models):
            self.models = models
            for i, m in models.items():
                self.modellbis.append(self.ui.addItemToListBox(self.model, "%d: %s" % (i, m)))
            return self.deferred
        return self.cgi.callRemote("getModels").addCallback(gotModels)


    def create(self, args):
        """
        User has clicked the create button
        """
        name = self.nametext.text.c_str()
        model = int(self.model.text.c_str().split(":")[0])
        d = self.cgi.callRemote("create", name, model)
        d.addCallback(lambda pcid: self.deferred.callback(pcid))
        d.addCallback(lambda _: self.stop())


    def stop(self):
        """
        Hide the window
        """
        self.ui.hideWindow(self.cgwindow)
        self.ui.stopUICallback(self.ccb)


created = 0

class PlayerMovement(UserInterface):
    """
    Handles the client side of the player's user interface
    """

    worldeditor = None

    def initialize(self, ui):
        """
        You need to add a docstring here ! 
        """
        global created
        self.ui = ui
        self.pc = self.ui.client
        if not created:
            self.chatwindow = ui.createComponent("WindowsLook/FrameWindow", 
                "ChatWindow", [0.2, 0.6], [0.6, 0.35], [0.01, 0.01], [1.0, 1.0], '')
            self.chatwindow.alpha = 0.8
            self.entertext = ui.createComponent("WindowsLook/Editbox", 
                "Chat", [0.05, 0.82], [0.9, 0.15], [0.01, 0.01], [1.0, 1.0], '')
            self.entertext.alpha = 0.8
            self.gottext = ui.createComponent("WindowsLook/MultiLineEditbox", 
                "Message", [0.05, 0.1], [0.9, 0.70], [0.01, 0.01], [1.0, 1.0], '')
            self.gottext.alpha = 0.8
            self.gottext.showVertScrollbar = True
            self.gottext.readOnly = True
            created = 1
            self.chatwindow.addChildWindow(self.entertext)
            self.chatwindow.addChildWindow(self.gottext)
            self.statswindow = ui.createComponent("WindowsLook/FrameWindow", 
                "StatsWindow", [0.7, 0.1], [0.25, 0.35], [0.01, 0.01], [1.0, 1.0], 'Statistics')
            self.statswindow.alpha = 0.8
            self.stat = []
            for x in range(9):
                self.stat.append(ui.createComponent("WindowsLook/StaticText", 
                    "Status%d" % x, [0, x / (0.85 / 0.075) + 0.15], [1.0, 0.1], [0.01, 0.01], [1.0, 1.0], ""))
                self.statswindow.addChildWindow(self.stat[ - 1])
        else:
            self.chatwindow = ui.getComponent("ChatWindow")
            self.entertext = ui.getComponent("Chat")
            self.gottext = ui.getComponent("Message")
            self.statswindow = ui.getComponent("StatsWindow")
            self.stat = []
            for x in range(9):
                self.stat.append(ui.getComponent("Status%d" % x))
        self.showstats = 0
        self.ui.client.statsready = 1
        self.chatup = 0
        self.camdist = 15
        self.camang = (1, 0, 0, 0)
        self.walking = [0, 0, 0]
        self.turning = [0, 0, 0]
        self.actions = {}
        a = Action()
        a.setServer(self.ui.client)
        a.setName("walkForwards")
        a.setType("Local Function")
        a.setAction(self.walkForwards)
        a.addKeyBinding(self.ui, "up")
        a.addKeyBinding(self.ui, "w")
        self.actions["walkForwards"] = a
        a = Action()
        a.setServer(self.ui.client)
        a.setName("walkBackwards")
        a.setType("Local Function")
        a.setAction(self.walkBackwards)
        a.addKeyBinding(self.ui, "down")
        a.addKeyBinding(self.ui, "s")
        self.actions["walkBackwards"] = a
        a = Action()
        a.setServer(self.ui.client)
        a.setName("strafeLeft")
        a.setType("Local Function")
        a.setAction(self.strafeLeft)
        a.addKeyBinding(self.ui, "a")
        self.actions["strafeLeft"] = a
        a = Action()
        a.setServer(self.ui.client)
        a.setName("strafeRight")
        a.setType("Local Function")
        a.setAction(self.strafeRight)
        a.addKeyBinding(self.ui, "d")
        self.actions["strafeRight"] = a
        a = Action()
        a.setServer(self.ui.client)
        a.setName("turnLeft")
        a.setType("Local Function")
        a.setAction(self.turnLeft)
        a.addKeyBinding(self.ui, "left")
        self.actions["turnLeft"] = a
        a = Action()
        a.setServer(self.ui.client)
        a.setName("turnRight")
        a.setType("Local Function")
        a.setAction(self.turnRight)
        a.addKeyBinding(self.ui, "right")
        self.actions["turnRight"] = a
        a = Action()
        a.setServer(self.ui.client)
        a.setName("jump")
        a.setType("Local Function")
        a.setAction(self.jump)
        a.addKeyBinding(self.ui, "space")
        self.actions["jump"] = a
        a = Action()
        a.setServer(self.ui.client)
        a.setName("Kick")
        a.setType("Local Function")
        a.setAction(self.kick)
        a.addKeyBinding(self.ui, "k")
        self.actions["Kick"] = a
        a = Action()
        a.setServer(self.ui.client)
        a.setName("Wave")
        a.setType("Local Function")
        a.setAction(self.wave)
        a.addKeyBinding(self.ui, "h")
        self.actions["Wave"] = a
        a = Action()
        a.setServer(self.ui.client)
        a.setName("worldEditor")
        a.setType("Local Function")
        a.setAction(self.worldEditor)
        a.addKeyBinding(self.ui, "tab")
        self.actions["worldEditor"] = a
        a = Action()
        a.setServer(self.ui.client)
        a.setName("toggleStats")
        a.setType("Local Function")
        a.setAction(self.toggleStats)
        a.addKeyBinding(self.ui, "f1")
        self.actions["toggleStats"] = a
        a = Action()
        a.setServer(self.ui.client)
        a.setName("toggleChat")
        a.setType("Local Function")
        a.setAction(self.toggleChat)
        a.addKeyBinding(self.ui, "f2")
        self.actions["toggleChat"] = a
        a = Action()
        a.setServer(self.ui.client)
        a.setName("toggleChat")
        a.setType("Local Function")
        a.setAction(self.toggleChat)
        a.addKeyBinding(self.ui, "f2")
        self.actions["toggleChat"] = a
        # a = Action()
        # a.setServer(self.ui.client)
        # a.setName("Connect")
        # a.setType("Local Function")
        # a.setAction(self.connect)
        # a.addKeyBinding(self.ui, "f4")
        # self.actions["Connect"] = a
        a = Action()
        a.setServer(self.ui.client)
        a.setName("AdjustCamera")
        a.setType("Local Function")
        a.setAction(self.adjustCamera)
        a.addKeyBinding(self.ui, "lshift")
        self.actions["AdjustCamera"] = a


    def finalize(self):
        """
        You need to add a docstring here ! 
        """
        pass


    def iterate(self, time):
        """
        You need to add a docstring here ! 
        """
        pass


    def start(self):
        """
        You need to add a docstring here ! 
        """
        for a in self.actions.values():
            self.ui.client.addAction(a)
        #self.ui.displayWindow(self.chatwindow)
        self.ui.requestUICallback("Chat", "EventTextAccepted", self, "say")
        self.ui.requestUICallback(self.chatwindow, "EventCloseClicked", self, "toggleChatCB")
        #self.toggleStats(1)
        self.chatup = 0


    def stop(self):
        """
        You need to add a docstring here ! 
        """
        for a in self.actions.values():
            self.ui.client.remAction(a)
        if self.chatup:
            self.ui.hideWindow(self.chatwindow)
            self.chatup = 0


    def receiveEvent(self, e):
        """
        You need to add a docstring here ! 
        """
        # pass
        # print "Got it"
        #self.gottext.text=e.object.getName()


    def walkForwards(self, d):
        """
        You need to add a docstring here ! 
        """
        try:
            # print self.walking
            if d:
                if self.walking[2]:
                    return
                self.walking[2] = 1
                self.pc.do("Walk", 1)
            else:
                if not self.walking[2]:
                    return
                self.walking[2] = 0
                self.pc.do("Walk",  - 1)
        except:
            import traceback
            traceback.print_exc()


    def walkBackwards(self, d):
        """
        You need to add a docstring here ! 
        """
        # print self.walking
        if d:
            if self.walking[2]:
                return
            self.walking[2] = 1
            self.pc.do("Walk",  - 1)
        else:
            if not self.walking[2]:
                return
            self.walking[2] = 0
            self.pc.do("Walk", 1)


    def strafeLeft(self, d):
        """
        You need to add a docstring here ! 
        """
        if d:
            if self.walking[0]:
                return
            self.walking[0] = 1
            self.pc.do("Strafe", 1)
        else:
            if not self.walking[0]:
                return
            self.walking[0] = 0
            self.pc.do("Strafe",  - 1)


    def strafeRight(self, d):
        """
        You need to add a docstring here ! 
        """
        if d:
            if self.walking[0]:
                return
            self.walking[0] = 1
            self.pc.do("Strafe",  - 1)
        else:
            if not self.walking[0]:
                return
            self.walking[0] = 0
            self.pc.do("Strafe", 1)


    def jump(self, d):
        """
        You need to add a docstring here ! 
        """
        if d:
            self.pc.do("Jump")


    def kick(self, d):
        """
        You need to add a docstring here ! 
        """
        if d:
            self.pc.do("Kick")


    def wave(self, d):
        """
        You need to add a docstring here ! 
        """
        if d:
            self.pc.do("Wave")


    def turnLeft(self, d):
        """
        You need to add a docstring here ! 
        """
        if d:
            if self.turning[0]:
                return
            self.turning[0] = 1
            self.pc.do("Turn",  - 1)
        else:
            if not self.turning[0]:
                return
            self.turning[0] = 0
            self.pc.do("Turn", 0)


    def turnRight(self, d):
        """
        You need to add a docstring here ! 
        """
        if d:
            if self.turning[0]:
                return
            self.turning[0] = 1
            self.pc.do("Turn", 1)
        else:
            if not self.turning[0]:
                return
            self.turning[0] = 0
            self.pc.do("Turn", 0)


    def worldEditor(self, d):
        """
        Toggle the world editor. Start it up if it hasn't
        been already
        """
        if d:
            if self.worldeditor is None:
                self.worldeditor = TopLevelEditor()
                d = self.worldeditor.initialize(self.ui,
                    self.ui.client)
                d.addCallback(lambda x:
                    self.worldeditor.start())
            else:
                self.worldeditor.toggleVisibility()


    def say(self, args):
        """
        You need to add a docstring here ! 
        """
        try:
            if self.entertext.text == "":
                return
            self.pc.do("Say", self.entertext.text.c_str())
            self.entertext.text = ""
        except:
            print "Problem saying stuff!"


    def toggleChatCB(self, *args):
        """
        You need to add a docstring here ! 
        """
        self.toggleChat(1)


    def toggleStats(self, d):
        """
        You need to add a docstring here ! 
        """
        if d:
            if self.showstats:
                self.ui.hideWindow(self.statswindow)
                self.showstats = 0
            else:
                self.ui.displayWindow(self.statswindow)
                self.showstats = 1


    def toggleChat(self, d):
        """
        You need to add a docstring here ! 
        """
        if d:
            if self.chatup:
                self.ui.hideWindow(self.chatwindow)
                self.chatup = 0
            else:
                self.ui.displayWindow(self.chatwindow)
                self.chatup = 1


    def connect(self, d):
        """
        You need to add a docstring here ! 
        """
        if d:
            self.lw = LoginWindow()
            self.lw.initialize(self.ui)
            d = self.lw.start()
            d.addCallback(lambda x: 
                self.ui.client.connectPC(self.ui.client.pcid))


    def adjustCamera(self, d):
        """
        You need to add a docstring here ! 
        """
        try:
            if d == 1:
                self.ui.system.setDefaultMouseCursor("WindowsLook", "Background")
                self.ui.requestMouseCallback(self.adjustCamera_Mouse)
            else:
                self.ui.system.setDefaultMouseCursor("WindowsLook", "MouseArrow")
                self.ui.stopMouseCallback(self.adjustCamera_Mouse)
        except:
            import traceback
            traceback.print_exc()


    def adjustCamera_Mouse(self, btn, pos, rel, mods):
        """
        You need to add a docstring here ! 
        """
        try:
            if btn[0]:
                #print self.camdist
                self.camdist -= rel[2] / 10.0
                #r=VectorMath.quaternionFromMatrix(VectorMath.matrixFromEuler(0,rel[1],0))
                #self.camang=VectorMath.normalizeQuaternion(VectorMath.multiplyQuaternions(self.camang,r))
                #self.camang=(self.camang[0]+rel[1],self.camang[1]+rel[0],self.camang[2])
                vx = VectorMath.quaternionFromMatrix(VectorMath.matrixFromEuler((rel[1] / 150.0, 0, 0)))
                vy = VectorMath.quaternionFromMatrix(VectorMath.matrixFromEuler((0, rel[0] / 150.0, 0)))
                vb = VectorMath.normalizeQuaternion(VectorMath.multiplyQuaternions(vx, vy))
                va = VectorMath.normalizeQuaternion(VectorMath.multiplyQuaternions(self.camang, vb))
                self.camang = va
                v = self.ui.client.parent.getLocalService(
                    ISimulationClient).views
                for vw in v.values():
                    vw.cameraoffs = (0,  - 2, self.camdist)#VectorMath.rotateVector((0, 0, self.camdist), va)
                    vw.camerarot = va
                    vw.updateCamera()
            if rel[2] != 0.0:
                self.camdist -= rel[2] / 10.0
                v = self.ui.client.parent.getLocalService(
                    ISimulationClient).views
                for vw in v.values():
                    vw.cameraoffs = (0,  - 2, self.camdist)#VectorMath.rotateVector((0, 0, self.camdist), va)
                    vw.camerarot = self.camang
                    vw.updateCamera()
            if btn[2]:
                if self.camdist > 0.3:
                    self.camdist = 0
                else:
                    self.camdist = 15
                self.camang = (1, 0, 0, 0)
                v = self.ui.client.parent.getLocalService(
                    ISimulationClient).views
                for vw in v.values():
                    vw.cameraoffs = (0,  - 2, self.camdist)#VectorMath.rotateVector((0, 0, self.camdist), va)
                    vw.camerarot = self.camang
                    vw.updateCamera()
        except:
            import traceback
            traceback.print_exc()


    def toggleLog(self,  * args):
        """
        You need to add a docstring here ! 
        """
        if args[0] == 1:
            # print "Toggle"
            c = self.ui.client
            for l in c.eventlogs:
                if isinstance(l, LogToGUI):
                    l.toggleVisible()
