# Copyright (C) 2006-2008 Mortal Coil Games
# See LICENSE for details.
""" -*- test-case-name: Test.test_visual -*-
"""

import ode

from ogre.renderer import OGRE as ogre

from twisted.internet import defer
from twisted.internet.task import coiterate

from mv3d.net.pb import Cacheable
from mv3d.phys.body import IVisualObjectView
from mv3d.phys.scenery import Scenery, PieceIterator
from mv3d.util.math3d import Vector, Quaternion
from mv3d.util import math3d as VectorMath
from mv3d.util.conductor import IAssetClient, ISimulationClient, IPlayerClient
from mv3d.client.ui.ogre3d import OgreNode, OgreGrass
from mv3d.client.ui.ogre3d import OgreObject, OgrePagedGeometry
from mv3d.client.view.object import ObjectView


class BasicVisualObjectView(IVisualObjectView, Cacheable):
    """
    A view of a basic visual object
    """
    renderer = None
    areaid = None
    node = None
    scale = (1.0, 1.0, 1.0)
    posoffset = (0,0,0)
    rotoffset = (1,0,0,0)

    def __init__(self):
        self.objects = {}
        self.assets = []
        self.currentanimations = {}


    def getID(self):
        return self.iid


    def create(self, area, pos, rot):
        """
        Create the graphical representation
        """
        # print "---- creating", self.posoffset
        # assert self.node is None
        if self.node is not None:
            self.destroy()
        d = defer.maybeDeferred(
            self.parent.parent.getLocalService(IAssetClient).acquireAssets, 
            self.assets)
        node = area.getNode()
        self.node = OgreNode(self.renderer, node.getNode())
        self.node.initialize("Node for %s" % str(self))
        self.node.setScale(self.scale)
        self.setPosition(pos)
        self.setRotation(rot)
        def gotAssets(assets):
            if not self.node:
                # we got destroyed likely
                return
            for asset in assets:
                # print "Adding asset", asset
                object = OgreObject(self.renderer, "Object %d of %s" % 
                    (len(self.objects), str(self)))
                object.createFromAsset(asset)
                self.node.addObject(object)
                self.objects[asset.getID()] = object
                for name, params in self.currentanimations.items():
                    self.startAnimation(name, *params)
        d.addCallback(gotAssets)
        return d


    def destroy(self):
        """
        Destroy the graphical representation
        """
        # print "Destroying", self
        self.node.finalize(1)
        # self.node = None


    def enterArea(self, area):
        """
        We've entered a new area
        """
        node = area.getNode()
        self.node.setParent(node)


    observe_enterArea = enterArea


    def observe_addAsset(self, aid):
        """
        Add an asset to the object
        """
        d = self.parent.parent.getLocalService(IAssetClient).acquireAsset(aid)
        def gotAsset(asset):
            object = OgreObject(self.renderer, "Object %d of %s" %
                (len(self.objects), str(self)))
            object.createFromAsset(asset)
            self.node.addObject(object)
            self.objects[asset.getID()] = object
        d.addCallback(gotAsset)
        return d


    def observe_removeAsset(self, aid):
        """
        Remove an asset from the object
        """
        self.objects[aid].destroy()
        del self.objects[aid]


    def startAnimation(self, name, loop=False, weight=1.0):
        """
        Start an animation
        """
        # print "start anim", name
        for o in self.objects.values():
            o.startAnimation(name, loop, weight)


    def stopAnimation(self, name):
        """
        Stop an animation
        """
        # print "stop anim", name
        for o in self.objects.values():
            o.stopAnimation(name)


    observe_startAnimation = startAnimation
    observe_stopAnimation = stopAnimation


    def setPosition(self, pos):
        """
        Set the position of this object
        """
        # print self.posoffset
        self.node.setPosition(Vector(self.posoffset) + pos)


    def setRotation(self, rot):
        """
        Set the rotation of this object
        """
        # print tuple(self.rotoffset)
        self.node.setRotation(Quaternion(rot) * self.rotoffset)


    def setCopyableState(self, state):
        # print state
        Cacheable.setCopyableState(self, state)
        self.renderer = self.parent.parent.getLocalService(
            IPlayerClient).getRenderer()
        # self.create()



class WaterPlaneVisualObjectView(IVisualObjectView, Cacheable):
    """
    A view of water
    """
    renderer = None
    areaid = None
    node = None
    distance = 0
    normal = (0,1,0)

    def __init__(self):
        self.objects = {}
        self.assets = []
        self.currentanimations = {}


    def getID(self):
        return self.iid


    def create(self, area, pos, rot):
        """
        Create the graphical representation
        """
        node = area.getNode()
        self.node = OgreNode(self.renderer, node.getNode())
        self.node.initialize("Node for %s" % str(self))
        self.node.setPosition((2500, self.distance, 2500))
        self.entity = OgreObject(self.renderer, "Water %s" % str(self))
        # self.entity.createPlane("Water %s"%str(self), "OceanCg_RTT", 
        self.entity.createPlane("Water %s"%str(self), "OceanCg_RTT", 
            self.distance, self.normal)
        self.node.getNode().attachObject(self.entity.plane) 
        self.node.addObject(self.entity)
        sceneManager = self.renderer.sceneManager
        camera = self.renderer.camera

        # Attach both the plane entity, and the plane definition
        # create RenderTexture
        rttTex = self.renderer.root.getRenderSystem().createRenderTexture(
            "RttTex"+str(self)[-4:], 256, 256, 
            ogre.TEX_TYPE_2D,ogre.PixelFormat.PF_R8G8B8)
        
        self.mReflectCam = sceneManager.createCamera("ReflectCam %s" % str(self))
        self.mReflectCam.setNearClipDistance(camera.getNearClipDistance())
        # self.mReflectCam.setFarClipDistance(camera.getFarClipDistance())
        self.mReflectCam.setFarClipDistance(0)
        self.mReflectCam.enableReflection(self.entity.plane)
        self.mReflectCam.enableCustomNearClipPlaneMP(self.entity.plane) 
        w = self.renderer.renderWindow.getViewport(0).getActualWidth() 
        h = self.renderer.renderWindow.getViewport(0).getActualHeight()
        self.mReflectCam.setAspectRatio ( float(w)/float(h) )
        # create ViewPort        
        v = rttTex.addViewport( self.mReflectCam )
        v.setClearEveryFrame ( True  )
        v.setBackgroundColour (ogre.ColourValue.White)

        # oceanMat = ogre.MaterialManager.getSingleton().getByName("OceanCg_RTT")
        oceanMat = ogre.MaterialManager.getSingleton().getByName("OceanCg_RTT")
        assert oceanMat is not None, "No material OceanCg_RTT"
        mPass = oceanMat.getTechnique(0).getPass(0)
        tState = mPass.getTextureUnitState(1)
        # tState.setTexture(rttTex)
        tState.setTextureName("RttTex"+str(self)[-4:])
        tState.setColourOperationEx( ogre.LBX_BLEND_MANUAL, ogre.LBS_TEXTURE, 
                               ogre.LBS_CURRENT, ogre.ColourValue.White,
                              ogre.ColourValue.White, 0.25)
        tState.setTextureAddressingMode (ogre.TextureUnitState.TAM_CLAMP)
        
        self.wrtl = WaterRTL()
        self.wrtl.mPlane = self.entity.plane
        rttTex.addListener(self.wrtl)

        # set up linked reflection
        # Also clip       
        self.rttfl = WaterFrameListener(self.renderer.renderWindow, 
            self.renderer.camera, self.mReflectCam, self.node.getNode())
        self.renderer.root.addFrameListener(self.rttfl)
        self.entity.setMaterial("OceanCg_RTT")


    def destroy(self):
        """
        Destroy the graphical representation
        """
        self.node.finalize(1)


    def enterArea(self, area):
        """
        We've entered a new area
        """
        node = area.getNode()
        self.node.setParent(node)


    observe_enterArea = enterArea


    def observe_addAsset(self, aid):
        """
        Add an asset to the object
        """


    def observe_removeAsset(self, aid):
        """
        Remove an asset from the object
        """


    def observe_setPlane(self, distance, normal):
        self.distance = distance
        self.normal = normal


    def setCopyableState(self, state):
        Cacheable.setCopyableState(self, state)
        self.renderer = self.parent.parent.getLocalService(
            IPlayerClient).getRenderer()



class WaterRTL(ogre.RenderTargetListener):
    def preRenderTargetUpdate(self,evt):
        self.mPlane.setVisible(False)


    def postRenderTargetUpdate(self,evt):
        self.mPlane.setVisible(True)



class WaterFrameListener(ogre.FrameListener):
    def __init__(self, renderWindow, camera, mReflectCam , planeSceneNode):
        ogre.FrameListener.__init__(self)
        self.camera = camera
        self.mReflectCam = mReflectCam       
        self.planeNode = planeSceneNode

    def frameStarted(self, frameEvent):
        """
        Make sure reflection camera is updated too
        """
        self.mReflectCam.setOrientation ( self.camera.getOrientation() )
        self.mReflectCam.setPosition (self.camera.getPosition())
        return True



class BasicPhysicalObjectView(ObjectView):
    """
    A generic object it has a physical and visual
    component and pretty much just reacts to the
    environment.
    """
    setCopyableState = ObjectView.setCopyableState


    def setParent(self, parent):
        self.parent = parent
        self.body.parent = parent
        for vob in self.body.vobs:
            vob.parent = self.parent


    def create(self):
        """
        Create everything
        """
        self.body.create(self.areas.values()[0].realm.world)
        for vob in self.body.vobs:
            vob.create(self.areas.values()[0], self.body.getPosition(), self.body.getRotation())
        self.body.enterArea(self.areas.values()[0])


    def destroy(self):
        """
        Destroy our body
        """
        # print "WTF, DESTROY", self.body
        self.body.destroy()
        # self.body = None


    def isEnabled(self):
        """
        Check if our body is enabled
        """
        return self.body.isEnabled()
        
        
    def iterate(self, t):
        self.body.iterate()


    def getViewPoint(self):
        """
        You need to add a docstring here ! 
        """
        if not self.body:
            return (0,0,0),(1,0,0,0)
        return Vector(0,2,0) + self.body.getPosition(),self.body.getRotation()



class WaterPlaneObjectView(ObjectView):
    """
    A generic object it has a physical and visual
    component and pretty much just reacts to the
    environment.
    """
    setCopyableState = ObjectView.setCopyableState
    gravity = None
    depth = -1
    disabled = 1
    density = 1
    viscosity = 0.95


    def setParent(self, parent):
        self.parent = parent


    def create(self):
        """
        Create everything
        """
        for c in self.body.colliders:
            c.oncollide = self.doBoyuoancy
        self.body.create(self.areas.values()[0].realm.world)
        for vob in self.body.vobs:
            vob.create(self.areas.values()[0], self.body.getPosition(), self.body.getRotation())
        self.body.enterArea(self.areas.values()[0])


    def destroy(self):
        """
        Destroy everything
        """
        self.body.destroy()


    def isEnabled(self):
        """
        Check if our body is enabled
        """
        return self.body.isEnabled()
        
        
    def iterate(self, t):
        self.body.iterate()


    def getViewPoint(self):
        """
        You need to add a docstring here ! 
        """
        if not self.body:
            return (0,0,0),(1,0,0,0)
        return self.body.getPosition(),self.body.getRotation()


    def doBoyuoancy(self, geom, mygeom):
        """
        In this function, we'll try and apply forces to
            the body attached to the geom that has collided with 
            us.
            The basic idea for simple boyuoancy is that we apply
            f = m2 * g * ((d2-d1) - 1)
            if both densities are equal, then nothing happens.
            Otherwise, if the body colliding is more dense than us,
            it gets sucked down further. If it is less dense, it
            gets pushed up.
        """
        b = geom.getBody()
        if b is None and hasattr(geom, "body"):
            b = geom.body
        if b is None:
            return 1
        contacts = ode.collide(geom, mygeom)
        maxd = 0
        for c in contacts:
            p, n, d, g1, g2 = c.getContactGeomParams()
            if d > maxd:
                maxd = d
        # if hasattr(b, "mvbody") and hasattr(b.mvbody, "setWaterDepth"):
        # b.waterDepth = maxd
        geom.waterDepth = maxd
        if maxd == 0:
            return 1
        if maxd > 4.0:
            maxd = 4.0
        # retard the vel
        v = self.viscosity + ((1.0 - self.viscosity) * (1.0 - (maxd / 4.0)))
        b.setLinearVel(VectorMath.scaleVector(b.getLinearVel(), v))
        b.setAngularVel(VectorMath.scaleVector(b.getAngularVel(), v))
        # now push against gravity
        m = b.getMass().mass
        d1 = self.density
        d2 = geom.density
        off = (m * ((d2 - d1) - 1.0)) * (maxd / 4.0)
        if self.gravity is None:
            # print "NO GRAV"
            self.gravity = self.parent.parent.getLocalService(
                ISimulationClient).realms.values(
                )[0].getPhysicsProperty("gravity")
        f = VectorMath.scaleVector(self.gravity, off)
        # print maxd, tuple(f), self.density, geom.density, self.viscosity, self.body.colliders[0].distance
        b.addForce(f)
        return 1



class MeshGenVisualObjectView(IVisualObjectView, Cacheable):
    """
    A view of a basic visual object
    """
    renderer = None
    areaid = None
    node = None
    meshGen = None
    materialAssetId = None
    mesh = None
    scale = (1.0, 1.0, 1.0)
    posoffset = (0,0,0)
    rotoffset = (1,0,0,0)
    lodlevel = 1.0
    created = False
    parent = None


    def getID(self):
        return self.iid


    def setParent(self, parent):
        self.parent = parent


    def create(self, area, pos, rot):
        """
        Create the graphical representation
        """
        assert not self.created
        self.created = True
        # print "CREATE CREATE CREATE", id(self), self.mesh, pos
        d = self.parent.parent.getLocalService(IAssetClient).acquireAsset(
                self.materialAssetId)
        node = area.getNode()
        self.node = OgreNode(self.renderer, node.getNode())
        self.node.initialize("Node for %s" % str(self))
        self.node.setScale(self.scale)
        self.setPosition(pos)
        self.setRotation(rot)
        assert self.mesh is None, self.mesh
        mesh = OgreObject(self.renderer)
        d.addCallback(self.gotAsset, mesh)
        return d


    def gotAsset(self, asset, mesh):
        # print "GOTASSET",mesh
        self.mesh = mesh
        d = self.mesh.createFromMeshGen("MeshGenVisualObjectView" + str(id(self)), self.meshGen, 
            asset, self.lodlevel)
        def madeMesh(ignored):
            # print "adding", self, "to node", self.node
            # self.mesh.castShadows(False)
            self.node.addObject(self.mesh)
        d.addCallback(madeMesh)
        return d


    def destroy(self):
        """
        Destroy the graphical representation
        """
        if self.node is not None:
            self.node.finalize(1)
        self.mesh = None
        self.created = False
        # print "Destroy (terrain)"


    def enterArea(self, area):
        """
        We've entered a new area
        """
        node = area.getNode()
        self.node.setParent(node)


    def leaveArea(self, aid):
        """
        Leave an area
        """
        self.destroy()


    observe_enterArea = enterArea
    observe_leaveArea = leaveArea


    def observe_setMeshGen(self, meshGen):
        """
        Set the mesh generator
        """
        self.meshGen = meshGen


    def observe_setMaterialAssetId(self, aid):
        """
        Set the asset for the material
        """
        d = self.parent.parent.getLocalService(IAssetClient).acquireAsset(aid)
        self.materialAssetId = aid
        def gotAsset(asset):
            self.materialAsset = asset
            self.object.setMaterial(asset)
        d.addCallback(gotAsset)
        return d


    def setPosition(self, pos):
        """
        Set the position of this object
        """
        self.node.setPosition(Vector(self.posoffset) + pos)


    def setRotation(self, rot):
        """
        Set the rotation of this object
        """
        self.node.setRotation(Quaternion(rot) * self.rotoffset)


    def setCopyableState(self, state):
        # print "set state", state
        Cacheable.setCopyableState(self, state)
        self.renderer = self.parent.parent.getLocalService(
            IPlayerClient).getRenderer()


class SceneryView(Scenery):
    """
    This version gets run on the client because it builds the
    visual aspects as well.
    """
    paged = None
    pageSize = 50
    batchPageLod = (100, 30)
    imposterPageLod = (4000, 50)
    vobs = tuple()

    def enterArea(self, area):
        """
        Enter an area
        """
        self.renderer = self.parent.parent.getLocalService(
            IPlayerClient).getRenderer()
        self.area = area
        self.node = area.getNode()
        self.space = area.getDisabledSpace()
        self.paged = OgrePagedGeometry(self.renderer, str(self))
        bounds = [None, None, None, None]
        assetList = []
        for p in self.pieces:
            for v in p.visualAssets:
                if not v["asset"] in assetList:
                    assetList.append(v["asset"])
                pos = v["position"]
                if bounds[0] is None or pos[0] < bounds[0]:
                    bounds[0] = pos[0] - 1
                if bounds[2] is None or pos[0] > bounds[2]:
                    bounds[2] = pos[0] + 1
                if bounds[1] is None or pos[2] < bounds[1]:
                    bounds[1] = pos[2] - 1
                if bounds[3] is None or pos[2] > bounds[3]:
                    bounds[3] = pos[2] + 1
        self.paged.create(bounds, self.pageSize, self.batchPageLod,
            self.imposterPageLod)
        d = self.parent.parent.getLocalService(
            IAssetClient).acquireAssets(assetList)
        def gotAssets(a):
            for asset in a:
                asset.create(asset.name)
            self.assets = dict(zip(assetList, a))
            return coiterate(PieceIterator(self.pieces,
                self.buildPiece))
        return d.addCallback(gotAssets)


    def buildPiece(self, piece):
        """
        Build a piece in our area
        """
        for c in piece.colliders:
            c.build(space=self.space)
        for v in piece.visualAssets:
            self.paged.addGeometry(self.assets[v["asset"]],
                v["position"], v.get("yRot", 0),
                v.get("scale", 1.0))



class GrassView(IVisualObjectView, Cacheable):
    """
    Client side visual representation of grass
    """
    grass = None
    bounds = None
    pageSize = 100
    maxDist = 100
    layers = None
    heightField = None


    def create(self, area, pos, rot):
        """
        Create the grass
        """
        self.renderer = self.parent.parent.getLocalService(
            IPlayerClient).getRenderer()
        asvc = self.parent.parent.getLocalService(IAssetClient)
        self.grass = OgreGrass(self.renderer)
        self.grass.create(self.bounds, self.getHeight,
            self.pageSize, self.maxDist)
        dd = []
        for l in self.layers:
            def gotAssets(data):
                asset, colorMapAsset, densityMapAsset = data
                kk = l.copy()
                del kk["aid"]
                if kk.has_key("colorMapAssetId"):
                    del kk["colorMapAssetId"]
                if kk.has_key("densityMapAssetId"):
                    del kk["densityMapAssetId"]
                kk["colorMapAsset"] = colorMapAsset
                kk["densityMapAsset"] = densityMapAsset
                return self.grass.addLayer(asset, **kk)
            a = [asvc.acquireAsset(l["aid"])]
            if l.get("colorMapAssetId") is not None:
                a.append(asvc.acquireAsset(l["colorMapAssetId"]))
            else:
                a.append(defer.succeed(None))
            if l.get("densityMapAssetId") is not None:
                a.append(asvc.acquireAsset(l["densityMapAssetId"]))
            else:
                a.append(defer.succeed(None))
            dd.append(defer.gatherResults(a).addCallback(gotAssets))
        return defer.gatherResults(dd)


    def enterArea(self, area):
        """
        We've entered a new area
        """


    def leaveArea(self, aid):
        """
        Leave an area
        """
        self.destroy()


    observe_enterArea = enterArea
    observe_leaveArea = leaveArea


    def destroy(self):
        """
        Destroy the grass. that's unpossible! :(
        """
        self.grass.destroy()
        self.grass = None


    def getHeight(self, pos):
        """
        Translate to local coords and get the height
        """
        p = ((pos[0] - self.bounds[0]) / (self.bounds[2] - self.bounds[0]),
            (pos[1] - self.bounds[1]) / (self.bounds[3] - self.bounds[1]))
        p = (p[0] * (self.heightField.resolution[0]-1),
            p[1] * (self.heightField.resolution[1]-1))
        return self.heightField.getSmoothedHeight(p, False)
