# Copyright (C) 2006-2008 Mortal Coil Games
# See LICENSE for details.
"""
"""

import logging

from time import time

from twisted.spread import pb
from twisted.internet import defer
from twisted.python.util import mergeFunctionMetadata

from mv3d.util.datastore.store import attributes, number, simpleList, text, autoStore
from mv3d.net.security import Securable

pb.MAX_BROKER_REFS = 65536
Priorities = {"Master":1, "Slave":2, "Cache":3}


class CacheableError(Exception):
    """
    Raised when there is a problem related to cacheable
    things
    """



def withClientUpdate(func):
    """
    Update all clients of this cacheable by
    remotely calling this function on them.
    """
    def wrapper(self, *args, **kw):
        r = func(self, *args, **kw)
        self.updateClients(func.__name__, *args, **kw)
        return r
    mergeFunctionMetadata(func, wrapper)
    return wrapper



def runOnMaster(func):
    """
    Make sure this function gets run on the master
    version of this object
    """
    def wrapper(self, *args, **kw):
        self.setupCacheable()
        if self.getPriority() != 1:
            if (hasattr(self.getManipulator(), "broker") and 
                self.getManipulator().broker.disconnected):
                raise CacheableError("Disconnected")
            return self.getManipulator().callRemote(
                "Manipulate", func.__name__, *args, **kw)
        r = func(self, *args, **kw)
        self.updateClients(func.__name__, *args, **kw)
        return r
    mergeFunctionMetadata(func, wrapper)
    return wrapper



def onlyOnSlave(func):
    """
    Make sure this function gets run on a slave
    and not on the master
    """
    def wrapper(self, *args, **kw):
        assert self.getPriority != 1, "This must only be called on a slave"
        return func(self, *args, **kw)
    mergeFunctionMetadata(func, wrapper)
    return wrapper



def onlyOnMaster(func):
    """
    Make sure this function gets run on the master
    and not on the slave
    """
    def wrapper(self, *args, **kw):
        assert self.getPriority == 1, "This must only be called on a master"
        return func(self, *args, **kw)
    mergeFunctionMetadata(func, wrapper)
    return wrapper



def observable(func):
    """
    This func runs only on the master but is also
    observable on clients. Returns two functions, the func and
    then the observer. Usage:

    def foo(bar):
        ...

    foo, observe_foo = runOnMaster(foo)

    """
    return runOnMaster(func), onlyOnSlave(withClientUpdate(func))



class Manipulator(Securable, pb.Viewable):
    """
    You need to add a docstring here ! 
    """
    def __init__(self, parent):
        self.parent = parent


    def view_getManipulator(self, con, nm):
        """
        You need to add a docstring here ! 
        """
        return self.parent.getManipulator(nm)(self.parent)


    def view_Manipulate(self, con, what, *a, **kw):
        """
        You need to add a docstring here ! 
        """
        if not what in dir(self.parent):
            return 0
        return getattr(self.parent, what)( * a, **kw)



class Cacheable(pb.Cacheable, pb.RemoteCache):
    """
    A convenience Cacheable class 
    """
    datastoreAttributes = attributes(copylevel=number, defaultmanipulator=text,
        exclude_from_state=simpleList, priority=number, lastuse=number,
        masterservers=simpleList, slaveservers=simpleList,
        manipulatorclasses=simpleList)

    server = None
    from_connection = None
    copylevel = 0
    manipulator = None
    defaultmanipulator = "Manipulator"
    exclude_from_state = None
    clients = None
    priority = Priorities["Master"]
    lastuse = None
    masterservers = None
    slaveservers = None
    replacefunctions = None
    manipulatorclasses = None


    def setupCacheable(self):
        if self.exclude_from_state is None:
            self.exclude_from_state = ["clients", "broker", 
                "from_connection", "server", "manipulator", "exclude_from_state",
                "replacefunctions"]
            self.lastuse = time()
        if self.masterservers is None:
            self.masterservers = []
            self.slaveservers = []
        if self.replacefunctions is None:
            self.replacefunctions = []
            self.priority = 1
        if self.manipulatorclasses is None:
            self.manipulatorclasses = {}
            if self.__class__.__name__ != "ClassGenerator":
                from mv3d.util.classgen import ClassGenerator
                self.addManipulatorClass("Manipulator", 
                    ClassGenerator(classname="Manipulator", 
                        modulename="mv3d.net.pb"))
                #class generators can't have default manip, so sorry
        if self.clients is None:
            self.clients = {}


    def __getstate__(self):
        self.setupCacheable()
        state=self.__dict__.copy()
        def killit(k):
            if state.has_key(k):
                del state[k]
        for xclude in self.exclude_from_state:
            # if not xclude in self.persist_ok:
            killit(xclude)
        return state


    def __setstate__(self, state):
        self.__init__()
        self.__dict__.update(state)


    def getManipulator(self, nm=None):
        """
        returns the default manipulator if nm is None
        otherwise, return a new manipulator
        """
        self.setupCacheable()
        if nm is None:
            if (self.manipulator is None and
                    self.__class__.__name__ != "ClassGenerator"):
                self.manipulator = (
                    self.manipulatorclasses[self.defaultmanipulator].construct(self))
            return self.manipulator
        if self.manipulatorclasses.has_key(nm):
            return self.manipulatorclasses[nm].construct(self)
        raise CacheableError("We don't have a %s manipulator in %s" % (
            nm, self.__class__.__name__))


    @autoStore
    def addManipulatorClass(self, nm, m):
        """
        Adds a manipulator class- nm is a string/name
        m is the manipulator class generator
        """
        self.setupCacheable()
        self.manipulatorclasses[nm] = m


    @autoStore
    def remManipulatorClass(self, nm):
        """
        Try to remove a manipulator class named nm (str)
        """
        self.setupCacheable()
        del self.manipulatorclasses[nm]


    def getDefaultManipulator(self):
        """
        Get the name of the default manipulator (str) 
        """
        return self.defaultmanipulator


    @autoStore
    def setDefaultManipulator(self, nm):
        """
        Sets the name of the default manipulator (str) 
        """
        self.defaultmanipulator = nm


    setDefaultManipulator, observe_setDefaultManipulator = observable(setDefaultManipulator)


    def getManipulatorClass(self, nm=None):
        """
        Returns manipulator class named nm-- if not specified, returns the
        default manipulator class. 
        """
        self.setupCacheable()
        if nm is None:
            nm = self.defaultmanipulator
        return self.manipulatorclasses[nm]


    def getCopyLevel(self):
        """
        Returns the amount of caches this object
        is in between here and the master server
        """
        return self.copylevel


    def getParentConnection(self):
        """
        returns the connection instance that 
        we came in over.
        """
        return self.from_connection


    def getServer(self):
        """
        Returns the server we belong to
        """
        return self.server


    def setServer(self, s):
        """
        Set the server that we belong to
        """
        self.server = s


    @autoStore
    def setPriority(self, p):
        """
        Sets whether this instance of the 
        cache is a master, slave, or just a plain old cache
        p can be either an int or a str
        """
        if isinstance(p, str):
            p = Priorities[p]
        self.priority = p


    def getPriority(self):
        """
        Return the priority (int) that we are set to 
        """
        return self.priority


    def addExclude(self, *elist):
        """
        Add an item to be excluded from transferring over the network 
        """
        self.setupCacheable()
        if isinstance(elist[0], list) or isinstance(elist[0], tuple):
            for e in elist[0]:
                if e not in self.exclude_from_state:
                    self.exclude_from_state.append(e)
            return
        for e in elist:
            if e not in self.exclude_from_state:
                self.exclude_from_state.append(e)


    def remExclude(self, e):
        """
        Remove a previously added item so that it will now be 
        transferred over the network.
        """
        self.setupCacheable()
        self.exclude_from_state.remove(e)


    @autoStore
    def addMasterServer(self, s):
        """
        Add a master server to this cache
        s should be a ServiceLoc instance
        Really, there should be only one master server
        """
        self.setupCacheable()
        self.masterservers.append(s)


    addMasterServer, observe_addMasterServer = observable(addMasterServer)


    @autoStore
    def remMasterServer(self, s):
        """
        Remove a master server from the list.
        s is a ServiceLoc instance
        returns 1 if s in masterservers
        """
        self.masterservers.remove(s)


    remMasterServer, observe_remMasterServer = observable(remMasterServer)


    def getMasterServers(self):
        """
        returns the master server list 
        """
        return self.masterservers


    @autoStore
    def addSlaveServer(self, s):
        """
        Add s as a slave server of this cache.
        s is a ServiceLoc instance
        """
        self.slaveservers.append(s)

    addSlaveServer, observe_addSlaveServer = observable(addSlaveServer)


    @autoStore
    def remSlaveServer(self, s):
        """
        Remove s from the list of slave servers.
        s is a ServiceLoc instance
        returns 1 if s in slaveservers
        """
        self.slaveservers.remove(s)

    remSlaveServer, observe_remSlaveServer = observable(remSlaveServer)


    def getSlaveServers(self):
        """
        Returns the list of slave servers 
        """
        return self.slaveservers


    def updateClients(self, rc, *a, **kw):
        """
        Call this to update the clients that are caching us.
        rc, a, and kw are basically the same args you'd pass to
        callRemote, so rc=a string of the remote procedure to run
        This function also updates the lastuse variable. That is used
        to determine if this cache is needed.
        """
        self.setupCacheable()
        dd = []
        badc = []
        for k,c in self.clients.items():
            if self.server is not None:
                self.server.cupdates += 1
            try:
                dd.append(c.callRemote(rc, *a, **kw))
            except AttributeError:
                badc.append(k)
            except pb.DeadReferenceError:
                badc.append(k)
        for k in badc:
            del self.clients[k]
        return defer.DeferredList(dd)


    def hasClients(self):
        """
        Return 1 if there are attached clients 
        """
        self.setupCacheable()
        return len(self.clients) != 0


    def getLastUse(self):
        """
        Return the last usage time 
        """
        return self.lastuse


    def timeSinceLastUse(self):
        """
        Return how much time has passed since the last use 
        """
        return time() - self.lastuse


    def canPrune(self):
        """
        This function will return 1 if it believes this cache
        is not needed any more. It will only return 1 for Cache priority
        objects that have no clients.
        ToDo: It should also check the last use time.
        """
        if self.priority != Priorities["Cache"]:
            return False
        if self.hasClients():
            return False
        return True


    def countClients(self):
        """
        Return the number of connected clients 
        """
        return len(self.clients)


    def getClients(self):
        """
        return the list of clients (which are RemoteReferences 
        """
        return self.clients


    def addClient(self, persp, con):
        """
        Add a client 
        persp=perspective
        con=observer (in twisted speak)
        """
        self.setupCacheable()
        self.clients[persp] = con


    def remClient(self, persp):
        """
        Remote a client
        persp=perspective
        """
        self.setupCacheable()
        try:
            del self.clients[persp]
        except KeyError:
            pass


    def hasLostMaster(self):
        """
        You need to add a docstring here ! 
        """
        # if we are the master, then 
        # everything is all good
        if self.priority == 1:
            return False
        if not hasattr(self.getManipulator(), "broker"):
            return False
        return self.broker.disconnected


    def getState(self):
        """
        Taking into account the exclude list, this function
        grabs the __dict___ and culls excluded items before returning it.
        """
        #print "GetState", self
        self.setupCacheable()
        state = self.__dict__.copy()
        for r in self.exclude_from_state:
            if state.has_key(r):
                del state[r]
        # increase the copy level, set the top server if needed
        # and set the priority down to cache.
        # note that the other end will need to manyally
        # select Slave as a priority level if it is desired.
        state["copylevel"] = self.copylevel + 1
        # if self.getPriority() ==1 and self.server is not None:
            # if self.masterservers ==[]:
                # state["masterservers"] = [ServiceLoc().FromServer(self.server)]
        state["priority"] = Priorities["Cache"]
        from mv3d.util.classgen import ClassGenerator
        if not isinstance(self, ClassGenerator):
            state["manipulator"] = self.manipulatorclasses[self.defaultmanipulator].construct(self)
        else:
            state["manipulator"] = 0
        #self.getManipulator()
        #print "Get state for", self, "is", state
        return state


    def setState(self, state):
        """
        This simply updates __dict___ with state 
        """
        self.__dict__.update(state)


    def getStateToCacheAndObserveFor(self, perspective, observer):
        """
        Called when this object is sent over the network.
        First update the last use.
        Then Check to make sure the remote user has permissions
        to reference us. If not, return {}
        Else, add her as a client and return the state after
        adding a item that will give the other end a reference
        to the receiving connection
        """
        assert perspective is not None, (
            "Perspective of %s is None" % self.__class__.__name__)
        self.lastuse = time()
        if not self.checkPermissions(perspective, "reference"):
            if perspective != None:
                if hasattr(self, "perspective"):
                    pp = self.perspective
                else:
                    pp = "Unknown"
                perspective.parent.parent.log("Access denied to"
                    "%s for reference of %s" % (pp, self.__class__.__name__),
                    logging.ERROR)
            raise CacheableError("Permission denied sending %s!" % 
                self.__class__.__name__)
        self.addClient(perspective, observer)
        st = self.getState()
        if hasattr(perspective, "controller"):
            st["cacheable_con"] = perspective.controller
        return st


    def setCopyableState(self, state):
        """
        Called when this item is received over the netwokr.
        First, update the last use.
        Then Get the special cacheable_con item and grab the Connection
        we came in over and set that as from_connection. Then from that
        connection, get its parent and set that as server.
        Continue on to SetState
        """
        self.lastuse = time()
        if state.has_key("cacheable_con"):
            self.from_connection = state["cacheable_con"].object
            self.parent = self.from_connection.parent
        else:
            if hasattr(self.broker, "mv3d_server"):
                self.server = self.broker.mv3d_server
        g = self.setState(state)
        return g


    def stoppedObserving(self, perspective, observer):
        """
        Called when a client is disconnecting from us for whatever reason. 
        """
        self.lastuse = time()
        self.remClient(perspective)



class Copyable(pb.Copyable, pb.RemoteCopy):
    """
    Defines an object that can be copied across
    the network via twisted.spread
    """
    datastoreAttributes = attributes(exclude_from_state=simpleList)

    exclude_from_state = None

    def setupCopyable(self):
        """
        Set this instance up if needed
        """
        if self.exclude_from_state is None:
            self.exclude_from_state = ["broker"]


    def addExclude(self, *elist):
        """
        Add an item to be excluded from transferring over the network 
        """
        self.setupCopyable()
        if isinstance(elist[0], list) or isinstance(elist[0], tuple):
            for e in elist[0]:
                if e not in self.exclude_from_state:
                    self.exclude_from_state.append(e)
            return
        for e in elist:
            if e not in self.exclude_from_state:
                self.exclude_from_state.append(e)


    def remExclude(self, e):
        """
        You need to add a docstring here ! 
        """
        self.setupCopyable()
        self.exclude_from_state.remove(e)


    def getStateToCopy(self):
        """
        Get the state of this object excluding
        previously mentioned exclude items
        """
        self.setupCopyable()
        state = self.__dict__.copy()
        for r in self.exclude_from_state:
            if state.has_key(r):
                del state[r]
        return state


    def setCopyableState(self, state):
        """
        Set the state of this object to what was
        sent over the wire
        """
        self.__init__()
        self.__dict__.update(state)



