# Copyright (C) 2006-2008 Mortal Coil Games
# See LICENSE for details.
""" -*- test-case-name: mv3d.test.test_securable -*-
Security Classes 
"""

import logging

from twisted.python.util import mergeFunctionMetadata

from mv3d.util.datastore.store import (
    attributes, simpleDict, number, autoStore
    )


Order = {"deny first":0, "allow first":1}


class AccessDenied(Exception):
    """
    Access was denied
    """



class SecurableError(Exception):
    """
    Raised when you tried to do something
    the Securable didn't like
    """



def requirePermissions(*perms):
    """
    Specify that certain permissions are required
    for this method assumes argument 2
    can be checked for security.
    """
    def decorator(func):
        def wrapper(self, *args, **kw):
            for perm in perms:
                if not self.checkPermissions(args[0], perm):
                    raise AccessDenied("You do not have %s permissions "
                        "to run %s." % (perm, func.__name__))
            return func(self, *args, **kw)
        mergeFunctionMetadata(func, wrapper)
        return wrapper
    return decorator



class Securable:
    """
    Securable is a class that defines subclasses
    as things that can be locked down with extendable
    security settings
    """
    datastoreAttributes = attributes(allowed=simpleDict, denied=simpleDict,
        availablepermissions=simpleDict, order=number)


    allowed = None
    denied = None
    availablepermissions = None
    order = Order["deny first"]


    def setupSecurable(self):
        if self.allowed is None:
            self.allowed = {}
            self.denied = {}
        if self.availablepermissions is None:
            self.availablepermissions = ["all", "reference", "read", 
                "modify", "delete", "read permissions"]


    def _securableupdateClients(self, w, *a, **kw):
        """
        Call updateClients if we have that method 
        """
        if hasattr(self, "updateClients"):
            self.updateClients(w,  * a, **kw)


    @autoStore
    def setOrder(self, o):
        """
        Set whether we deny or allow first 
        """
        self.order = o
        self._securableupdateClients('setOrder', o)


    @autoStore
    def addPermission(self, name):
        """
        Add a permission type
        """
        self.setupSecurable()
        assert name not in self.availablepermissions
        self.availablepermissions.append(name)
        self._securableupdateClients('addPermission', name)


    @autoStore
    def remPermission(self, name):
        """
        Remote a permission
        """
        self.setupSecurable()
        if not name in self.availablepermissions:
            raise SecurableError("There is no permission named %s" % name)
        self.availablepermissions.remove(name)
        if self.allowed.has_key(name):
            del self.allowed[name]
        self._securableupdateClients('remPermission', name)


    @autoStore
    def grantPermission(self, permission, who):
        """
        Grant permission to this object
        """
        self.setupSecurable()
        if not permission in self.availablepermissions:
            raise SecurableError("There is no permission named %s" % 
                permission)
        if permission in self.allowed.keys():
            if not who in self.allowed[permission]:
                self.allowed[permission].append(who)
        else:
            self.allowed[permission] = [who]
        self._securableupdateClients('grantPermission', permission, who)


    def observe_grantPermission(self, permission, who):
        return self.grantPermission(permission, who)


    @autoStore
    def denyPermission(self, permission, who):
        """
        Deny access to this object
        """
        self.setupSecurable()
        if not permission in self.availablepermissions:
            raise SecurableError("There is no permission named %s" % 
                permission)
        if permission in self.denied.keys():
            if not who in self.denied[permission]:
                self.denied[permission].append(who)
        else:
            self.denied[permission] = [who]
        self._securableupdateClients('denyPermission', permission, who)


    def observe_denyPermission(self, permission, who):
        return self.denyPermission(permission, who)


    @autoStore
    def revokePermission(self, permission, who):
        """
        Remove a previously granted permission
        """
        self.setupSecurable()
        if not permission in self.availablepermissions:
            raise SecurableError("There is no permission named %s" % 
                permission)
        self.allowed[permission].remove(who)
        self._securableupdateClients('revokePermission', permission, who)


    @autoStore
    def remDeniedPermission(self, permission, who):
        """
        Remove a previously denied permission
        """
        self.setupSecurable()
        if not permission in self.availablepermissions:
            raise SecurableError("There is no permission named %s" % 
                permission)
        self.denied[permission].remove(who)
        self._securableupdateClients('remDeniedPermission', permission, 
            who)


    def checkDenyPermissions(self, con, act):
        """
        Check if con's user is in our denied list for the
        given ACTion
        """
        self.setupSecurable()
        if "all" in self.denied.keys():
            if con.getName() in self.denied["all"]:
                return True
            for g in con.getGroups():
                if g in self.denied["all"]:
                    return True
        if not act in self.denied.keys():
            return False
        pp = self.denied[act]
        if "all" in pp:
            return True
        if con.getName() in pp:
            return True
        gr = con.getGroups()
        for g in gr:
            if g in pp:
                return True
        return False


    def checkAllowPermissions(self, con, act):
        """
        Check if con's user is in our allowed list for the given
        ACTion
        """
        self.setupSecurable()
        if "all" in self.allowed.keys():
            if con.getName() in self.allowed["all"]:
                return True
            for g in con.getGroups():
                if g in self.allowed["all"]:
                    return True
        if not act in self.allowed.keys():
            return False
        pp = self.allowed[act]
        if "all" in pp:
            return True
        if con.getName() in pp:
            return True
        gr = con.getGroups()
        for g in gr:
            if g in pp:
                return True
        return False


    def checkPermissions(self, con, act):
        """
        Returns true if con is allowed to perform act on
        this object
        """
        self.setupSecurable()
        if hasattr(self, "server"):
            svr = self.server
        elif hasattr(self, "log"):
            svr = self
        elif hasattr(self, "client"):
            svr = self.client
        elif hasattr(self, "parent"):
            svr = self.parent
        else:
            svr = None
        if not act in self.availablepermissions:
            raise SecurableError("There is no permission named %s" % act)
        if self.order == Order["deny first"]:
            if self.checkDenyPermissions(con, act):
                if svr is not None:
                    svr.log("%s access denied for %s to %s" % (act,
                        con, self.__class__.__name__), logging.WARNING)
                return False
            return self.checkAllowPermissions(con, act)
        if self.order == Order["allow first"]:
            if self.checkAllowPermissions(con, act):
                return True
            return not self.checkDenyPermissions(con, act)
        if svr is not None:
            svr.log("%s access denied for %s to %s" % (act,
                con, self.__class__.__name__), logging.WARNING)
        return False
