# -*- test-case-name: mv3d.test.test_biped -*-
# Copyright (C) 2007-2008 Mortal Coil Games
# See LICENSE for details.
"""

Biped related functionality

"""

from time import time

import ode

from mv3d.util.datastore.store import (attributes, 
    simpleList, referenceList, classHash, number, boolean,
    autoStore, randomHash)
from mv3d.util.datastore.axiomstore import (
    DirectStore, AxiomVector, AxiomQuaternion
    )
from mv3d.phys.collider import odeCollider, castRay
from mv3d.phys.body import IBody, IVisualObjectView, StringFlinger
from mv3d.util.math3d import Vector, matrixFromEuler, Quaternion
from mv3d.net.pb import Cacheable, withClientUpdate
from mv3d.net.security import Securable
from mv3d.util.classgen import ClassGenerator


class BipedBody(IBody, Cacheable, Securable):
    """
    A biped body is a combination of two
    bodies. One has a physical form (the body)
    and the other doens't (the head)
    """
    datastoreAttributes = attributes(IBody, Cacheable, Securable,
        height=number, climbheight=number, radius=number, 
        bodyoffset=number, headoffset=number, bodylength=number,
        maxwalkspeed=number, maxturnspeed=number, jumpcounter=number,
        gradualOffset=number, enabled=boolean, walking=number, 
        turning=number, walkdir=simpleList, turndir=simpleList,
        oid=number, areaid=number, vobs=referenceList,
        initial_position=DirectStore(AxiomVector, True,
            u"getPosition"),
        initial_rotation=DirectStore(AxiomQuaternion, True,
            u"getRotation"),
        initial_linearVel=DirectStore(AxiomVector, True,
            u"getLinearVelocity"),
        initial_angularVel=DirectStore(AxiomVector, True,
            u"getAngularVelocity"),
        objectid=number,
        )
    datastoreIndexes = tuple()
    datastoreHash = ("oid", "areaid", classHash, randomHash)
    datastoreVersion = 1

    setCopyableState = Cacheable.setCopyableState
    height = 7
    climbheight = 1
    radius = 2.5
    bodyoffset = 0
    headoffset = 0
    bodylength = 0
    maxwalkspeed = 25
    maxturnspeed = 2.0
    jumpcounter = 0
    gradualOffset = 0
    lastIter = None
    lastpos = None
    enabled = True
    onground = False
    standing = False
    walking = False
    turning = False
    stopturning = False
    walkdir = (0,0,0)
    turndir = (0,0,0)
    body = None
    head = None
    headjoint = None
    world = None
    oid = None
    bodycollider = None
    colliderspace = None
    space = None
    area = None
    areaid = None
    parent = None
    initial_position = None
    initial_rotation = None
    initial_linearVel = None
    initial_angularVel = None


    def __init__(self):
        self.otherspacecolliders = {}
        self.otherspacecolliderspaces = {}
        self.vobs = []
        self.aggregators = []
        self.addExclude("lastpos", "onground", "body", "head", "headjoint",
            "world", "bodyoffset", "headoffset", "otherspacecolliders",
            "otherspacecolliderspaces", "space", "area", "colliderspace",
            "aggregators", "parent")


    @autoStore
    @withClientUpdate
    def setHeight(self, h):
        """
        Set the height
        """
        self.height = h


    @autoStore
    @withClientUpdate
    def setClimbHeight(self, h):
        """
        Set the climbing height
        """
        self.climbheight = h


    @autoStore
    @withClientUpdate
    def setRadius(self, r):
        """
        Set the radius of the body
        """
        self.radius = r


    observe_setHeight = setHeight
    observe_setClimbHeight = setClimbHeight
    observe_setRadius = setRadius


    def create(self, world):
        """
        Create this BipedBody 
        """
        assert self.body is None
        # print "CREATE CALLED!"#, tuple(self.initial_position)
        self.bodylength = self.height - self.climbheight - (2 * self.radius)
        self.bodyoffset = self.height - (self.bodylength / 2)
        self.headoffset = self.height - self.radius
        self.body = ode.Body(world)
        self.head = ode.Body(world)
        # print "my body", self.body
        self.headjoint = ode.BallJoint(world)
        self.body.setPosition((0, self.bodyoffset, 0))
        self.head.setPosition((0, self.headoffset, 0))
        self.headjoint.attach(self.body, self.head)
        self.headjoint.setAnchor((0, self.headoffset, 0))
        self.bodycollider = odeCollider("Cylinder", 
            length=self.bodylength, 
            radius=self.radius,
            rotation=matrixFromEuler((3.1415 / 2, 0, 0)))
        if self.oid is not None:
            self.bodycollider.oid = self.oid
        # print "pos",self.initial_position
        if self.initial_position is not None:
            self.body.setPosition(self.initial_position)
            self.head.setPosition(Vector(0, 
                self.headoffset - self.bodyoffset, 0) + self.initial_position)
            # self.body.setPosition((0,0,0))
        if self.initial_rotation is not None:
            self.head.setQuaternion(self.initial_rotation)
        if self.initial_linearVel is not None:
            self.body.setLinearVel(self.initial_linearVel)
        if self.initial_angularVel is not None:
            self.head.setAngularVel(self.initial_angularVel)
        if not self.enabled:
            self.body.disable()
            self.head.disable()


    def destroy(self):
        """
        Destroy this body
        """
        # print "Destroy Biped"
        self.bodycollider.destroy()
        self.headjoint.attach(None, None)
        self.headjoint = None
        self.head = None
        self.body = None
        for v in self.vobs:
            v.destroy()
        # print "Done"


    def loadedFromStore(self):
        """
        Called when loaded from a datastore
        """
        self.aggregators = []


    def prepareForStore(self, store):
        """
        Called before storing this object into a store
        """
        # self.initial_position = self.getPosition()
        # self.initial_rotation = self.getRotation()
        # self.initial_linearVel = self.getLinearVelocity()
        # self.initial_angularVel = self.getAngularVelocity()


    def standUp(self):
        """
        Make the body float above the ground
        to simulate feet
        """
        dp = 5
        if self.lastpos is not None:
            dp = self.body.getPosition()[1] - self.lastpos[1]
        self.lastpos = self.body.getPosition()
        p = self.body.getPosition()
        lv = self.body.getLinearVel()
        self.onground = False
        bodo = self.bodyoffset#(self.height / 2.0) + self.radius# +0.2 HERE
        f = 0
        extracheck = 0.5
        # print "meee",self.bodycollider.geom.waterdepth, self.bodycollider.tgeom
        # if self.bodycollider.tgeom.waterDepth > 0:
            # print "WATER"
            # extracheck = -0.5
        def rayCheck(offs=(0, 0, 0), checkspaces=[]):
            """
            Shoot a ray for feet
            """
            checkspaces.append(self.area.getDisabledSpace())
            checkspaces.append(self.area.getEnabledSpace())
            d = None
            geom = None
            pos = None
            mindepth = 9999999
            p1 = Vector(p) + Vector(offs) #+ Vector(0,2,0)
            p2 = (p1[0], p1[1] - (bodo + extracheck), p1[2])
            # print bodo, offs[1]
            for space in checkspaces:
                r = castRay(space, p1, p2, [self.bodycollider])
                if r.mindepth < mindepth:
                    mindepth = r.mindepth
                    d = bodo - r.mindepth
                    geom = r.mingeom
                    pos = r.minpos
            if d is not None and self.jumpcounter == 0:
                # print geom, d
                # if d - extracheck > 0.1 or lv[1] > 500 * self.body.getMass().mass:
                    # f = (((d - extracheck) - lv[1]) * self.body.getMass().mass) / 0.004#* 250 #
                    # self.body.addForce((0, f, 0))
                    # print "i never get here", f
                # else:
                # uncomment these two lines to have no knees
                if d - extracheck > 0 or self.bodycollider.tgeom.waterDepth == 0:
                    self.body.setPosition((p[0], p[1] + (d - extracheck), p[2]))
                    self.body.setLinearVel((lv[0], 0, lv[2]))
                    bod = geom.getBody()
                    if bod is None and hasattr(geom, "body"):
                        bod = geom.body
                    if bod is not None:
                        # apply force to it
                        # f = (((d - extracheck) - lv[1]) * self.body.getMass().mass) * 0.5
                        f = self.body.getMass().mass * (lv[1] + 1) #* 0.05
                        # print "Pushing", f
                        bod.addForceAtPos((0,  f, 0), pos)
                        bod.enable()
                        # should also check for rotation
                    self.onground = True
            # else:
                # print "ground?", p1[1], p2[1]
            return d, geom, pos
        for offs in ((0,0,0), (0.3,0,0.3), (0.1,0,0.1)):
            d, geom, pos = rayCheck(offs)
            if d is not None:
                break
        else:
            self.jumpcounter = 0
        # print "On ground", self.bodycollider.tgeom.waterDepth, tuple(self.body.getPosition())
        # if not self.onground:
            # print self.onground
        # print self.body.getNumJoints()
        self.body.setRotation((1,0,0,0,1,0,0,0,1))
        self.body.setAngularVel((0,0,0))
        # lvel=self.body.getLinearVel()
        # avel=self.head.getAngularVel()
        # slv=abs(lvel[0])+abs(lvel[2])+abs(dp)
        # sav=abs(avel[0])+abs(avel[1])+abs(avel[2])
        # if (slv<0.01 and sav==0 and self.onground 
            # and not self.walking and not self.turning and dp<=0):
            # if self.enabled:
                # self.disable()
                # print "IS THIS"
        return f


    def isOnGround(self):
        """
        Returns true if the biped is standing on the groun. 
        """
        return self.onground


    # @withClientUpdate
    def standStill(self):
        """
        Attempt to retard the lateral movement of the biped
        as if it was trying with its feet to stand still.
        Note: This function starts this behavior and will stop
        all walking behavior.
        """
        # if self.getPriority() != 1:
            # m = self.getManipulator()
            # m.callRemote("Manipulate", "standStill")
            # return
        # print "Standing!", self.getPosition()
        self.standing = True
        self.walking = False
        self.walkdir = (0, 0, 0)
        for v in self.vobs:
            v.startAnimation("idle", True)
            v.stopAnimation("walk")


    @withClientUpdate
    def walk(self, d):
        """
        Start walking in the direction specified
        by d. This function only starts the behavior.
        """
        # if self.getPriority() != 1:
            # m = self.getManipulator()
            # if m.broker.disconnected:
                # return
            # m.callRemote("Manipulate", "walk", d)
            # return
        for v in self.vobs:
            v.startAnimation("walk", True)
            v.stopAnimation("idle")
        self.enable()
        self.walking = True
        self.standing = False
        # print "old=", list(self.walkdir)
        wd = Vector(self.walkdir) + d
        wd0, wd1, wd2 = wd
        if wd0 == 0:
            wd0 = 1
        if wd1 == 0:
            wd1 = 1
        if wd2 == 0:
            wd2 = 1
        self.walkdir = (wd[0] / abs(wd0), wd[1] / abs(wd1), wd[2] / abs(wd2))
        # print "Walk", list(wd), "d=",d, self.walkdir, self.getPosition()
        if self.walkdir == (0, 0, 0):
            self.standStill()
            return


    @withClientUpdate
    def turn(self, d):
        """
        Start turning in the direction specified by
        d. This starts the behavior.
        """
        # print "Turn"
        # if self.getPriority() != 1:
            # m = self.getManipulator()
            # if m.broker.disconnected:
                # return
            # m.callRemote("Manipulate", "turn", d)
            # return
        self.enable()
        nd = []
        for x in range(len(d)):
            if d[x]!=0:
                nd.append(d[x]/abs(d[x]))
            else:
                nd.append(d[x])
        d = tuple(nd)
        otd = self.turndir
        self.turndir = Vector(d) * 50
        # print "Turning", d, list(otd), list(self.turndir)
        self.turning = True
        self.stopturning = False
        self.lastturn = time()


    @withClientUpdate
    def jump(self):
        """
        Apply a force sufficient to make the biped jump
        into the air.
        """
        # if self.getPriority() != 1:
            # m = self.getManipulator()
            # m.callRemote("Manipulate", "jump")
            # return
        # if not self.isOnGround():
          # return
        if self.enabled or not self.body.isEnabled():
            self.enable()
        self.body.addForce((0, 200 * self.body.getMass().mass, 0))
        self.jumpcounter = 1


    @withClientUpdate
    def stopTurning(self):
        """
        Start slowing down angular momentum. 
        """
        # print "stopTurning", list(self.turndir)
        # if self.getPriority() != 1:
            # m = self.getManipulator()
            # m.callRemote("Manipulate", "stopTurning")
            # return
        self.stopturning = 1
        self.turning = False
        self.turndir = (0,0,0)


    # observe_standStill = standStill
    # observe_walk = walk
    observe_turn = turn
    observe_jump = jump
    observe_stopTurning = stopTurning
    
    def observe_walk(self, d):
        # print "observe walk", self.getPosition()
        self.walk(d)


    def observe_standStill(self):
        # print "observe_standStill"
        self.standStill()


    def doWalk(self, dt):
        """
        Called each iteration, this function
        applies the various forces associated with
        walking.
        """
        if not self.isOnGround():
            # print "Not on ground", self.getPosition()
            return
        lv = self.body.getLinearVel()
        spd = Vector(lv[0], 0, lv[2]).length()
        d = Quaternion(self.head.getQuaternion())
        # print "Walking", spd
        if spd >= self.maxwalkspeed:
            spd = self.maxwalkspeed
        else:
            rv = d.rotate(self.walkdir)
            f = Vector(rv) * (self.body.getMass().mass * 1000 * dt)
            self.body.addForce(f)
            return
        nlv = d.rotate(self.walkdir) * spd
        nlv = (nlv[0], lv[1], nlv[2])
        self.body.setLinearVel(nlv)


    def doStand(self, p=0.3):
        """
        Called every iteration, this function
        attempts to stop the biped. When it it stopped,
        turn off the standing behavior.
        """
        if not self.onground:
            # print "not stnading on grond"
            return
        # print "Standing"
        lv = self.body.getLinearVel()
        vv = lv[0] + lv[2]
        if vv < 0.1 and self.onground:
            self.body.setLinearVel((0, 0, 0))
            if not self.turning and not self.stopturning:
                self.disable()
                # print "disabling", self.bodycollider.tgeom, self.bodycollider.tgeom.getSpace()
            self.standing = False
            return
        v = Vector(lv) * (p * 0.05)
        v = (v[0], lv[1], v[2]) # can't stop gravity!
        self.body.setLinearVel(v)


    def doTurn(self, dt):
        """
        Updated every Iteration, this function applies
        the various forces needed to continue turning.
        """
        td = Vector(self.turndir) * (5.0 * dt)
        self.head.addTorque(td)
        lv = self.head.getAngularVel()
        spd = (lv[0] * lv[0]) + (lv[1] * lv[1]) + (lv[2] * lv[2])
        # print "turning", list(td), spd, dt
        if spd > self.maxturnspeed:
            self.head.setAngularVel(Vector(lv) * (self.maxturnspeed / spd))
            nspd = Vector(self.head.getAngularVel()).length()
            # print "nope", spd, self.maxturnspeed#, self.body.getMass().mass
            # print self.bodylength, self.radius 
            return


    def doStopTurning(self):
        """
        Called every iteration, this function will decrease
        angular momentum until it is nearly none. Then it stops
        the behavior.
        """
        av = Vector(self.head.getAngularVel()) * 0.05#(20.0 * 0.05)
        self.head.setAngularVel(av)
        if self.stopturning > 2:
            # print "stop", self.getAngularVelocity()
            self.head.setAngularVel((0, 0, 0))
            # print "post", self.head.getAngularVel()
            self.stopturning -= 1
            if self.stopturning == 2:
                self.stopturning = False
            return
        spd = Vector(self.head.getAngularVel()).squaredDistance((0,0,0))
        # print "stopturning", spd
        # print spd
        if spd < 0.01:
            self.stopturning = 15
            self.head.setAngularVel((0, 0, 0))


    def enterSpace(self, space):
        """
        Create colliders in this space
        """
        # print "Enter Space", space, self.space
        assert self.body is not None
        # if self.colliderspace is not None:
            # self.colliderspace.destroy()
        # self.colliderspace = odeCollider("Space")
        # self.colliderspace.Build(body=None, space=space)
        self.bodycollider.destroy()
        self.bodycollider.build(body=self.body, space=space)
        self.bodycollider.geom.waterDepth = 0
        self.bodycollider.tgeom.waterDepth = 0
        self.space = space

    
    def enterSecondarySpace(self, ownerid, space):
        """
        Create secondary colliders in this space
        """
        # print "WTF"
        self.otherspacecolliders[ownerid] = []
        self.otherspacespaces[ownerid] = space
        if self.otherspacecolliderspaces.has_key(ownerid):
            self.otherspacecolliderspaces.destroy()
        # self.otherspacecolliderspaces[ownerid] = odeCollider("Space")
        # self.otherspacecolliderspaces[ownerid].Build(space=space)
        self.otherspacecolliders[ownerid] = self.bodycollider.copy()
        self.otherspacecolliders[ownerid].setParameter("density", 0)
        self.otherspacecolliders[ownerid].Build(self.body, space)


    def enterArea(self, area):
        """
        Determine the space most appropriate
        in are and enter it. Area should be an
        area object
        """
        if self.body is None:
            self.create(area.world) #Hack!
        if self.isEnabled():
            space = area.getEnabledSpace(self.getPosition())
        else:
            space = area.getDisabledSpace(self.getPosition())
        self.enterSpace(space)
        self.areaid = area.getID()
        self.area = area
        for v in self.vobs:
            v.enterArea(area)


    def leaveArea(self, area):
        """
        This body is leaving the area, remove
        everything in the area
        """
        # self.colliderspace.destroy()
        self.bodycollider.destroy()
        self.space = None


    def leaveSecondarySpace(self, ownerid):
        """
        Remove our secondary info from this space
        """
        self.otherspacecolliders[ownerid].destroy()
        del self.otherspacecolliders[ownerid]
        del self.otherspacespaces[ownerid]
        # self.otherspacecolliderspaces[ownerid].destroy()
        # del self.otherspacecolliderspaces[ownerid]


    def hasSecondarySpace(self, ownerid):
        """
        Returns true if we have colliders in this secondary space
        """
        return ownerid in self.otherspacespaces.keys()


    def attach(self, other, joint):
        """
        Attach this body to another one using the joint
        """
        joint.attach(self, other)
        self.joints.append(joint)
        other.joints.append(joint)


    @withClientUpdate
    def setPosition(self, pos):
        """
        Set the position of this body
        """
        self.body.setPosition(pos)
        self.updateVobs()


    def getPosition(self):
        """
        Get the position of this body
        """
        return self.body.getPosition()


    @withClientUpdate
    def setRotation(self, rot):
        """
        Set the quaternion rotation of this body
        """
        self.head.setQuaternion(rot)
        self.updateVobs()


    def getRotation(self):
        """
        Get the quaternion rotation of this body
        """
        return self.head.getQuaternion()


    @withClientUpdate
    def setLinearVelocity(self, lv):
        """
        Set the linear velocity of this body
        """
        self.body.setLinearVel(lv)
        
        
    def getLinearVelocity(self):
        """
        Returns the linear velocity of this body
        """
        return self.body.getLinearVel()


    @withClientUpdate
    def setAngularVelocity(self, av):
        """
        Set the angular Velocity of this body
        """
        self.head.setAngularVel(av)
        
        
    def getAngularVelocity(self):
        """
        Get the angular velocity of this body
        """
        return self.head.getAngularVel()


    @withClientUpdate
    def enable(self):
        """
        Set the state of this body to be enabled
        """
        # print "Enable biped", self.getPosition()
        self.body.enable()
        self.head.enable()
        self.setLinearVelocity(self.getLinearVelocity())
        self.setAngularVelocity(self.getAngularVelocity())
        # self.setPosition(self.getPosition())
        # self.setRotation(self.getRotation())
        if not self.enabled:
            # print self.server === none poop
            # print self.oid, hasattr(self, "objectid")
            if self.objectid is not None:
                self.parent.enableItem(self.objectid)
            if self.area is not None:
                self.enterSpace(self.area.getEnabledSpace(self.getPosition()))
        self.enabled = True


    @withClientUpdate
    def disable(self):
        """
        Set the state of this body to be disabled
        """
        # import traceback
        # print traceback.print_stack()
        # print "Disable body", self.getPosition()
        self.body.disable()
        self.head.disable()
        self.setLinearVelocity((0,0,0))
        self.setAngularVelocity((0,0,0))
        self.setPosition(self.getPosition())
        self.setRotation(self.getRotation())
        if self.enabled and self.area is not None:
            self.enterSpace(self.area.getDisabledSpace(self.getPosition()))
        self.enabled = False


    def isEnabled(self):
        """
        Returns true if this body is enabled
        """
        # if self.body is None:
            # return self.enabled
        return self.body.isEnabled()


    def setWaterDepth(self, depth):
        """
        Set how far under water this is
        """
        self.waterdepth = depth


    def updateDynamics(self, pos, rot, lv, av):
        """
        Update the position, rotation, linear and angular velocities
        """
        # assert self.enabled, "Not enabled"
        # dp = Vector(self.getPosition()) - pos
        # ln = dp.squaredDistance((0,0,0))
        # if ln > 0 and ln < 10:
            # self.gradualUpdate = dp * (1 / (ln * 5))
            # self.gradualOffset = ln * 5
        # else:
        # print "got update", list(av), list(lv)
        if not self.isEnabled():
            l = (Vector(av).squaredDistance((0,0,0)) + 
                Vector(lv).squaredDistance((0,0,0)))
            if l > 0.1:
                self.enable()
        self.body.setPosition(pos)
        self.head.setQuaternion(rot)
        self.body.setLinearVel(lv)
        self.head.setAngularVel(av)
        self.updateVobs()


    def getPreferredSpaceType(self):
        """
        Return our preferred space type
            0 = Enabled
            1 = Disabled
        """
        if self.body.isEnabled():
            return 0
        return 1


    @autoStore
    @withClientUpdate
    def addVisualObject(self, vob):
        """
        Add a visual object
        """
        self.vobs.append(vob)
    

    @autoStore
    @withClientUpdate
    def removeVisualObject(self, vob):
        """
        Remove a visual object
        """
        self.vobs.remove(vob)
    
    
    observe_removeVisualObject = removeVisualObject
    observe_addVisualObject = addVisualObject


    def getVisualObjects(self):
        """
        Return our list of visual objects
        """
        return self.vobs


    def updateVobs(self):
        """
        Update all the visual objects
        """
        # print "vobby"
        p = self.getPosition()
        r = self.getRotation()
        for vob in self.vobs:
            if isinstance(vob, IVisualObjectView):
                vob.setPosition(p)
                vob.setRotation(r)


    def iterate(self):
        """
        iterate this body
        """
        # print "iterate", self
        if self.lastIter is None:
            self.lastIter = time() - 0.05
        dt = time() - self.lastIter
        self.lastIter = time()
        if self.gradualOffset > 0:
            # print self.gradualOffset
            self.gradualOffset -= 1
            self.body.setPosition(self.gradualUpdate + self.body.getPosition())
        if not self.isEnabled() and not self.walking and not self.turning:
            if self.enabled:
                self.disable()
                # print "WHY"
            return
        # print self.getPosition(), self.enabled, self.parent.enabled
        if not self.enabled:
            self.enable()
        p = self.body.getPosition()
        # print p
        r = self.head.getQuaternion()
        lv = self.body.getLinearVel()
        av = self.head.getAngularVel()
        if self.walking:
            self.doWalk(dt)
        elif self.standing:
            self.doStand()
        elif self.body is not None:
            lv = self.body.getLinearVel()
            if (lv[0] * lv[0]) + (lv[2] * lv[2]) > 0:
                self.standStill()
        if self.turning:
            self.doTurn(dt)
        elif self.stopturning:
            self.doStopTurning()
        else:
            av = Vector(self.head.getAngularVel())
            if av.squaredDistance((0,0,0)) > 0:
                self.stopTurning()
                # print "FORCING IT!"
        self.standUp()
        # print "iterate",self.aggregators
        for a in self.aggregators:
            # print "Updating", self.getAngularVelocity()
            a.giveUpdate(self, p, r, lv, av)
        self.updateVobs()
        self.bodycollider.tgeom.waterDepth = 0


    def getServerClassGenerators(self):
        """
        Return class generators
        """
        cgs = [ClassGenerator("mv3d.phys.body.StringCatcher",
            sourceclass="mv3d.phys.body.StringFlinger")]
        cgs.append(ClassGenerator().setFrom(self))
        for v in self.vobs:
            cgs.append(v.getServerClassGenerators())
        return cgs


    def getClientClassGenerators(self):
        """
        Return class generators
        """
        cgs = [ClassGenerator("mv3d.phys.body.StringCatcher",
            sourceclass="mv3d.phys.body.StringFlinger")]
        cgs.append(ClassGenerator().setFrom(self))
        for v in self.vobs:
            cgs.append(v.getClientClassGenerators())
        return cgs


    def getStateToCacheAndObserveFor(self, perspective, observer):
        st = Cacheable.getStateToCacheAndObserveFor(self, perspective, observer)
        # print "yo"
        assert perspective is not None
        # print "hi"
        if not hasattr(perspective,"bodydynamicsaggregator"):
            perspective.bodydynamicsaggregator = StringFlinger()
            perspective.bodydynamicsaggregator.grantPermission("read","all")
            perspective.bodydynamicsaggregator.grantPermission("reference","all")
            perspective.parent.parent.addIterator(perspective.bodydynamicsaggregator.iterate)
        st["bodydynamicsreceiver"] = perspective.bodydynamicsaggregator
        self.aggregators.append(perspective.bodydynamicsaggregator)
        # print self.aggregators
        st["initial_position"] = self.getPosition()
        st["initial_rotation"] = self.getRotation()
        st["initial_linearVel"] = self.getLinearVelocity()
        st["initial_angularVel"] = self.getAngularVelocity()
        return st


    def stoppedObserving(self, perspective, observer):
        assert perspective is not None
        # print "stopped observing", perspective
        perspective.bodydynamicsaggregator.removeBody(self)
        self.aggregators.remove(perspective.bodydynamicsaggregator)
        return Cacheable.stoppedObserving(self, perspective, observer)

