# -*- test-case-name: mv3d.test.test_body -*-
# Copyright (C) 2007-2008 Mortal Coil Games
# See LICENSE for details.
"""
Defines a wrapper around various ode related objects so that
they automatically handle various things related to MV3D such 
as transitioning to new areas, persisting, sent over the network, etc.
"""


from array import array
import ode
from time import time

from twisted.spread import pb

from mv3d.util.datastore.store import (attributes, 
    simpleList, referenceList, classHash, number, boolean,
    randomHash, autoStore)
from mv3d.util.datastore.axiomstore import (
    DirectStore, AxiomVector, AxiomQuaternion)
from mv3d.net.pb import Cacheable, withClientUpdate
from mv3d.net.security import Securable
from mv3d.util.classgen import ClassGenerator
from mv3d.util.math3d import Vector, Quaternion
from mv3d.phys.collider import odeCollider



class BodyError(Exception):
    """
    An exception related to a Body
    """


class IBodyDynamicsAgregator:
    """
    Something that agregates dynamics updates from multiple sources
    for one connection
    """
    def giveUpdate(self, body, p, r, lv, av):
        """
        Give an update for an object
        """


    def sendNextBatch(self):
        """
        Process and send the next batch
        of updates
        """


class IBodyDynamicsReceiver:
    """
    This receives body dynamics updates
    """
    def observe_receiveBatch(self, batch):
        """
        Receive a batch of updates
        """


class DefaultRanker:
    """
    Just dumbly assign a rank of 1
    to everything
    """
    def __init__(self, agregator):
        self.agregator = agregator


    def rank(self, bid, pos, rot, lv, av):
        return 1


    def sent(self, bid):
        pass


class TimeRanker:
    """
    Rank updates by the last time the object was
    updated
    """
    maxtime = 2.0


    def __init__(self, agregator):
        self.agregator = agregator
        self.lastseen = {}


    def rank(self, bid, pos, rot, lv, av):
        last = self.lastseen.get(bid, 0.0)
        diff = time() - last
        if diff > self.maxtime:
            diff = self.maxtime
        return 1 - (diff / self.maxtime)


    def sent(self, bid):
        self.lastseen[bid] = time()



class DefaultReducer:
    """
    Send only the top x items
    """
    limit = 10


    def __init__(self, agregator):
        self.agregator = agregator


    def setLimit(self, limit):
        self.limit = limit


    def reduce(self, updates):
        """
        Take a list of updates that are assigned
        by (rank, (bid, p, r, lv, av)) and only return
        ones that are < our rank
        """
        # reduced = filter(lambda u: u[0] < self.rank, updates)
        updates.sort()
        return updates[:self.limit]



class StringFlinger(IBodyDynamicsAgregator, Cacheable, Securable):
    """
    A dynamics agregator that works by putting
    dynamics info into a string
    """
    sendevery = 10
    itercount = 0
    autoiter = True
    coiterator = None
    ranker = None
    stoppedObserving = Cacheable.stoppedObserving
    
    
    def __init__(self, ranker=None, reducer=None):
        self.updates = {}
        self.bodies = {}
        self.nextid = 0
        if ranker is not None:
            self.ranker = ranker
        else:
            self.ranker = TimeRanker(self)
        if reducer is not None:
            self.reducer = reducer
        else:
            self.reducer = DefaultReducer(self)
        self.addExclude("queue", "autoiter", "bodies", "nextid",
            "ranker", "reducer")


    # @withClientUpdate
    def addBody(self, body):
        """
        Add a new body to our list
        """
        # print "addbody", body
        assert body not in self.bodies
        self.bodies[body] = self.nextid
        self.updateClients("addBody", self.nextid, body)
        self.nextid += 1
        return self.nextid - 1


    def removeBody(self, body):
        """
        Remove a body from our list
        """
        # print "Removing body"
        if self.bodies.has_key(body):
            self.updateClients("removeBody", self.bodies[body])
            del self.bodies[body]


    def removeByOID(self, oid):
        """
        Try and remove a body by it's objectid
        """
        kill = []
        for b, i in self.bodies.iteritems():
            if hasattr(b, "objectid") and b.objectid == oid:
                kill.append(b)
        for b in kill:
            self.removeBody(b)


    def giveUpdate(self, body, p, r, lv, av):
        """
        Queue up an update
        """
        try:
            bid = self.bodies[body]
        except KeyError:
            # need to add it
            bid = self.addBody(body)
        self.updates[bid] = (self.ranker.rank(bid,
            p, r, lv, av), bid, p, r, lv, av)


    def updateToString(self, data):
        """
        Transform an update into a string
        """
        ida = array("L", [data[1]])
        dyna = array("f")
        for d in data[2:]:
            dyna.fromlist(list(d))
        return ida.tostring() + dyna.tostring()


    def sendNextBatch(self):
        """
        Process and send the next batch
        of updates
        """
        batch = self.reducer.reduce(list(self.updates.values()))
        [self.ranker.sent(u[1]) for u in batch]
        data = [self.updateToString(u) for u in batch]
        string = "".join(data)
        self.updateClients("receiveBatch", "".join(data))
        self.updates = {}
        return len(batch)


    def iterate(self, t):
        if self.itercount % self.sendevery == 0:
            self.sendNextBatch()
        self.itercount += 1


    def getStateToCacheAndObserveFor(self, perspective, observer):
        st = Cacheable.getStateToCacheAndObserveFor(self, perspective, observer)
        # Reverse the bodies dict
        st["bodies"] = {}
        for k, v in self.bodies.items():
            st["bodies"][v] = k
        # print "sending bodies", st["bodies"]
        return st


    def setCopyableState(self, state):
        raise BodyError("Can't set the state of an IDynamicsBodyFlinger!")



class StringCatcher(IBodyDynamicsReceiver, Cacheable):
    """
    Receive dynamics updates for bodies
    """
    updatesize = 56
    # setCopyableState = Cacheable.setCopyableState
    def setCopyableState(self, state):
        # print "-=-==-- ", state
        Cacheable.setCopyableState(self, state)

    
    def observe_addBody(self, bid, body=None):
        # print "adding body!", bid
        self.bodies[bid] = body


    def observe_removeBody(self, bid):
        # print "Removing body", bid
        del self.bodies[bid]


    def processUpdate(self, upstr):
        """
        Take an string for one update and
        apply it to an object
        """
        if hasattr(self.server, "lastups"):
            self.server.lastups += 1
        bid = array("L", upstr[:4])[0]
        dyna = array("f", upstr[4:])
        p = Vector(tuple(dyna[:3]))
        r = Quaternion(tuple(dyna[3:7]))
        lv = Vector(tuple(dyna[7:10]))
        av = Vector(tuple(dyna[10:13]))
        # print bid, self.bodies
        # print "Updated", bid
        try:
            self.bodies[bid].updateDynamics(p, r, lv, av)
        except KeyError:
            # ignore non-existent body updates
            # print "missing", bid
            pass

    
    def observe_receiveBatch(self, data):
        """
        Process a batch of updates
        """
        # return
        # print "got batch"
        for x in range(0, len(data), self.updatesize):
            up = data[x:x+self.updatesize]
            self.processUpdate(up)



class IBody:
    """
    Defines a body which can have physical,
    visual, and even audio representations
    """
    datastoreAttributes = attributes(colliders=referenceList)

    colliders = None
    
    
    def create(self):
        """
        Create the body
        """


    def enterArea(self, area):
        """
        This body has been inserted into an area
        """
        
        
    def leaveArea(self, area):
        """
        This body is leaving the area
        """


    def hasSecondaryArea(self, areaid):
        """
        Returns true if this body has a secondary
        representation in this area
        """
        return False


    @withClientUpdate
    def addCollider(self, collider):
        """
        Add a collider
        """
        if self.colliders is None:
            self.colliders = []
        self.colliders.append(collider)


    def observe_addCollider(self, collider):
        self.addCollider(collider)
        
        
    def getColliders(self):
        """
        Return all of our colliders
        """
        if self.colliders is None:
            self.colliders = []
        return self.colliders


    def getPosition(self):
        """
        Returns the position of this body
        """
        raise BodyError("%s doesn't support getPosition." % 
            self.__class__.__name__)


    def setPosition(self):
        """
        Sets the position of this body
        """
        raise BodyError("%s doesn't support setPosition." % 
            self.__class__.__name__)


    def getRotation(self):
        """
        Returns the rotation of this body
        """
        raise BodyError("%s doesn't support getRotation." % 
            self.__class__.__name__)


    def setRotation(self):
        """
        Sets the rotation of this body
        """
        raise BodyError("%s doesn't support setRotation." % 
            self.__class__.__name__)


    def getLinearVelocity(self):
        """
        Returns the linear velocity of this body
        """
        raise BodyError("%s doesn't support getLinearVelocity." % 
            self.__class__.__name__)


    def setLinearVelocity(self):
        """
        Sets the linear velocity of this body
        """
        raise BodyError("%s doesn't support setLinearVelocity." % 
            self.__class__.__name__)


    def getAngularVelocity(self):
        """
        Returns the angular velocity of this body
        """
        raise BodyError("%s doesn't support getAngularVelocity." % 
            self.__class__.__name__)


    def updateDynamics(self, p, r, lv, av):
        """
        Update all dynamics properties
        """
        raise BodyError("%s doesn't support updateDynamics." %
            self.__class__.__name__)


    def setAngularVelocity(self):
        """
        Sets the angular velocity of this body
        """
        raise BodyError("%s doesn't support setAngularVelocity." % 
            self.__class__.__name__)


    def getPreferredSpaceType(self):
        """
        Return our preferred space type
            0 = Enabled
            1 = Disabled
            2 = Special
        """
        raise BodyError("%s doesn't support getPreferredSpaceType." % 
            self.__class__.__name__)


    def attach(self, other, joint):
        """
        Attach this body to another one using the joint
        """
        raise BodyError("%s doesn't support attach." % 
            self.__class__.__name__)


    def enable(self):
        """
        Set the state of this body to be enabled
        """
        raise BodyError("%s doesn't support enable." % 
            self.__class__.__name__)


    def disable(self):
        """
        Set the state of this body to be disabled
        """
        raise BodyError("%s doesn't support disable." % 
            self.__class__.__name__)


    def isEnabled(self):
        """
        Returns true if this body is enabled
        """
        raise BodyError("%s doesn't support isEnabled." % 
            self.__class__.__name__)


    def setWaterDepth(self, depth):
        """
        Set the depth of water on this object
        """
        raise BodyError("%s doesn't support setWaterDepth." % 
            self.__class__.__name__)


    def getWaterDepth(self):
        """
        Returns the depth of water on this object
        """
        raise BodyError("%s doesn't support getWaterDepth." % 
            self.__class__.__name__)


    def iterate(self):
        """
        iterate this body
        """


    def getServerClassGenerators(self):
        """
        Return class generators
        """


    def getClientClassGenerators(self):
        """
        Return class generators
        """


    def observe_setPosition(self, pos):
        self.setPosition(pos)


    def observe_setRotation(self, rot):
        self.setRotation(rot)


    def observe_setLinearVelocity(self, lv):
        self.setLinearVelocity(lv)


    def observe_setAngularVelocity(self, av):
        self.setAngularVelocity(av)


    def observe_disable(self):
        self.disable()


    def observe_enable(self):
        self.enable()



class MobileBody(IBody, Cacheable, Securable):
    """
    A physical and visual body that can exist in multiple areas
    at the same time
    """
    datastoreAttributes = attributes(IBody, Cacheable, Securable,
        areaid=number, objectid=number, enabled=boolean, 
        joints=referenceList, vobs=referenceList,
        initial_position=DirectStore(AxiomVector, True,
            u"getPosition"),
        initial_rotation=DirectStore(AxiomQuaternion, True,
            u"getRotation"),
        initial_linearVel=DirectStore(AxiomVector, True,
            u"getLinearVelocity"),
        initial_angularVel=DirectStore(AxiomVector, True,
            u"getAngularVelocity")
        )
    datastoreIndexes = tuple()
    datastoreHash = ("objectid", classHash, randomHash)
    datastoreVersion = 1

    body = None
    space = None
    colliderspace = None
    areaid = None
    area = None
    objectid = None
    initial_position = None
    initial_rotation = None
    initial_linearVel = None
    initial_angularVel = None
    enabled = True
    waterdepth = 0


    def __init__(self, objectid=None):
        self.objectid = objectid
        self.joints = []
        self.colliders = []
        self.vobs = []
        self.otherspacecolliders = {}
        self.otherspacespaces = {}
        self.otherspacecolliderspaces = {}
        self.aggregators = []
        self.addExclude("body", "world", "space", "colliderspace", 
            "otherspacecolliders", "otherspacespaces", "aggregators",
            "area", "parent")


    def loadedFromStore(self):
        self.otherspacecolliders = {}
        self.otherspacespaces = {}
        self.otherspacecolliderspaces = {}
        self.aggregators = []


    def create(self, world, space=None):
        """
        Create the body and optionally the colliders if space is 
        defined
        """
        self.world = world
        self.body = ode.Body(self.world)
        self.body.mvbody = self
        if self.initial_position is not None:
            self.body.setPosition(self.initial_position)
        if self.initial_rotation is not None:
            self.body.setQuaternion(self.initial_rotation)
        if self.initial_linearVel is not None:
            self.body.setLinearVel(self.initial_linearVel)
        if self.initial_angularVel is not None:
            self.body.setAngularVel(self.initial_angularVel)
        if space is not None:
            self.enterSpace(space)
        if not self.enabled:
            self.body.disable()


    def destroy(self):
        """
        Destroy the body
        """
        self.world = None
        for c in self.colliders:
            c.destroy()
        if self.colliderspace is not None:
            self.colliderspace.destroy()
        self.body = None
        self.space = None
        for v in self.vobs:
            v.destroy()


    def enterSpace(self, space):
        """
        Create colliders in this space
        """
        # print "---Enter Space", space, self.space, self.colliders
        assert self.body is not None
        if self.colliderspace is not None:
            self.colliderspace.destroy()
        self.colliderspace = odeCollider("Space")
        self.colliderspace.build(None, space)
        # print self.colliderspace.geom, self.body
        for c in self.colliders:
            c.destroy()
            c.build(body=self.body, space=self.colliderspace.geom)
        self.space = space

    
    def enterSecondarySpace(self, ownerid, space):
        """
        Create secondary colliders in this space
        """
        self.otherspacecolliders[ownerid] = []
        self.otherspacespaces[ownerid] = space
        if self.otherspacecolliderspaces.has_key(ownerid):
            self.otherspacecolliderspaces.destroy()
        self.otherspacecolliderspaces[ownerid] = odeCollider("Space")
        self.otherspacecolliderspaces[ownerid].build(space=space)
        for collider in self.colliders:
            c = collider.copy()
            c.density = 0.0
            c.build(self.body, self.otherspacecolliderspaces[ownerid].geom)
            self.otherspacecolliders[ownerid].append(c)


    def enterArea(self, area):
        """
        Determine the space most appropriate
        in are and enter it. Area should be an
        area object
        """
        # print "Enter Area", area
        if self.body is None:
            self.create(area.world) #Hack!
        if self.isEnabled():
            space = area.getEnabledSpace(self.getPosition())
        else:
            space = area.getDisabledSpace(self.getPosition())
        self.enterSpace(space)
        self.areaid = area.getID()
        self.area = area
        for v in self.vobs:
            v.enterArea(area)


    def leaveArea(self, area):
        """
        This body is leaving the area, remove
        everything in the area
        """
        self.colliderspace.destroy()
        for c in self.colliders:
            c.destroy()
        for v in self.vobs:
            v.leaveArea(area)
        self.space = None


    def leaveSecondarySpace(self, ownerid):
        """
        Remove our secondary info from this space
        """
        for c in self.otherspacecolliders[ownerid]:
            c.destroy()
        del self.otherspacecolliders[ownerid]
        del self.otherspacespaces[ownerid]
        self.otherspacecolliderspaces[ownerid].destroy()
        del self.otherspacecolliderspaces[ownerid]


    def hasSecondarySpace(self, ownerid):
        """
        Returns true if we have colliders in this secondary space
        """
        return ownerid in self.otherspacespaces.keys()


    @autoStore
    def attach(self, other, joint):
        """
        Attach this body to another one using the joint
        """
        joint.attach(self, other)
        self.joints.append(joint)
        other.joints.append(joint)


    @withClientUpdate
    def setPosition(self, pos):
        """
        Set the position of this body
        """
        self.body.setPosition(pos)
        self.updateVobs()


    def getPosition(self):
        """
        Get the position of this body
        """
        return self.body.getPosition()


    @withClientUpdate
    def setRotation(self, rot):
        """
        Set the quaternion rotation of this body
        """
        self.body.setQuaternion(rot)
        self.updateVobs()


    def getRotation(self):
        """
        Get the quaternion rotation of this body
        """
        return self.body.getQuaternion()


    @withClientUpdate
    def setLinearVelocity(self, lv):
        """
        Set the linear velocity of this body
        """
        self.body.setLinearVel(lv)
        
        
    def getLinearVelocity(self):
        """
        Returns the linear velocity of this body
        """
        return self.body.getLinearVel()


    @withClientUpdate
    def setAngularVelocity(self, av):
        """
        Set the angular Velocity of this body
        """
        self.body.setAngularVel(av)
        
        
    def getAngularVelocity(self):
        """
        Get the angular velocity of this body
        """
        return self.body.getAngularVel()


    @withClientUpdate
    def enable(self):
        """
        Set the state of this body to be enabled
        """
        # print "Enabling", self.objectid, time()
        self.body.enable()
        self.setLinearVelocity(self.getLinearVelocity())
        self.setAngularVelocity(self.getAngularVelocity())
        # self.setPosition(self.getPosition())
        # self.setRotation(self.getRotation())
        if not self.enabled:
            if self.objectid is not None:
                self.parent.enableItem(self.objectid)
                # print "Boung"
            if self.area is not None:
                self.enterSpace(self.area.getEnabledSpace(self.getPosition()))
        self.enabled = True


    @withClientUpdate
    def disable(self):
        """
        Set the state of this body to be disabled
        """
        # print "Disabling body"
        self.body.disable()
        self.setLinearVelocity((0,0,0))
        self.setAngularVelocity((0,0,0))
        self.setPosition(self.getPosition())
        self.setRotation(self.getRotation())
        if self.enabled and self.area is not None:
            self.enterSpace(self.area.getDisabledSpace(self.getPosition()))
        self.enabled = False


    def isEnabled(self):
        """
        Returns true if this body is enabled
        """
        return self.body.isEnabled()


    def updateDynamics(self, pos, rot, lv, av):
        """
        Update the position, rotation, linear and angular velocities
        """
        if self.body is None:
            # print "Update for dead body.", self.objectid
            self.initial_position = pos
            self.initial_rotation = rot
        else:
            self.body.setPosition(pos)
            self.body.setQuaternion(rot)
            self.body.setLinearVel(lv)
            self.body.setAngularVel(av)
        # self.updateVobs()


    def getPreferredSpaceType(self):
        """
        Return our preferred space type
            0 = Enabled
            1 = Disabled
        """
        if self.body.isEnabled():
            return 0
        return 1


    @autoStore
    @withClientUpdate
    def addVisualObject(self, vob):
        """
        Add a visual object
        """
        self.vobs.append(vob)
    

    @autoStore
    @withClientUpdate
    def removeVisualObject(self, vob):
        """
        Remove a visual object
        """
        self.vobs.remove(vob)
    
    
    observe_removeVisualObject = removeVisualObject
    observe_addVisualObject = addVisualObject


    def getVisualObjects(self):
        """
        Return our list of visual objects
        """
        return self.vobs


    def updateVobs(self):
        """
        Update all the visual objects
        """
        p = self.getPosition()
        r = self.getRotation()
        for vob in self.vobs:
            if isinstance(vob, IVisualObjectView):
                vob.setPosition(p)
                vob.setRotation(r)


    def setWaterDepth(self, depth):
        """
        Set the depth of water
        """
        self.waterdepth = depth


    def getWaterDepth(self, depth):
        """
        Get the depth of water
        """
        return self.waterdepth


    def iterate(self):
        """
        iterate this body
        """
        if self.isEnabled():
            # print "iterate", self
            if not self.enabled:
                self.enable()
            p = self.body.getPosition()
            # print p
            r = self.body.getQuaternion()
            lv = self.body.getLinearVel()
            av = self.body.getAngularVel()
            for a in self.aggregators:
                a.giveUpdate(self, p, r, lv, av)
            self.updateVobs()
        else:
            if self.enabled:
                self.disable()


    def getServerClassGenerators(self):
        """
        Return class generators
        """
        cgs = [ClassGenerator("mv3d.phys.body.StringCatcher",
            sourceclass="mv3d.phys.body.StringFlinger")]
        cgs.append(ClassGenerator().setFrom(self))
        for v in self.vobs:
            cgs.append(v.getServerClassGenerators())
        return cgs


    def getClientClassGenerators(self):
        """
        Return class generators
        """
        cgs = [ClassGenerator("mv3d.phys.body.StringCatcher",
            sourceclass="mv3d.phys.body.StringFlinger")]
        cgs.append(ClassGenerator().setFrom(self))
        for v in self.vobs:
            cgs.append(v.getClientClassGenerators())
        return cgs


    def getStateToCacheAndObserveFor(self, perspective, observer):
        st = Cacheable.getStateToCacheAndObserveFor(self, perspective, observer)
        assert perspective is not None
        if not hasattr(perspective,"bodydynamicsaggregator"):
            # print "_+_+_+_+_+_+Creating new bodydynamicsaggregator"
            perspective.bodydynamicsaggregator = StringFlinger()
            perspective.bodydynamicsaggregator.grantPermission("read","all")
            perspective.bodydynamicsaggregator.grantPermission("reference","all")
            perspective.parent.addIterator(perspective.bodydynamicsaggregator.iterate)
        st["bodydynamicsreceiver"] = perspective.bodydynamicsaggregator
        self.aggregators.append(perspective.bodydynamicsaggregator)
        # perspective.bodydynamicsaggregator.addBody(self)
        st["initial_position"] = self.getPosition()
        st["initial_rotation"] = self.getRotation()
        # print "=---------------- getStateToCacheAndObserveFor", st
        return st


    def stoppedObserving(self, perspective, observer):
        assert perspective is not None
        perspective.bodydynamicsaggregator.removeBody(self)
        self.aggregators.remove(perspective.bodydynamicsaggregator)
        return Cacheable.stoppedObserving(self, perspective, observer)


    def setCopyableState(self, state):
        # print "Body state", state
        Cacheable.setCopyableState(self, state)
        #create(world?)



class StationaryBody(IBody, Cacheable, Securable):
    """
    A body that is stationary
    """
    datastoreAttributes = attributes(IBody, Cacheable, Securable,
        areaid=number, special=number, position=simpleList, 
        rotation=simpleList, vobs=referenceList)
    datastoreIndexes = tuple()
    datastoreHash = ("areaid", classHash, randomHash)
    datastoreVersion = 1

    getStateToCacheAndObserveFor = Cacheable.getStateToCacheAndObserveFor
    stoppedObserving = Cacheable.stoppedObserving
    setCopyableState = Cacheable.setCopyableState

    space = None
    special = False
    colliderspace = None
    areaid = None
    position = (0,0,0)
    rotation = (1,0,0,0)


    def __init__(self):       
        self.vobs = []
        self.colliders = []
        self.transformedcolliders = []
        self.addExclude("parent")


    def loadedFromStore(self):
        self.transformedcolliders = []


    def create(self, world, space=None):
        """
        Create the body and optionally the colliders if space is 
        defined
        """
        if space is not None:
            self.enterSpace(space)


    def getPreferredSpaceType(self):
        """
        Return our preferred space type
            1 = Disabled
            2 = Special
        """
        if self.special:
            return 2
        return 1


    @autoStore
    @withClientUpdate
    def setPosition(self, pos):
        """
        Set the position of this body
        """
        self.position = pos
        self.updateVobs()
        self.enterSpace(self.space)


    def getPosition(self):
        """
        Get the position of this body
        """
        return self.position


    @autoStore
    @withClientUpdate
    def setRotation(self, rot):
        """
        Set the quaternion rotation of this body
        """
        self.rotation = rot
        self.updateVobs()
        self.enterSpace(self.space)


    def getRotation(self):
        """
        Get the quaternion rotation of this body
        """
        return self.rotation


    def getColliders(self):
        """
        Return the transformed colliders if there are any
        otherwise return the defined colliders
        """
        if len(self.transformedcolliders):
            return self.transformedcolliders
        return self.colliders


    def enable(self):
        """
        Can't be enabled
        """


    def disable(self):
        """
        Can't be disabled
        """


    def isEnabled(self):
        """
        We can't be enabled
        """
        return False


    def enterSpace(self, space):
        """
        Create colliders in this space
        """
        if self.colliderspace is not None:
            self.colliderspace.destroy()
        self.colliderspace = odeCollider("Space")
        self.colliderspace.build(space=space)
        for c in self.transformedcolliders:
            c.destroy()
        self.transformedcolliders = []
        for c in self.colliders:
            newc = c.copy()
            if hasattr(newc, "position"):
                newc.setPosition(c.position + self.position)
            if hasattr(newc, "rotation"):
                newc.setRotation(c.rotation * self.rotation)
            newc.build(space=self.colliderspace.geom)
            self.transformedcolliders.append(newc)
        self.space = space


    def enterArea(self, area):
        """
        Determine the space most appropriate
        in are and enter it. Area should be an
        area object
        """
        if self.special:
            space = area.getSpecialSpace(self.getPosition())
        else:
            space = area.getDisabledSpace(self.getPosition())
        self.enterSpace(space)
        self.areaid = area.getID()


    @autoStore
    @withClientUpdate
    def addVisualObject(self, vob):
        """
        Add a visual object
        """
        self.vobs.append(vob)
    

    @autoStore
    @withClientUpdate
    def removeVisualObject(self, vob):
        """
        Remove a visual object
        """
        self.vobs.remove(vob)


    observe_removeVisualObject = removeVisualObject
    observe_addVisualObject = addVisualObject


    def getVisualObjects(self):
        """
        Get all the visual objects
        """
        return self.vobs


    def updateVobs(self):
        """
        Update all the visual objects
        """
        p = self.getPosition()
        r = self.getRotation()
        for vob in self.vobs:
            if isinstance(vob, IVisualObjectView):
                vob.setPosition(p)
                vob.setRotation(r)


    def getServerClassGenerators(self):
        """
        Return class generators
        """
        cgs = []
        cgs.append(ClassGenerator().setFrom(self))
        for v in self.vobs:
            cgs.append(v.getServerClassGenerators())
        return cgs


    def getClientClassGenerators(self):
        """
        Return class generators
        """
        cgs = []
        cgs.append(ClassGenerator().setFrom(self))
        for v in self.vobs:
            cgs.append(v.getClientClassGenerators())
        return cgs



class IJoint:
    """
    A force that holds two Bodies together
    """



class Joint(IJoint, Cacheable):
    """
    A standard joint
    """
    type = None
    joint = None

    def __init__(self, type=None):
        self.type = type


    def attach(self, world, bodya, bodyb):
        self.createJoint(world)
        self.joint.attach(bodya.body, bodyb.body)


    def createBallJoint(self, world):
        self.joint = ode.BallJoint(world)


    def createJoint(self, world):
        assert self.type is not None
        return getattr(self, "create"+self.type)(world)


    def setAnchor(self, pos):
        self.joint.setAnchor(pos)



class IVisualObject:
    """
    This can be used to describe how an object looks to the client
    """
    def enterArea(self, aid):
        """
        Enter an area
        """


    def destroy(self):
        """
        Destroy this object
        """


    def getAssets(self):
        """
        Return all the assets required to
        display this object
        """


    def addAsset(self, aid):
        """
        Add an asset to the object
        """


    def removeAsset(self, aid):
        """
        Renive an asset from the object
        """


    def setScale(self, scale):
        """
        Set our scsale
        """


    def startAnimation(self, name, loop=False, weight=100):
        """
        Start an animation on this object
        """


    def stopAnimation(self, name):
        """
        Stop an animation on this object
        """


    observe_setScale = setScale
    observe_startAnimation = startAnimation
    observe_stopAnimation = stopAnimation
    observe_addAsset = addAsset
    observe_removeAsset = removeAsset
    observe_enterArea = enterArea


    def getServerClassGenerators(self):
        """
        Returns class generators for a server
        """


    def getClientClassGenerators(self):
        """
        Returns class generators for a client
        """



class IVisualObjectView:
    """
    The client side of a visual object
    """
    def observe_enterArea(self, aid):
        """
        Enter an area
        """


    def observe_leaveArea(self, aid):
        """
        Leave an area
        """


    def observe_addAsset(self, aid):
        """
        Add an asset to the object
        """


    def observe_setScale(self, scale):
        """
        Set our scsale
        """


    def observe_startAnimation(self, name, loop, weight):
        """
        Start an animation
        """


    def observe_stopAnimation(self, name):
        """
        Stop an animation
        """


    def setPosition(self, pos):
        """
        Set the position of this object
        """


    def setRotation(self, rot):
        """
        Set the rotation of this object
        """



class BasicVisualObject(IVisualObject, Cacheable, Securable):
    """
    A basic visual object
    """
    datastoreAttributes = attributes(Cacheable, Securable,
        areaid=number, scale=simpleList, posoffset=simpleList, 
        rotoffset=simpleList, assets=simpleList,
        currentanimations=simpleList)
    datastoreIndexes = tuple()
    datastoreHash = ("areaid", classHash, randomHash)
    datastoreVersion = 1

    stoppedObserving = Cacheable.stoppedObserving
    getStateToCacheAndObserveFor = Cacheable.getStateToCacheAndObserveFor
    setCopyableState = Cacheable.setCopyableState

    areaid = None
    scale = Vector(1.0, 1.0, 1.0)
    posoffset = Vector(0,0,0)
    rotoffset = Quaternion(1,0,0,0)


    def __init__(self, **kw):
        self.assets =  []
        self.currentanimations = {}


    def enterArea(self, area):
        """
        Enter an area
        """
        self.areaid = area.getID()


    def leaveArea(self, area):
        """
        Leave an area
        """
        self.areaid = None


    def getAssets(self):
        return self.asserts


    @withClientUpdate
    def addAsset(self, aid):
        """
        Add an asset
        """
        self.assets.append(aid)


    @withClientUpdate
    def removeAsset(self, aid):
        """
        Remove an asset from the object
        """
        self.assets.remove(aid)


    @withClientUpdate
    def setScale(self, scale):
        """
        Set our scsale
        """
        self.scale = scale


    @withClientUpdate
    def setOffset(self, p, r):
        """
        Set the offset for this node
        that gets applied when we set
        the position or rotation
        """
        self.posoffset = p
        self.rotoffset = r


    @withClientUpdate
    def startAnimation(self, name, loop=False, weight=100):
        """
        Start an animation
        """
        if loop:
            self.currentanimations[name] = (loop, weight)


    @withClientUpdate
    def stopAnimation(self, name):
        """
        Stop an animation
        """
        if self.currentanimations.has_key(name):
            del self.currentanimations[name]


    observe_setScale = setScale
    observe_setOffset = setOffset
    observe_startAnimation = startAnimation
    observe_stopAnimation = stopAnimation


    def getServerClassGenerators(self):
        """
        Returns class generators for a server
        """
        return ClassGenerator(classname="BasicVisualObject", 
            modulename="mv3d.phys.body", sourceclass="mv3d.phys.body.BasicVisualObject")


    def getClientClassGenerators(self):
        """
        Returns class generators for a client
        """
        return ClassGenerator(classname="BasicVisualObjectView", 
            modulename="mv3d.client.view.visual", sourceclass="mv3d.phys.body.BasicVisualObject")


class WaterPlaneVisualObject(IVisualObject, Cacheable, Securable):
    """
    A plane that is made of water
    This is special because water is done with RTT
    """
    datastoreAttributes = attributes(Cacheable, Securable,
        areaid=number, distance=number, normal=number, 
        assets=referenceList)
    datastoreIndexes = tuple()
    datastoreHash = ("areaid", classHash, randomHash)
    datastoreVersion = 1

    stoppedObserving = Cacheable.stoppedObserving
    getStateToCacheAndObserveFor = Cacheable.getStateToCacheAndObserveFor
    setCopyableState = Cacheable.setCopyableState

    areaid = None
    distance = 0.0
    normal = Vector(0,1,0)


    def __init__(self):
        self.assets = []


    def enterArea(self, area):
        """
        Enter an area
        """
        self.areaid = area.getID()


    def getAssets(self):
        return self.asserts


    @withClientUpdate
    def addAsset(self, aid):
        """
        Add an asset
        """
        self.assets.append(aid)


    @withClientUpdate
    def removeAsset(self, aid):
        """
        Remove an asset from the object
        """
        self.assets.remove(aid)


    @withClientUpdate
    def setPlane(self, distance, normal):
        """
        Set the plane
        """
        self.distance = distance
        self.normal = normal


    observe_setPlane = setPlane


    def getServerClassGenerators(self):
        """
        Returns class generators for a server
        """
        return ClassGenerator(classname="WaterPlaneVisualObject", 
            modulename="mv3d.phys.body", 
            sourceclass="mv3d.phys.body.WaterPlaneVisualObject")


    def getClientClassGenerators(self):
        """
        Returns class generators for a client
        """
        return ClassGenerator(classname="WaterPlaneVisualObjectView", 
            modulename="mv3d.client.view.visual", 
            sourceclass="mv3d.phys.body.WaterPlaneVisualObject")



pb.setUnjellyableForClass(StringFlinger, StringCatcher)
