# Copyright (C) 2006-2008 Mortal Coil Games
# See LICENSE for details.
""" -*- test-case-name: mv3d.test.test_collider -*-

"""

import sys
import ode

from twisted.spread import pb

from mv3d.util.datastore.store import (attributes, text,
    simpleList, reference, classHash, number, boolean, randomHash)
from mv3d.util.math3d import Vector, Quaternion
from mv3d.net.pb import Copyable


class ColliderError(Exception):
    """
    Raised when there is an issue with a
    collider
    """


class ICollider:
    """
    Something that can be collided with in
    the physical simulation
    """
    def __getstate__(self):
        """
        Skip a few things
        """
        state = self.__dict__.copy()
        if state.has_key("geom"):
            del state["geom"]
        if state.has_key("tgeom"):
            del state["tgeom"]
        if state.has_key("oncollide"):
            del state["oncollide"]
        return state


    def create(self, body=None, space=None):
        """
        Create this collider in space (if provided)
        and attach it to body (if provided)
        """


    def setGeomParams(self, geom):
        """
        Set misc parameters for the geom. 
        """
        geom.contact_type = 0
        for x in self.contacttype:
            geom.contact_type |= getattr(ode, x)
        geom.contact_mu = self.mu
        geom.contact_mu2 = self.mu2
        geom.contact_bounce = self.bounce
        geom.contact_bounce_vel = self.bounce_vel
        geom.contact_soft_erp = self.soft_erp
        geom.contact_soft_cfm = self.soft_cfm
        geom.contact_slip1 = self.slip1
        geom.contact_slip2 = self.slip2
        geom.oid = self.oid
        geom.oncollide = self.oncollide
        if hasattr(self, "density"):
            geom.density = self.density
        geom.rollfrict = self.rollfrict


    def destroy(self):
        """
        Destroy this geom
        """
        if self.geom is not None:
            deleteGeom(self.geom)
            self.geom = None
        if hasattr(self, "tgeom") and self.tgeom is not None:
            deleteGeom(self.tgeom)
            self.tgeom = None


    def getPosition(self):
        """
        Return the position of the collider
        """
        return self.position


    def setPosition(self, pos):
        """
        Set the position of the geom
        """
        self.position = pos
        if self.geom is not None:
            self.geom.setPosition(pos)


    def getRotation(self):
        """
        Return the rotation of the collider
        """
        return self.rotation


    def setRotation(self, rot):
        """
        Set the rotation of the geom
        """
        self.rotation = rot
        if self.geom is not None:
            self.geom.setQuaternion(rot)



class odeBox(ICollider, Copyable):
    """
    A box from the ode Physics engine
    """
    datastoreAttributes = attributes(Copyable,
        position=simpleList, rotation=simpleList, size=simpleList,
        density=number, stationary=boolean, contacttype=simpleList,
        mu=number, mu2=number, bounce=number, bounce_vel=number,
        soft_erp=number, soft_cfm=number, slip1=number,
        slip2=number, oid=number, rollfrict=number)
    datastoreIndexes = tuple()
    datastoreHash = ("oid", "position", classHash, randomHash)
    datastoreVersion = 1
    datastore = None

    position = Vector(0,0,0)
    rotation = Quaternion(1,0,0,0)
    size = Vector(1,1,1)
    density = 1.0
    stationary = False
    mu = 10000.0
    mu2 = 0.0
    bounce = 0.1
    bounce_vel = 0.0
    soft_erp = 0.0
    soft_cfm = 0.0
    slip1 = 0.0
    slip2 = 0.0
    rollfrict = 0.0
    oid = None
    oncollide = None
    geom = None
    tgeom = None
    contacttype = None


    def __init__(self, **kw):
        self.contacttype = ["ContactBounce"]
        for k, v in kw.iteritems():
            if not hasattr(self.__class__, k):
                raise ValueError("Inalid keyword argument %s to odeBox" % k)
        self.__dict__.update(kw)
        self.addExclude("geom", "tgeom", "oncollide")


    def copy(self):
        """
        Return a copy of this collider
        """
        return odeBox(datastore=self.datastore, position=self.position,
            rotation=self.rotation, size=self.size, density=self.density,
            stationary=self.stationary, contacttype=self.contacttype,
            mu=self.mu, mu2=self.mu2, bounce=self.bounce,
            bounce_vel=self.bounce_vel, soft_erp=self.soft_erp,
            soft_cfm=self.soft_cfm, slip1=self.slip1, slip2=self.slip2,
            oid=self.oid, oncollide=self.oncollide, rollfrict=self.rollfrict)


    def build(self, body=None, space=None):
        """
        Create the box in the space attached to the
        body (if either is provided)
        """
        if self.stationary:
            body = 0
        # print "box size", size, "dens", density, "bod", body
        if self.density == 0:
            self.density = 1.0
        M = ode.Mass()
        M.setBox(self.density, *list(self.size))
        if body is not None:
            body.setMass(M + body.getMass())
            self.geom = ode.GeomBox(lengths=self.size)
        else:
            self.geom = ode.GeomBox(space, lengths=self.size)
        self.geom.setPosition(self.position)
        if len(self.rotation) != 4:
            self.geom.setRotation(self.rotation)
        else:
            self.geom.setQuaternion(self.rotation)
        self.setGeomParams(self.geom)
        if body is not None:
            self.tgeom = ode.GeomTransform(space)
            self.tgeom.setGeom(self.geom)
            self.tgeom.setBody(body)
            self.geom.body = body
            self.setGeomParams(self.tgeom)


class odeSphere(ICollider, Copyable):
    """
    A box from the ode Physics engine
    """
    datastoreAttributes = attributes(Copyable,
        position=simpleList, rotation=simpleList, radius=number,
        density=number, stationary=boolean, contacttype=simpleList,
        mu=number, mu2=number, bounce=number, bounce_vel=number,
        soft_erp=number, soft_cfm=number, slip1=number,
        slip2=number, oid=number, rollfrict=number)
    datastoreIndexes = tuple()
    datastoreHash = ("oid", "position", classHash, randomHash)
    datastoreVersion = 1
    datastore = None

    position = Vector(0,0,0)
    rotation = Quaternion(1,0,0,0)
    radius = 1.0
    density = 1.0
    stationary = False
    geom = None
    tgeom = None
    mu = 10000.0
    mu2 = 0.0
    bounce = 0.1
    bounce_vel = 0.0
    soft_erp = 0.0
    soft_cfm = 0.0
    slip1 = 0.0
    slip2 = 0.0
    oid = None
    oncollide = None
    rollfrict = 0.0
    contacttype = None


    def __init__(self, **kw):
        self.contacttype = ["ContactBounce"]
        for k, v in kw.iteritems():
            if not hasattr(self.__class__, k):
                raise ValueError("Inalid keyword argument %s to odeSphere" % k)
        self.__dict__.update(kw)
        self.addExclude("geom", "tgeom", "oncollide")


    def copy(self):
        """
        Return a copy of this collider
        """
        return odeSphere(datastore=self.datastore, position=self.position,
            rotation=self.rotation, radius=self.radius, density=self.density,
            stationary=self.stationary, contacttype=self.contacttype,
            mu=self.mu, mu2=self.mu2, bounce=self.bounce,
            bounce_vel=self.bounce_vel, soft_erp=self.soft_erp,
            soft_cfm=self.soft_cfm, slip1=self.slip1, slip2=self.slip2,
            oid=self.oid, oncollide=self.oncollide, rollfrict=self.rollfrict)


    def build(self, body=None, space=None):
        """
        Create the cylinder
        """
        if self.stationary:
            body = None
        M = ode.Mass()
        M.setSphere(self.density, self.radius)
        if body is not None:
            self.geom = ode.GeomSphere(radius=self.radius)
            body.setMass(M + body.getMass())
        else:
            self.geom = ode.GeomSphere(space, radius=self.radius)
        self.geom.setPosition(self.position)
        if len(self.rotation) != 4:
            self.geom.setRotation(self.rotation)
        else:
            self.geom.setQuaternion(self.rotation)
        self.setGeomParams(self.geom)
        if body is not None:
            self.tgeom = ode.GeomTransform(space)
            self.tgeom.setGeom(self.geom)
            self.tgeom.setBody(body)
            self.geom.body = body
            self.setGeomParams(self.tgeom)



class odeCylinder(ICollider, Copyable):
    """
    A box from the ode Physics engine
    """
    datastoreAttributes = attributes(Copyable,
        position=simpleList, rotation=simpleList, radius=simpleList,
        length=number, density=number, stationary=boolean, 
        contacttype=simpleList, mu=number, mu2=number, bounce=number, 
        bounce_vel=number, soft_erp=number, soft_cfm=number, slip1=number,
        slip2=number, oid=number, rollfrict=number)
    datastoreIndexes = tuple()
    datastoreHash = ("oid", "radius", classHash, randomHash)
    datastoreVersion = 1
    datastore = None

    position = Vector(0,0,0)
    rotation = Quaternion(1,0,0,0)
    radius = 1.0
    length = 1.0
    density = 1.0
    stationary = False
    geom = None
    tgeom = None
    mu = 10000.0
    mu2 = 0.0
    bounce = 0.1
    bounce_vel = 0.0
    soft_erp = 0.0
    soft_cfm = 0.0
    slip1 = 0.0
    slip2 = 0.0
    oid = None
    oncollide = None
    rollfrict = 0.0
    contacttype = None


    def __init__(self, **kw):
        self.contacttype = ["ContactBounce"]
        for k, v in kw.iteritems():
            if not hasattr(self.__class__, k):
                raise ValueError("Inalid keyword argument %s to odeCylinder" % k)
        self.__dict__.update(kw)
        self.addExclude("geom", "tgeom", "oncollide")


    def copy(self):
        """
        Return a copy of this collider
        """
        return odeCylinder(datastore=self.datastore, position=self.position,
            rotation=self.rotation, radius=self.radius, length=self.length, 
            density=self.density,
            stationary=self.stationary, contacttype=self.contacttype,
            mu=self.mu, mu2=self.mu2, bounce=self.bounce,
            bounce_vel=self.bounce_vel, soft_erp=self.soft_erp,
            soft_cfm=self.soft_cfm, slip1=self.slip1, slip2=self.slip2,
            oid=self.oid, oncollide=self.oncollide, rollfrict=self.rollfrict)


    def build(self, body=None, space=None):
        """
        Build the cylinder
        """
        if self.stationary:
            body = None
        M = ode.Mass()
        M.setCappedCylinder(self.density, 1, self.radius, self.length)
        if body is not None:
            self.geom = ode.GeomCCylinder(radius=self.radius, 
                length=self.length)
            body.setMass(M + body.getMass())
        else:
            self.geom = ode.GeomCCylinder(space, radius=self.radius, 
                length=self.length)
        self.geom.setPosition(self.position)
        if len(self.rotation) != 4:
            self.geom.setRotation(self.rotation)
        else:
            self.geom.setQuaternion(self.rotation)
        self.setGeomParams(self.geom)
        if body is not None:
            self.tgeom = ode.GeomTransform(space)
            self.tgeom.setGeom(self.geom)
            self.tgeom.setBody(body)
            self.geom.body = body
            self.setGeomParams(self.tgeom)



class odePlane(ICollider, Copyable):
    """
    A box from the ode Physics engine
    """
    datastoreAttributes = attributes(Copyable,
        normal=simpleList, distance=number,
        contacttype=simpleList,
        mu=number, mu2=number, bounce=number, bounce_vel=number,
        soft_erp=number, soft_cfm=number, slip1=number,
        slip2=number, oid=number, rollfrict=number)
    datastoreIndexes = tuple()
    datastoreHash = ("oid", classHash, randomHash)
    datastoreVersion = 1
    datastore = None

    normal= Vector(0,1,0)
    distance = 1.0
    geom = None
    mu = 10000.0
    mu2 = 0.0
    bounce = 0.1
    bounce_vel = 0.0
    soft_erp = 0.0
    soft_cfm = 0.0
    slip1 = 0.0
    slip2 = 0.0
    oid = None
    oncollide = None
    contacttype = None
    rollfrict = 0.0


    def __init__(self, **kw):
        self.contacttype = ["ContactBounce"]
        for k, v in kw.iteritems():
            if not hasattr(self.__class__, k):
                raise ValueError("Inalid keyword argument %s to odePlane" % k)
        self.__dict__.update(kw)
        self.addExclude("geom", "tgeom", "oncollide")


    def copy(self):
        """
        Return a copy of this collider
        """
        return odePlane(datastore=self.datastore, 
            normal=self.normal, distance=self.distance,
            contacttype=self.contacttype,
            mu=self.mu, mu2=self.mu2, bounce=self.bounce,
            bounce_vel=self.bounce_vel, soft_erp=self.soft_erp,
            soft_cfm=self.soft_cfm, slip1=self.slip1, slip2=self.slip2,
            oid=self.oid, oncollide=self.oncollide, rollfrict=self.rollfrict)


    def build(self, body=None, space=None):
        """
        Build the plane
        """
        assert body is None
        self.geom = ode.GeomPlane(space, normal=self.normal, dist=self.distance)
        self.setGeomParams(self.geom)



class odeMesh(ICollider, Copyable):
    """
    A mesh from the ode Physics engine
    """
    datastoreAttributes = attributes(Copyable, tris=simpleList, 
        verts=simpleList,
        position=simpleList, rotation=simpleList, size=simpleList,
        density=number, stationary=boolean, contacttype=simpleList,
        mu=number, mu2=number, bounce=number, bounce_vel=number,
        soft_erp=number, soft_cfm=number, slip1=number,
        slip2=number, oid=number, rollfrict=number)
    datastoreIndexes = tuple()
    datastoreHash = ("oid", classHash, randomHash)
    datastoreVersion = 1
    datastore = None

    tris = None
    verts = None
    geom = None
    tgeom = None
    density = 1.0
    position = Vector(0,0,0)
    rotation = Quaternion(1,0,0,0)
    stationary = False
    contacttype = None
    mu = 10000.0
    mu2 = 0.0
    bounce = 0.1
    bounce_vel = 0.0
    soft_erp = 0.0
    soft_cfm = 0.0
    slip1 = 0.0
    slip2 = 0.0
    oid = None
    oncollide = None
    rollfrict = 0.0


    def __init__(self, **kw):
        self.contacttype = ["ContactBounce"]
        for k, v in kw.iteritems():
            if not hasattr(self.__class__, k):
                raise ValueError("Inalid keyword argument %s to odeMesh" % k)
        self.__dict__.update(kw)
        self.addExclude("geom", "tgeom", "oncollide")


    def copy(self):
        """
        Return a copy of this collider
        """
        return odeMesh(datastore=self.datastore, 
            tris=self.tris, verts=self.verts, density=self.density,
            contacttype=self.contacttype, stationary=self.stationary,
            mu=self.mu, mu2=self.mu2, bounce=self.bounce,
            bounce_vel=self.bounce_vel, soft_erp=self.soft_erp,
            soft_cfm=self.soft_cfm, slip1=self.slip1, slip2=self.slip2,
            oid=self.oid, oncollide=self.oncollide, rollfrict=self.rollfrict)


    def build(self, body=None, space=None):
        """
        Build the mesh
        """
        if self.stationary:
            body = None
        M = ode.Mass()
        M.setBox(self.density, 10, 10, 10)
        data = ode.TriMeshData()
        data.build(self.verts, self.tris)
        if body is not None:
            self.geom = ode.GeomTriMesh(data)
            body.setMass(M + body.getMass())
        else:
            self.geom = ode.GeomTriMesh(data, space)
        self.geom.setPosition(self.position)
        if len(self.rotation) != 4:
            self.geom.setRotation(self.rotation)
        else:
            self.geom.setQuaternion(self.rotation)
        self.setGeomParams(self.geom)
        if body is not None:
            self.tgeom = ode.GeomTransform(space)
            self.tgeom.setGeom(self.geom)
            self.tgeom.setBody(body)
            self.geom.body = body
            self.setGeomParams(self.tgeom)



class odeMeshGen(ICollider, Copyable):
    """
    A mesh from the ode Physics engine
    """
    datastoreAttributes = attributes(Copyable, meshGen=reference,
        lodfactor=number,
        position=simpleList, rotation=simpleList, 
        density=number, stationary=boolean, contacttype=simpleList,
        mu=number, mu2=number, bounce=number, bounce_vel=number,
        soft_erp=number, soft_cfm=number, slip1=number,
        slip2=number, oid=number, rollfrict=number)
    datastoreIndexes = tuple()
    datastoreHash = ("oid", classHash, randomHash)
    datastoreVersion = 1
    datastore = None

    meshGen = None
    lodfactor = 1.0
    geom = None
    tgeom = None
    density = 1.0
    position = Vector(0,0,0)
    rotation = Quaternion(1,0,0,0)
    stationary = False
    contacttype = None
    mu = 10000.0
    mu2 = 0.0
    bounce = 0.1
    bounce_vel = 0.0
    soft_erp = 0.0
    soft_cfm = 0.0
    slip1 = 0.0
    slip2 = 0.0
    oid = None
    oncollide = None
    rollfrict = 0.0


    def __init__(self, **kw):
        self.contacttype = ["ContactBounce"]
        for k, v in kw.iteritems():
            if not hasattr(self.__class__, k):
                raise ValueError("Inalid keyword argument %s to odeMeshGen" % k)
        self.__dict__.update(kw)
        self.addExclude("geom", "tgeom", "oncollide")


    def copy(self):
        """
        Return a copy of this collider
        """
        return odeMeshGen(datastore=self.datastore, meshGen=self.meshGen,
            lodfactor=self.lodfactor, density=self.density,
            contacttype=self.contacttype, stationary=self.stationary,
            mu=self.mu, mu2=self.mu2, bounce=self.bounce,
            bounce_vel=self.bounce_vel, soft_erp=self.soft_erp,
            soft_cfm=self.soft_cfm, slip1=self.slip1, slip2=self.slip2,
            oid=self.oid, oncollide=self.oncollide, rollfrict=self.rollfrict)


    def build(self, body=None, space=None):
        """
        Build this collider from a mesh generator
        """
        d = self.meshGen.genMesh(lodfactor=self.lodfactor)
        def gotMesh(mesh, body, space):
            if mesh is None:
                return
            if self.stationary:
                body = None
            M = ode.Mass()
            M.setBox(self.density, 10, 10, 10)
            data = ode.TriMeshData()
            data.build(mesh.verts, mesh.tris)
            if body is not None:
                self.geom = ode.GeomTriMesh(data)
                body.setMass(M + body.getMass())
            else:
                self.geom = ode.GeomTriMesh(data, space)
            self.geom.setPosition(self.position)
            if len(self.rotation) != 4:
                self.geom.setRotation(self.rotation)
            else:
                self.geom.setQuaternion(self.rotation)
            self.setGeomParams(self.geom)
            if body is not None:
                self.tgeom = ode.GeomTransform(space)
                self.tgeom.setGeom(self.geom)
                self.tgeom.setBody(body)
                self.geom.body = body
                self.setGeomParams(self.tgeom)
        d.addCallback(gotMesh, body, space)
        return d


class odeRay(ICollider, Copyable):
    """
    A ray from the ode Physics engine
    """
    datastoreAttributes = attributes(Copyable,
        position=simpleList, length=number, delta=simpleList,
        oid=number)
    datastoreIndexes = tuple()
    datastoreHash = ("oid", classHash, randomHash)
    datastoreVersion = 1
    datastore = None

    geom = None
    tgeom = None
    position = Vector(0,0,0)
    length = 1.0
    delta = Vector(0,0,0)
    oid = None
    oncollide = None


    def __init__(self, **kw):
        for k, v in kw.iteritems():
            if not hasattr(self.__class__, k):
                raise ValueError("Inalid keyword argument %s to odeRay" % k)
        self.__dict__.update(kw)
        self.addExclude("geom", "tgeom", "oncollide")


    def copy(self):
        """
        Return a copy of this collider
        """
        return odeRay(datastore=self.datastore, position=self.position,
            length=self.length, delta=self.delta, oid=self.oid,
            oncollide=self.oncollide)


    def build(self, body=None, space=None):
        """
        Create the ray
        """
        assert body is None
        assert self.delta != Vector(0,0,0)
        self.geom = ode.GeomRay(space, rlen=self.length)
        self.geom.set(self.position, self.delta)
        self.geom.oid = self.oid
        self.geom.oncollide = self.oncollide


class odeSpace(ICollider, Copyable):
    """
    A space from the ode Physics engine
    """
    datastoreAttributes = attributes(Copyable,
        type=text, center=simpleList, extents=simpleList,
        depth=number, oid=simpleList)
    datastoreIndexes = tuple()
    datastoreHash = ("oid", "type", classHash, randomHash)
    datastoreVersion = 1
    datastore = None

    geom = None
    type = u"simple"
    center = Vector(0,0,0)
    extents = Vector(100000,100000,100000)
    depth = 4
    oid = None


    def __init__(self, **kw):
        for k, v in kw.iteritems():
            if not hasattr(self.__class__, k):
                raise ValueError("Inalid keyword argument %s to odeSpace" % k)
        self.__dict__.update(kw)
        self.addExclude("geom")


    def copy(self):
        """
        Return a copy of this collider
        """
        return odeSpace(datastore=self.datastore, center=self.center,
            extents=self.extents, depth=self.depth,
            oid=self.oid, type=self.type)


    def build(self, body=None, space=None):
        """
        Create the space
        """
        assert body is None
        if self.type == "simple":
            self.geom = ode.SimpleSpace(space)
        elif self.type == "hash":
            self.geom = ode.HashSpace(space)
        elif self.type == "quadtree":
            self.geom = ode.QuadTreeSpace(self.center, self.extents,
                self.depth, space)  # ?? different arg order from docs
        self.geom.oid = self.oid



def deleteGeom(geom):
    """
    Really remove a geom 
    """
    if geom is None:
        return
    # first remove it from the space its in
    if geom.getSpace():
        geom.getSpace().remove(geom)
    if ode._geom_c2py_lut.has_key(geom._id()):
        del ode._geom_c2py_lut[geom._id()]
        return
    raise ColliderError("Couldn't delete geom!")



class RayCast:
    """
    Helper class for casting a ray
    """
    def __init__(self):
        self.mindepth = 9999999999
        self.mingeom = 0
        self.maxdepth = -999999999
        self.maxgeom = 0
        self.hitlist = []
        self.ignorelist = []
        self.hitonly = []


    def callback(self, r, geom1, geom2):
        """
        Called when there is a potential collision
        """
        if geom1 in self.ignorelist:
            return
        if geom2 in self.ignorelist:
            return
        if len(self.hitonly):
            if not geom2 in self.hitonly:
                return
            if not geom1 in self.hitonly:
                return
        contacts = ode.collide(geom1, geom2)
        for c in contacts:
            p, n, d, g1, g2 = c.getContactGeomParams()
            if d > self.maxdepth:
                self.maxdepth = d
                if g1._id() ==self.ray._id():
                    self.maxgeom = g2
                else:
                    self.maxgeom = g1
            if d < self.mindepth:
                self.mindepth = d
                self.minpos = p
                if g1._id() ==self.ray._id():
                    self.mingeom = g2
                else:
                    self.mingeom = g1
            if not g2 in self.hitlist:
                self.hitlist.append(g1)


def castRay(space, spos, epos, ignore=[], hitonly=[]):
    """
    Cast a ray in space from spos to epos. Any geoms
    in ignore will be ignored. If hitonly is specified result
    will only include geoms from that list. Returns an
    instance of the RayCast class
    """
    rc = RayCast()
    rc.ignorelist = ignore
    l = Vector(epos) - spos
    length = l.length()
    delta = l.normalize()
    myray = ode.GeomRay()
    myray.set(spos, delta)
    myray.setLength(length)
    rc.ray = myray
    if len(hitonly):
        hitonly.append(myray)
    ode.collide2(myray, space, 0, rc.callback)
    deleteGeom(myray)
    return rc



class CollisionMaker:
    """
    Retains contacts and properly handles collisions
    """
    
    def __init__(self, world):
        self.world = world
        self.toenable = []
        self.cglen = 0
        self.contactgroup = ode.JointGroup()
        self.exception = None
        
        
    def collideInternal(self, space):
        """
        Collide all geoms in space with each other
        """
        self.exception = None
        space.collide(0, self.nearCallback)
        if self.exception is not None:
            raise self.exception[0], self.exception[1], self.exception[2]
        
        
    def collide(self, geom1, geom2):
        """
        Collide two geoms (spaces or otherwise)
        """
        self.exception = None
        ode.collide2(geom1, geom2, 0, self.nearCallback)
        if self.exception is not None:
            raise self.exception[0], self.exception[1], self.exception[2]
            
            
    def clearCollisions(self):
        """
        Clear any contacts made by collisions
        """
        # if self.cglen:
            # print "contacts", self.cglen
        self.contactgroup.empty()
        self.cglen = 0


    def nearCallback(self, args, geom1, geom2):
        """
        This function is called when two geoms might
        collide.
        """
        try:
            # exceptions caught in C callbacks cause strange things
            if isinstance(geom1, ode.SpaceBase):
                if len(geom1) ==0:
                    return
                return ode.collide2(geom1, geom2, 0, self.nearCallback)
            if isinstance(geom2, ode.SpaceBase):
                return ode.collide2(geom2, geom1, 0, self.nearCallback)
            body1, body2 = geom1.getBody(), geom2.getBody()
            if (body1 == None or not body1.isEnabled()) and (body2 == None or not body2.isEnabled()):
                return
            if ode.areConnected(body1, body2):
                return
            contacts = ode.collide(geom1, geom2)
            if not len(contacts):
                return
            if geom1.oncollide:
                if geom1.oncollide(geom2, geom1):
                    if geom2.oncollide:
                        geom2.oncollide(geom1, geom2)
                    return
            if geom2.oncollide:
                if geom2.oncollide(geom1, geom2):
                    return
            ctype = geom1.contact_type|geom2.contact_type
            mu = geom1.contact_mu + geom2.contact_mu
            mu2 = geom1.contact_mu2 + geom2.contact_mu2
            bounce = (geom1.contact_bounce + geom2.contact_bounce) / 2.0
            bouncevel = (geom1.contact_bounce_vel + geom2.contact_bounce_vel) / 2.0
            softcfm = (geom1.contact_soft_cfm + geom2.contact_soft_cfm) / 2.0
            softerp = (geom1.contact_soft_erp + geom2.contact_soft_erp) / 2.0
            slip1 = (geom1.contact_slip1 + geom2.contact_slip1)
            slip2 = (geom1.contact_slip2 + geom2.contact_slip2)
            if geom1.oid and not geom1.oid in self.toenable:
                self.toenable.append(geom1.oid)
            if geom2.oid and not geom2.oid in self.toenable:
                self.toenable.append(geom2.oid)
            for c in contacts:
                c.setMode(ctype)
                c.setBounce(bounce)
                c.setMu2(mu2)
                c.setMu(mu)
                c.setBounceVel(bouncevel)
                c.setSoftCFM(softcfm)
                c.setSoftERP(softerp)
                c.setSlip1(slip1)
                c.setSlip2(slip2)
                # p, n, d, g1, g2 = c.getContactGeomParams()
                # print d
                ode.ContactJoint(self.world, self.contactgroup, c).attach(body1, body2)
                self.cglen += 1
        except:
            self.exception = sys.exc_info()


# Set the default types
Box = odeBox
Sphere = odeSphere
Cylinder = odeCylinder
Plane = odePlane
Mesh = odeMesh
Space = odeSpace
MeshGen = odeMeshGen
Ray = odeRay


def odeCollider(type="Box", **params):
    """
    Pretend we are the init for the old odeCollider class
    but instead return the appropriate collider
    """
    if type == "Box":
        return Box(**params)
    elif type == "Sphere":
        return Sphere(**params)
    elif type == "Cylinder":
        return Cylinder(**params)
    elif type == "Plane":
        return Plane(**params)
    elif type == "Mesh":
        return Mesh(**params)
    elif type == "Space":
        return Space(**params)
    elif type == "MeshGen":
        return MeshGen(**params)
    raise ColliderError("Unkown collider type: %s" % type)


pb.setUnjellyableForClass(odeBox, odeBox)
pb.setUnjellyableForClass(odeSphere, odeSphere)
pb.setUnjellyableForClass(odeCylinder, odeCylinder)
pb.setUnjellyableForClass(odePlane, odePlane)
pb.setUnjellyableForClass(odeMesh, odeMesh)
pb.setUnjellyableForClass(odeMeshGen, odeMeshGen)
pb.setUnjellyableForClass(odeSpace, odeSpace)
pb.setUnjellyableForClass(odeRay, odeRay)
