# Copyright (C) 2007-2008 Mortal Coil Games
# See LICENSE for details.
""" -*- test-case-name: mv3d.test.test_mesh -*-

"""

from time import time

from array import array

from twisted.internet import defer

from mv3d.util.datastore.store import (attributes, referenceList, setAttrs,
    simpleList, reference, classHash, number, boolean, randomHash)
from mv3d.net.pb import Cacheable, withClientUpdate
from mv3d.net.security import Securable
from mv3d.util.math3d import Vector, Quaternion
from mv3d.util.noise import PerlinNoise
from mv3d.phys.body import IBody, IVisualObject, IVisualObjectView
from mv3d.phys.collider import odeCollider
from mv3d.util.classgen import ClassGenerator


class MeshError(Exception):
    """
    An error that occured with the mesh
    """



def frange(start, stop, n):
    nm1 = n - 1
    nm1inv = 1.0 / nm1
    return [nm1inv * (start * (nm1 - i) + stop * i) for i in range(n)]



class IHeightGenerator:
    """
    This can be used to generate heights for 
    a height field
    """
    def generate(self, pos):
        """
        Generate a height at pos (x,z)
        Returns a float
        """



class NoiseHeightGenerator(IHeightGenerator, Cacheable, Securable):
    """
    Generate heights via Perlin Noise
    """
    datastoreAttributes = attributes(Cacheable, Securable,
        multiplier=simpleList, noise=reference, offset=simpleList)
    datastoreIndexes = tuple()
    datastoreHash = ("multiplier", "offset", classHash, randomHash)
    datastoreVersion = 1
    datastore = None

    getStateToCacheAndObserveFor = Cacheable.getStateToCacheAndObserveFor
    stoppedObserving = Cacheable.stoppedObserving
    setCopyableState = Cacheable.setCopyableState

    multiplier = (4.0, 100.0, 4.0)
    noise = None
    offset = (0,0,0)


    def __init__(self, seed=None, persistence=0.5, octaves=4, map=None,
            multiplier=(5.0, 100.0, 4.0), noise=None, offset=(0,0,0)):
        self.multiplier = multiplier
        self.offset = offset
        self.noise = noise or PerlinNoise(seed=seed or int(time()),
            persistence=persistence, octaves=octaves)
        self.noise.grantPermission("read", "all")
        self.noise.grantPermission("reference", "all")
        if map is not None:
            self.noise.map = map


    def generate(self, pos):
        """
        Generate a height using our noise generator
        """
        p = ((pos[0] + self.offset[0]) * self.multiplier[0], 
            (pos[1] + self.offset[2]) * self.multiplier[2])
        n = (self.noise.perlinNoise(p) + self.offset[1]) * self.multiplier[1]
        # print p, n
        return n
        # return self.noise.PerlinNoise(pos) * 100


    def getClassGenerators(self):
        return [ClassGenerator().setFrom(self), ClassGenerator().setFrom(self.noise)]



class IMeshGen:
    """
    Interface definition for mesh generators
    """
    def genMesh(self, lodfactor=1.0, justTris=True):
        """
        Generate the mesh. If justTris is False, 
        then normals and uv coords, and color
        map will be generated as well
        """



class MeshLodLevel:
    """
    Defines a mesh at a specific lod
    """
    def __init__(self, verts=None, colors=None, norms=None, uvs=None, 
            tris=None, trinorms=None, justTris=False):
        self.verts = verts
        self.tris = tris
        self.colors = colors
        self.norms = norms
        self.uvs = uvs
        self.trinorms = trinorms
        self.justTris = justTris



class HeightField(IMeshGen, Cacheable, Securable):
    """
    A heightfield mesh
    """
    datastoreAttributes = attributes(Cacheable, Securable,
        heightGenerator=reference, size=simpleList,
        resolution=simpleList, heights=simpleList)
    datastoreIndexes = tuple()
    datastoreHash = ("size", "resolution", classHash, randomHash)
    datastoreVersion = 1
    datastore = None

    stoppedObserving = Cacheable.stoppedObserving

    heightGenerator = None
    lodlevels = None
    size = (1.0, 1.0, 1.0)
    resolution = (10, 10, 10)
    heights = None

    def __init__(self, heightGenerator=None, size=None, resolution=None,
            heights=None):
        self.heightGenerator = heightGenerator
        self.size = size
        self.resolution = resolution
        self.heights = heights or array('f')
        self.lodlevels = dict()
        self.addExclude("lodlevels")


    def loadedFromStore(self):
        self.lodlevels = dict()


    @withClientUpdate
    def setHeights(self, resolution, heights):
        """
        Sets the heights of this HeightField
        """
        self.resolution = resolution
        self.heights = heights
        self.lodlevels = dict()
        self.updateClients("setHeights", resolution, heights.tostring())


    def observe_setHeights(self, resolution, heights):
        h = array('f')
        h.fromstring(heights)
        return self.setHeights(resolution, h)


    @withClientUpdate
    def setHeightGenerator(self, heightGenerator):
        """
        Set our height Generator
        """
        self.heightGenerator = heightGenerator


    @withClientUpdate
    def setResolution(self, resolution):
        """
        Set our resolution
        """
        self.resolution = resolution


    observe_setHeightGenerator = setHeightGenerator
    observe_setResolution = setResolution


    def calcIndex(self, point):
        """
        Calc the index in our height
        array of point (x,y)
        """
        return (int(point[0]) + int(point[1]) * self.resolution[0])


    def getHeight(self, point):
        """
        Get the height at this point
        """
        return self.heights[int(self.calcIndex(point))]


    def getSmoothedHeight(self, point, useGenerator=True):
        """
        Get the height at this point,
        but smooth it out with bilinear interpolation
        """
        if self.heightGenerator is not None and useGenerator:
            return self.heightGenerator.generate((point[0] / float(self.resolution[0]),
                point[1] / float(self.resolution[1])))
        tw = self.resolution[0] - 1
        th = self.resolution[1] - 1
        ww = point[0]
        hh = point[1]
        iw = int(ww)
        ih = int(hh)
        if iw == tw and ih == th:
            return self.getHeight(point)
        pw = ww - iw
        ph = hh - ih
        interp = lambda a, b, p: (a * (1.0 - p)) + (b * p)
        if ih >= th:
            if iw >= tw:
                return self.heights[tw + th * self.resolution[0]]
            return interp(self.heights[iw + th * self.resolution[0]], 
                self.heights[iw + 1 + th * self.resolution[0]], pw)
        if iw >= tw:
            return interp(self.heights[tw + ih * self.resolution[0]], 
                self.heights[tw + (ih + 1)  * self.resolution[0]], ph)
        h00 = self.heights[iw + ih * self.resolution[0]]
        h01 = self.heights[iw + (ih + 1) * self.resolution[0]]
        h10 = self.heights[iw + 1 + ih * self.resolution[0]]
        h11 = self.heights[iw + 1 + (ih + 1) * self.resolution[0]]
        ww1 = interp(h00, h01, ph)
        ww2 = interp(h10, h11, ph)
        return interp(ww1, ww2, pw)


    def genMesh(self, lodfactor=1.0, justTris=True):
        """
        This will generate the mesh. If justTris is False, then
        it will only generate the triangles.
        """
        if self.lodlevels.has_key(lodfactor) and self.lodlevels[lodfactor].justTris == justTris:
            return defer.succeed(self.lodlevels[lodfactor])
        if lodfactor != 1.0:
            return self.genMeshAtLOD(lodfactor, justTris)
        if len(self.heights) == 0 and self.heightGenerator is not None:
            self.generateHeights()
        verts = []
        if not justTris:
            colors = [(1, 1, 1)] * (self.resolution[0] * self.resolution[1])
            norms = [(0, 0, 0)] * (self.resolution[0] * self.resolution[1])
        else:
            colors = []
            norms = []
        uvs = []
        tris = []
        trinorms = []
        self.lodlevels[lodfactor] = MeshLodLevel(verts, colors, norms, uvs, tris, trinorms, justTris)
        vertc = 0
        for z in frange(0, self.size[1], self.resolution[1]):
            for x in frange(0, self.size[0], self.resolution[0]):
                y = self.heights[vertc]
                # print vertc
                verts.append(Vector(x, y, z))
                # print tuple(verts[-1])
                if not justTris:
                    uvs.append((x / float(self.size[0]), z / float(self.size[1])))
                vertc += 1
        for tri in range(self.resolution[0] * self.resolution[1] - (1 + self.resolution[1])):
            if divmod(tri, self.resolution[0])[1] == self.resolution[0] - 1:
                continue
            tris.append((tri, tri + self.resolution[0], tri + self.resolution[0] + 1))
            tris.append((tri + 1, tri, tri + 1 + self.resolution[0]))
            if not justTris:
                v3 = verts[tri + self.resolution[0] + 1]
                v2 = verts[tri + self.resolution[0]]
                v1 = verts[tri]
                s1 = v1 - v2
                s2 = v2 - v3
                v3 = verts[tri + self.resolution[0] + 1]
                v2 = verts[tri]
                v1 = verts[tri + 1]
                trinorms.append((s1 ** s2).normalize())
                norms[tri + self.resolution[0] + 1] = Vector(norms[tri + self.resolution[0] + 1]) + trinorms[-1]
                norms[tri + self.resolution[0]] = Vector(norms[tri + self.resolution[0]]) + trinorms[-1]
                norms[tri] = Vector(norms[tri]) + trinorms[-1]
                s3 = v2 - v1
                s4 = v3 - v2
                trinorms.append((s3 ** s4).normalize())
        # print tris
        return defer.succeed(self.lodlevels[lodfactor])


    def genMeshAtLOD(self, lodfactor=1.0, justTris=True):
        """
        This will generate the mesh. If justTris is False, then
        it will only generate the triangles.
        """
        if self.lodlevels.has_key(lodfactor) and self.lodlevels[lodfactor].justTris == justTris:
            return defer.succeed(self.lodlevels[lodfactor])
        if len(self.heights) == 0 and self.heightGenerator is not None:
            self.generateHeights()
        verts = []
        rez = (int(self.resolution[0] * lodfactor), int(self.resolution[1] * lodfactor))
        if not justTris:
            colors = [(1, 1, 1)] * (rez[0] * rez[1])
            norms = [(0, 0, 0)] * (rez[0] * rez[1])
        else:
            colors = []
            norms = []
        uvs = []
        tris = []
        trinorms = []
        self.lodlevels[lodfactor] = MeshLodLevel(verts, colors, norms, uvs, tris, trinorms, justTris)
        for z in frange(0, 1, rez[1]):
            for x in frange(0, 1, int(self.resolution[0] * lodfactor)):
                y = self.getSmoothedHeight((x * self.resolution[0], z * self.resolution[1])) 
                verts.append(Vector(x * self.size[0], y, z * self.size[1]))
                if not justTris:
                    # print "uvs"
                    uvs.append((x, z))
        for tri in range(rez[0] * rez[1] - (1 + rez[1])):
            if divmod(tri, rez[0])[1] == rez[0] - 1:
                continue
            tris.append((tri, 
                tri + rez[0], tri + rez[0] + 1))
            tris.append((tri + 1, tri, tri + 1 + rez[0]))
            if not justTris:
                v3 = verts[tri + rez[0] + 1]
                v2 = verts[tri + rez[0]]
                v1 = verts[tri]
                s1 = v1 - v2
                s2 = v2 - v3
                v3 = verts[tri + rez[0] + 1]
                v2 = verts[tri]
                v1 =verts[tri + 1]
                trinorms.append((s1 ** s2).normalize())
                norms[tri + rez[0]+ 1] = (
                    Vector(norms[tri + rez[0] + 1]) + trinorms[-1])
                norms[tri + rez[0]] = Vector(norms[tri + rez[0]]) + trinorms[-1]
                norms[tri] = Vector(norms[tri]) + trinorms[-1]
                s3 = v2 - v1
                s4 = v3 - v2
                trinorms.append((s3 ** s4).normalize())
        return defer.succeed(self.lodlevels[lodfactor])


    def generateHeights(self, heightGenerator=None, resolution=None):
        """
        Generate heights for this height field using a generator
        """
        if heightGenerator is not None:
            self.setHeightGenerator(heightGenerator)
        if resolution is not None:
            self.setResolution(resolution)
        self.heights = array('f')
        for z in frange(0, 1, self.resolution[1]):
            for x in frange(0, 1, self.resolution[0]):
                self.heights.append(self.heightGenerator.generate((x,z)))
        self.updateClients("setHeights", self.heights.tostring())
        return defer.succeed(True)


    def getClassGenerators(self):
        """
        Get all the class generators needed 
        for this meshGen
        """
        cgs = [ClassGenerator().setFrom(self)]
        if self.heightGenerator is not None:
            cgs.append(self.heightGenerator.getClassGenerators())
        return cgs


    def getStateToCacheAndObserveFor(self, p, o):
        state = Cacheable.getStateToCacheAndObserveFor(self, p, o)
        if self.heightGenerator is not None:
            state["heights"] = None
        else:
            state["heights"] = self.heights.tostring()
        return state


    def setCopyableState(self, state):
        hts = state["heights"]
        state["heights"] = array('f')
        if hts is not None:
            state["heights"].fromstring(hts)
        Cacheable.setCopyableState(self, state)



class MeshGenStaticBody(IBody, Cacheable, Securable):
    """
    A static body that uses a mesh generator
    """
    datastoreAttributes = attributes(IBody, Cacheable, Securable,
        special=boolean, areaid=simpleList, position=simpleList,
        rotation=simpleList, meshGen=reference, physlod=number,
        vobs=referenceList, objectid=simpleList)
    datastoreIndexes = tuple()
    datastoreHash = ("objectid", "position", classHash, randomHash)
    datastoreVersion = 1
    datastore = None

    getStateToCacheAndObserveFor = Cacheable.getStateToCacheAndObserveFor
    stoppedObserving = Cacheable.stoppedObserving
    setCopyableState = Cacheable.setCopyableState

    space = None
    special = False
    areaid = None
    transformGeom = None
    collider = None
    transformedcollider = None
    position = (0,0,0)
    rotation = (1,0,0,0)
    meshGen = None
    physlod = 1.0
    vobs = None
    objectid = None
    
    getStateToCacheAndObserveFor = Cacheable.getStateToCacheAndObserveFor
    setCopyableState = Cacheable.setCopyableState


    def __init__(self, **kw):
        setAttrs(self, **kw)
        self.vobs = []


    def create(self, world=None, meshGen=None, space=None):
        """
        Create the the colliders if space is 
        defined
        """
        if meshGen is not None:
            self.meshGen = meshGen
        if self.meshGen is not None:
            self.collider = odeCollider("MeshGen", meshGen=self.meshGen, lodfactor=self.physlod)
        if space is not None:
            self.enterSpace(space)


    def destroy(self):
        """
        Destroy the colliders
        """
        if self.collider is not None:
            self.collider.destroy()
        for v in self.vobs:
            v.destroy()


    @withClientUpdate
    def setMeshGen(self, meshGen):
        """
        Set the mesh generator
        """
        self.meshGen = meshGen


    def getMeshGen(self):
        """
        Returns the meshGen
        """
        return self.meshGen


    @withClientUpdate
    def setPhysicalLod(self, lodfactor):
        """
        Set the physical lodfactor
        """
        self.physlod = lodfactor


    observe_setMeshGen = setMeshGen
    observe_setPhysicalLod = setPhysicalLod


    def getPreferredSpaceType(self):
        """
        Return our preferred space type
            1 = Disabled
            2 = Special
        """
        if self.special:
            return 2
        return 1


    @withClientUpdate
    def setPosition(self, pos):
        """
        Set the position of this body
        """
        self.position = pos
        self.updateVobs()
        self.enterSpace(self.space)


    def getPosition(self):
        """
        Get the position of this body
        """
        return self.position


    @withClientUpdate
    def setRotation(self, rot):
        """
        Set the quaternion rotation of this body
        """
        self.rotation = rot
        self.updateVobs()
        self.enterSpace(self.space)


    def getRotation(self):
        """
        Get the quaternion rotation of this body
        """
        return self.rotation


    def enable(self):
        """
        Can't be enabled
        """


    def disable(self):
        """
        Can't be disabled
        """


    def isEnabled(self):
        """
        We can't be enabled
        """
        return False


    def enterSpace(self, space):
        """
        Create colliders in this space
        """
        # print "meshgen enter space", space
        # print self.__dict__
        if self.collider is None and self.meshGen is not None:
            self.collider = odeCollider("MeshGen", meshGen=self.meshGen, 
                lodfactor=self.physlod)
        if self.transformedcollider is not None:
            self.transformedcollider.destroy()
        self.transformedcollider = self.collider.copy()
        self.transformedcollider.setPosition(Vector(
            self.collider.position) + self.position)
        self.transformedcollider.setRotation(Quaternion(
            self.collider.rotation) * self.rotation)
        # print "Building Collider", self
        self.transformedcollider.build(space=space)
        # print "built"
        self.space = space


    def enterArea(self, area):
        """
        Determine the space most appropriate
        in are and enter it. Area should be an
        area object
        """
        if self.special:
            space = area.getSpecialSpace(self.getPosition())
        else:
            space = area.getDisabledSpace(self.getPosition())
        self.enterSpace(space)
        self.areaid = area.getID()


    def leaveArea(self, area):
        """
        Just destroy ourselves
        """
        return self.destroy()


    @withClientUpdate
    def addVisualObject(self, vob):
        """
        Add a visual object
        """
        self.vobs.append(vob)
    

    @withClientUpdate
    def removeVisualObject(self, vob):
        """
        Remove a visual object
        """
        self.vobs.remove(vob)


    observe_removeVisualObject = removeVisualObject
    observe_addVisualObject = addVisualObject


    def getVisualObjects(self):
        """
        Get all the visual objects
        """
        return self.vobs


    def updateVobs(self):
        """
        Update all the visual objects
        """
        p = self.getPosition()
        r = self.getRotation()
        for vob in self.vobs:
            if isinstance(vob, IVisualObjectView):
                vob.setPosition(p)
                vob.setRotation(r)


    def getServerClassGenerators(self):
        """
        Return class generators
        """
        cgs = []
        cgs.append(ClassGenerator().setFrom(self))
        cgs += self.meshGen.getClassGenerators()
        for v in self.vobs:
            cgs += v.getServerClassGenerators()
        return cgs


    def getClientClassGenerators(self):
        """
        Return class generators
        """
        cgs = []
        cgs.append(ClassGenerator().setFrom(self))
        cgs += self.meshGen.getClassGenerators()
        for v in self.vobs:
            cgs += v.getClientClassGenerators()
        return cgs



class MeshGenVisualObject(IVisualObject, Cacheable, Securable):
    """
    A Visual Object that uses a mesh generator
    """
    datastoreAttributes = attributes(Cacheable, Securable,
        areaid=simpleList, meshGen=reference, materialAssetId=simpleList)
    datastoreIndexes = tuple()
    datastoreHash = ("areaid", classHash, randomHash)
    datastoreVersion = 1
    datastore = None

    stoppedObserving = Cacheable.stoppedObserving
    getStateToCacheAndObserveFor = Cacheable.getStateToCacheAndObserveFor
    setCopyableState = Cacheable.setCopyableState

    areaid = None
    meshGen = None
    materialAssetId = None

    __init__ = setAttrs


    def enterArea(self, area):
        """
        Enter an area
        """
        self.areaid = area.getID()


    def getMeshGen(self):
        return self.meshGen


    @withClientUpdate
    def setMeshGen(self, meshGen):
        """
        Set the mesh generator
        """
        self.meshGen = meshGen


    def getMaterialAsset(self):
        """
        Returns the material asset  id
        """
        return self.materialAssetId


    @withClientUpdate
    def setMaterialAsset(self, aid):
        """
        Set the material asset for the meshgen
        """
        self.materialAssetId = aid


    observe_setMeshGen = setMeshGen
    observe_setMaterialAsset = setMaterialAsset


    def getServerClassGenerators(self):
        """
        Returns class generators for a server
        """
        return [ClassGenerator(classname="MeshGenVisualObject", 
            modulename="mv3d.phys.mesh", 
            sourceclass="mv3d.phys.mesh.MeshGenVisualObject")]


    def getClientClassGenerators(self):
        """
        Returns class generators for a client
        """
        return [ClassGenerator(classname="MeshGenVisualObjectView", 
            modulename="mv3d.client.view.visual", 
            sourceclass="mv3d.phys.mesh.MeshGenVisualObject")]
