# Copyright (C) 2008 Mortal Coil Games
# See LICENSE for details.
#
# -*- test-case-name: mv3d.test.test_scenery -*-
"""
Scenery is basically something that can not move, but has a
physical and visual representation.

The top level item is the Scenery body, which contains
multiple SceneryPieces. Each piece can have multiple colliders
and visual objects allowing for a single scenery piece to be
composited from several pieces.

On the client side, the SceneryClient will have the paged geometry
that the pieces will attach to.
"""

from twisted.internet.task import coiterate

from mv3d.util.datastore.store import (attributes, 
    Reference, simpleList, classHash, number, randomHash)
from mv3d.util.classgen import ClassGenerator
from mv3d.net.pb import Cacheable, Copyable, observable
from mv3d.net.security import Securable
from mv3d.phys.body import IBody, IVisualObject


class SceneryPiece(Copyable, Securable):
    """
    An individual scenery piece
    """
    def __init__(self):
        self.colliders = []
        self.visualAssets = []


    def addVisualAsset(self, aid, position, yRot=0.0, scale=1.0):
        """
        Add a new visual asset
        """
        self.visualAssets.append(dict(asset=aid, position=position,
            yRot=yRot, scale=scale))



class PieceIterator:
    """
    Iterates through pieces and calls a func on them
    """
    def __init__(self, pieces, func, *a, **kw):
        self.func = (func, a, kw)
        self.pieces = pieces.__iter__()


    def __iter__(self):
        return self


    def next(self):
        """
        Implemnent the iter interface.
        Just get the next piece and call the func on it
        """
        p = self.pieces.next()
        return self.func[0](p, *self.func[1], **self.func[2])



class Scenery(IBody, Cacheable, Securable):
    """
    Holds scenery
    """
    datastoreAttributes = attributes(IBody, Cacheable, Securable,
        pieces=simpleList, pageSize=number, batchPageLod=simpleList,
        imposterPageLod=simpleList)
    datastoreIndexes = tuple()
    datastoreHash = (classHash, randomHash)
    datastoreVersion = 1

    area = None
    space = None
    pageSize = 50
    batchPageLod = (100, 30)
    imposterPageLod = (4000, 50)

    def __init__(self):
        self.pieces = []
        self.grantPermission("read", "all")
        self.grantPermission("reference", "all")
        self.addExclude("area", "space")


    def create(self, world):
        """
        We don't need to do anything
        """


    def addPiece(self, piece):
        """
        Add a scenery piece
        """
        self.pieces.append(piece)


    addPiece, observe_addPiece = observable(addPiece)


    def enterSpace(self, space):
        """
        Build the scenery in a space
        """
        self.space = space
        return coiterate(
            PieceIterator(self.pieces, self.buildPiece))


    def destroy(self):
        """
        Destroy the scenery
        """
        self.space = None
        return coiterate(
            PieceIterator(self.pieces, self.destroyPiece))


    def enterArea(self, area):
        """
        Enter an area
        """
        self.area = area
        return self.enterSpace(area.getDisabledSpace())


    def buildPiece(self, piece):
        """
        Build a piece in our area
        """
        for c in piece.colliders:
            c.build(space=self.space)


    def destroyPiece(self, piece):
        """
        Destroy one piece
        """
        for c in piece.colliders:
            c.destroy()


    def isEnabled(self):
        """
        Always disabled
        """
        return False


    def getPreferredSpaceType(self):
        """
        Always disabled
        """
        return 1


    def getPosition(self):
        """
        Return something... 
        """
        return (0, 0, 0)


    def getServerClassGenerators(self):
        """
        Return class generators 
        """
        cgs = []
        cgs.append(ClassGenerator().setFrom(self))
        cgs.append(ClassGenerator().setFrom(SceneryPiece))
        return cgs


    def getClientClassGenerators(self):
        """
        Return class generators
        """
        cgs = []
        cgs.append(ClassGenerator(modulename="mv3d.client.view.visual",
            classname="SceneryView", sourceclass="mv3d.phys.scenery.Scenery"))
        cgs.append(ClassGenerator().setFrom(SceneryPiece))
        return cgs



class Grass(IVisualObject, Cacheable, Securable):
    """
    Defines a patch of grass (or flowers)
    """
    datastoreAttributes = attributes(Cacheable, Securable,
        bounds=simpleList, layers=simpleList, #heightField=Reference, 
        pageSize=number, maxDist=number)
    datastoreIndexes = tuple()
    datastoreHash = ("bounds", classHash, randomHash)
    datastoreVersion = 1

    stoppedObserving = Cacheable.stoppedObserving
    getStateToCacheAndObserveFor = Cacheable.getStateToCacheAndObserveFor
    setCopyableState = Cacheable.setCopyableState


    def __init__(self, bounds, heightField, pageSize=100,
            maxDist=100):
        self.layers = []
        self.bounds = bounds
        self.heightField = heightField
        self.pageSize = pageSize
        self.maxDist = maxDist


    def addLayer(self, aid, **kw):
        """
        Add a layer of grass
        """
        allowed = ["density", "bounds", "animated", "swaySpeed",
            "swayLength", "swayDistrib", "colorMapAssetId",
            "densityMapAssetId", "minSize", "maxSize", "heightRange"]
        for k in kw.keys():
            if k not in allowed:
                raise TypeError("""Unexpected keyword argument %s
                    to addLayer""" % k)
        kk = kw.copy()
        kk["aid"] = aid
        self.layers.append(kk)
    addLayer, observe_addLayer = observable(addLayer)


    def getServerClassGenerators(self):
        """
        Return class generators 
        """
        cgs = self.heightField.getClassGenerators()
        cgs.append(ClassGenerator().setFrom(self))
        return cgs


    def getClientClassGenerators(self):
        """
        Return class generators
        """
        cgs = self.heightField.getClassGenerators()
        cgs.append(ClassGenerator(modulename="mv3d.client.view.visual",
            classname="GrassView", sourceclass="mv3d.phys.scenery.Grass"))
        return cgs
