# Copyright (C) 2006-2008 Mortal Coil Games
# See LICENSE for details.
""" -*- test-case-name: mv3d.test.test_assetgroup -*-
Classes for Assets and AssetGroups
    An Asset can represent any static content available in the game. Some
    examples are Images, 3D Objects, Sounds, and Animations.
    The general idea is that AssetGroups will only live on Asset Servers.
    They will be referenced down to other servers when needed (or
    cached to slave servers). The individual
    Assets will be copied down to the end user as they will include a way
    to acquire the asset (via downloading, hard drive, osmosis, or whatnot)
"""


from twisted.internet import defer

from mv3d.util.datastore.store import (attributes, text, number, simpleList, 
    autoStore)
from mv3d.util.classgen import ClassGenerator
from mv3d.net.security import Securable
from mv3d.net.pb import Cacheable, withClientUpdate
from mv3d.util.date import Date


class Asset(Securable, Cacheable):
    """
    Assets will have 2 number IDs:
        id[0] = AssetGroup id
        id[1] = AssetId (within the group)
        Generally, users of this class will subclass it
        and then override the AcquireAsset function.
    """
    datastoreAttributes = attributes(Securable, Cacheable, aid=number,
        name=text, author=text, copyright=text, license=text, description=text,
        revisor=text, comment=text, checksum=text, dependencies=simpleList)
    
    aid = None
    asset = None
    name = u""
    author = u""
    copyright = u""
    license = u""
    description = u""
    revisor = u""
    comment = u""
    checksum = None
    parent = None


    def __init__(self):
        """
        Initialize to default values
            The asset is set to 0 and added to the exclude list so it
            isn't sent over the network.
        """
        self.date = Date()
        self.dependencies = []
        self.addExclude("asset")


    def getID(self):
        """
        return the full id for this asset 
        """
        return self.aid


    @autoStore
    @withClientUpdate
    def setID(self, i):
        """
        set the full id for this asset 
        """
        self.aid = i


    observe_setID = setID


    def getName(self):
        """
        Returns my name 
        """
        return self.name


    @autoStore
    @withClientUpdate
    def setName(self, n):
        """
        Set my name to n. Names are non unique 
        """
        self.name = n


    observe_setName = setName


    def getAuthor(self):
        """
        returns the author (str) 
        """
        return self.author


    @autoStore
    @withClientUpdate
    def setAuthor(self, a):
        """
        sets the author to a (str) 
        """
        self.author = a


    observe_setAuthor = setAuthor


    def getCopyright(self):
        """
        Returns the copyright notice (str) 
        """
        return self.copyright


    @autoStore
    @withClientUpdate
    def setCopyright(self, c):
        """
        sets the copyright to c (str) 
        """
        self.copyright = c


    observe_setCopyright = setCopyright


    def getLicense(self):
        """
        Returns the license information (str) 
        """
        return self.license


    @autoStore
    @withClientUpdate
    def setLicense(self, l):
        """
        sets the license to l (str) 
        """
        self.license = l


    observe_setLicense = setLicense


    def getDescription(self):
        """
        Returns the description (str) 
        """
        return self.description


    @autoStore
    @withClientUpdate
    def setDescription(self, d):
        """
        sets the description to d (str) 
        """
        self.description = d


    observe_setDescription = setDescription


    def getRevisor(self):
        """
        returns the revisor (str) 
        """
        return self.revisor


    @autoStore
    @withClientUpdate
    def setRevisor(self, r):
        """
        sets the revisor to r (str) 
        """
        self.revisor = r


    observe_setRevisor = setRevisor


    def getDate(self):
        """
        returns the date (Date) 
        """
        return self.date


    @autoStore
    @withClientUpdate
    def setDate(self, d):
        """
        sets the date to d (anything parsable by Date) 
        """
        self.date = Date().parse(d)


    observe_setDate = setDate


    def getComment(self):
        """
        returns the comment associated with this revision (str) 
        """
        return self.comment


    @autoStore
    @withClientUpdate
    def setComment(self, c):
        """
        Sets the comment associated with this revision to c (str) 
        """
        self.comment = c


    observe_setComment = setComment


    def haveAsset(self):
        """
        Returns true if was have downloaded or otherwise acquired
            an asset
        """
        return self.asset != 0


    def acquireAsset(self, **kw):
        """
        This function will get the asset loaded by whatever means
            is appropriate.
            kw would be things like lod=10, bitrate="44k", etc.
            Normally, it'll return a deferred which will callback
            when the asset is finished downloading. If there is an instant
            error, it will return 0 right away. If the asset is already
            available, it'll return 1 right away.
        """
        # this Asset class doesn't actually get any assets so
        # just return 1 if the asset is present
        return self.haveAsset()


    def getAsset(self):
        """
        This function will return the reference to the asset if it
            has been loaded. If not, it returns 0.
        """
        if self.haveAsset():
            return self.asset
        return 0


    def listDependencies(self):
        """
        Return a list of the aids that we depend on
        """
        return self.dependencies


    def acquireDependencies(self, **kw):
        """
        This function will make sure that all the
        dependencies of this asset are loaded...
        It probably returns a deferred
        """
        dd = []
        i = self.parent
        for a in self.dependencies:
            d = i.acquireAsset(a, **kw)
            assert d != 0
            if isinstance(d, defer.Deferred):
                dd.append(d)
        if len(dd):
            dl = defer.DeferredList(dd)
            return dl
        return 1


    @autoStore
    @withClientUpdate
    def addDependency(self, d):
        """
        Add the aid of an asset that needs to be loaded
        before we can be loaded.
        """
        self.dependencies.append(d)


    @autoStore
    @withClientUpdate
    def remDependency(self, d):
        """
        Remove d from the list of asset ids that we
        depend on
        return 0 if d isn't a current dependency.
        """
        self.dependencies.remove(d)
        return True


    observe_addDependency = addDependency
    observe_remDependency = remDependency


    def getChecksum(self):
        """
        Get the checksum for this asset 
        """
        return self.checksum


    @autoStore
    @withClientUpdate
    def setChecksum(self, c):
        """
        Get the checksum for this asset 
        """
        self.checksum = c


    observe_setChecksum = setChecksum


    def refreshChecksum(self):
        """
        Causes the checksum to be reloaded for this asset 
        """
        pass


    def getClassGenerator(self):
        """
        Returns a class generator for this asset class 
        """
        return ClassGenerator(classname = self.__class__.__name__, modulename = self.__class__.__module__)


    def getStateToCacheAndObserveFor(self, perspective, observer):
        """
        You need to add a docstring here ! 
        """
        return Cacheable.getStateToCacheAndObserveFor(self, perspective, observer)


    def setCopyableState(self, state):
        """
        You need to add a docstring here ! 
        """
        return Cacheable.setCopyableState(self, state)


# define some basic types of assets

class ImageAsset(Asset):
    """
    You need to add a docstring here ! 
    """
    pass



class MaterialAsset(Asset):
    """
    You need to add a docstring here ! 
    """
    pass



class MeshAsset(Asset):
    """
    You need to add a docstring here ! 
    """
    pass



class PhysicsAsset(Asset):
    """
    You need to add a docstring here ! 
    """
    pass



class CodeAsset(Asset):
    """
    You need to add a docstring here ! 
    """
    pass



class SoundAsset(Asset):
    """
    You need to add a docstring here ! 
    """
    pass



class FileAsset(Asset):
    """
    You need to add a docstring here ! 
    """
    pass



# pb.setUnjellyableForClass(AssetGroup, AssetGroup)
# pb.setUnjellyableForClass(Asset, Asset)
