# -*- test-case-name: mv3d.test.test_asset -*-
# Copyright (C) 2006-2008 Mortal Coil Games
# See LICENSE for details.
"""
A class for an Asset that can be downloaded by a URL.
    The URL is in standard format, and can be one of the following types:
        http://
        https://
        ftp://
        file://
        gopher:// (I thought this was dead, but we do support it)
"""


import md5
import urllib
import os
import sys
import logging

from twisted.internet import threads
from twisted.internet import defer, reactor, ssl
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.web.client import downloadPage
from twisted.web import error

from mv3d.util.datastore.store import attributes, text, classHash, autoStore
from mv3d.net.pb import withClientUpdate
from mv3d.net.pb import Cacheable
from mv3d.util.classgen import ClassGenerator
from mv3d.resource.asset import ImageAsset, MaterialAsset, MeshAsset, PhysicsAsset
from mv3d.resource.asset import CodeAsset, SoundAsset, FileAsset


class URLAssetError(Exception):
    """
    Raised when there is a problem with a URL Asset
    """


class URLAsset:
    """
    An asset that can be downloaded via a URL
    Files are saved in basedir/GroupXX/localfile  (XX = Asset Group number)
    """
    datastoreAttributes = attributes(basedir=text, url=text, localfile=text)


    basedir = u"Extern"


    def getURL(self):
        """
        Get the URL this asset can be downloaded from 
        """
        return self.url


    @autoStore
    @withClientUpdate
    def setURL(self, u):
        """
        Set the URL this Asset can be downloaded from 
        """
        self.url = u


    def getLocalFile(self):
        """
        Gets the local file name for this asset 
        """
        return os.path.join(*self.localfile)


    @autoStore
    @withClientUpdate
    def setLocalFile(self, f):
        """
        Sets the local filename for this asset 
        """
        self.localfile = f


    def getBaseDir(self):
        """
        return the current base dir 
        """
        return self.basedir


    @autoStore
    @withClientUpdate
    def setBaseDir(self, b):
        """
        Set the base directory for this asset 
        """
        self.basedir = b


    observe_setURL = setURL
    observe_setLocalFile = setLocalFile
    observe_setBaseDir = setBaseDir


    def getFullPath(self):
        """
        Returns the full path to where our file is located 
        """
        return os.path.join(self.basedir, "Group%d" % self.getID()[0],
            *self.localfile[:-1])


    def checkPath(self):
        """
        Checks to make sure that the path to our file exists.
        if not, create it.
        """
        p = self.getFullPath()
        if os.path.exists(p):
            return True
        os.makedirs(p)
        return True


    def getFullFile(self):
        """
        Returns full path + local file name 
        """
        return os.path.join(self.basedir, "Group%d" % self.getID()[0],
            *self.localfile)


    def getFile(self):
        """
        Returns just the file name
        """
        return self.localfile[-1]


    @inlineCallbacks
    def urlAcquireAsset(self, **kw):
        """
        Download this asset using urllib 
        """
        yield self.acquireDependencies(**kw)
        if self.urlHaveAsset():
            returnValue(True)
        self.parent.parent.log("Downloading %s from %s." % (
            self.getFile(), self.url),
            logLevel=logging.INFO, system=self.parent.name)
        if not self.checkPath():
            raise URLAssetError("Check path failed")
        try:
            yield downloadPage(str(self.url), self.getFullFile(),
                agent="MV3D Asset Downloader")
        except error.Error, e:
            raise URLAssetError(str(e))
        self.parent.parent.log("Downloaded %s." % (
            self.getFile()),
            logLevel=logging.INFO, system=self.parent.name)
        returnValue(True)


    def urlHaveAsset(self):
        """
        This version of haveAsset checks if the file
        exists and then generates the md5 sum of it to check
        the CRC
        """
        try:
            # print "Haveasset start"
            #print "File", self.getID(), self.getFullFile()
            if not os.path.exists(self.getFullFile()):
                return False
            if self.checksum is None:
                # no checksum, so we have the asset I guess.
                # print "do have asset", repr(self.getFullFile()), "isfile", os.path.isfile(self.getFullFile()), "isdir", os.path.isdir(self.getFullFile())
                return True
            #print "continuting", self.checksum
            f = open(self.getFullFile(), "rb")
            s = f.read()
            f.close()
            m = md5.new()
            m.update(s)
            if m.hexdigest() != self.checksum:
                return False
            #print "Have asset full"
            return True
        except:
            import traceback
            traceback.print_exc()
            return False



class URLImageAsset(ImageAsset, URLAsset):
    """
    You need to add a docstring here ! 
    """
    datastoreAttributes = attributes(ImageAsset, URLAsset)
    datastoreHash = ("aid", classHash)
    datastoreVersion = 1


    def __init__(self, datastore=None, url=None, localfile=None, server=None, 
            basedir=u"Extern"):
        ImageAsset.__init__(self)
        self.datastore = datastore
        self.url = url
        self.localfile = localfile
        self.basedir = basedir
        self.parent = server


    def haveAsset(self):
        """
        You need to add a docstring here ! 
        """
        # print "haveAsset", self.name
        return URLAsset.urlHaveAsset(self)


    def acquireAsset(self, **kw):
        """
        You need to add a docstring here ! 
        """
        d = URLAsset.urlAcquireAsset(self, **kw)
        def register(r):
            """
            You need to add a docstring here ! 
            """
            # print "REGISTERING!", self.name
            from ogre.renderer import OGRE
            OGRE.ResourceGroupManager.getSingleton().addResourceLocation(os.path.split(self.getFullFile())[0], "FileSystem")
            return r


        if isinstance(d, defer.Deferred):
            d.addCallback(register)
            return d
        return register(d)


    def GetWebForm(self, url, args):
        """
        You need to add a docstring here ! 
        """
        return ImageAsset.GetWebForm(self, url, args) + \
            URLAsset.urlGetWebForm(self, url, args)



class URLMaterialAsset(MaterialAsset, URLAsset):
    """
    You need to add a docstring here ! 
    """
    datastoreAttributes = attributes(MaterialAsset, URLAsset)
    datastoreHash = ("aid", classHash)
    datastoreVersion = 1


    def __init__(self, datastore=None, url=None, localfile=None, server=None, 
            basedir=u"Extern"):
        MaterialAsset.__init__(self)
        self.datastore = datastore
        self.url = url
        self.localfile = localfile
        self.basedir = basedir
        self.parent = server


    def haveAsset(self):
        """
        You need to add a docstring here ! 
        """
        return URLAsset.urlHaveAsset(self)


    def acquireAsset(self):
        """
        You need to add a docstring here ! 
        """
        return URLAsset.urlAcquireAsset(self)



class URLMeshAsset(MeshAsset, URLAsset):
    """
    You need to add a docstring here ! 
    """
    datastoreAttributes = attributes(MeshAsset, URLAsset)
    datastoreHash = ("aid", classHash)
    datastoreVersion = 1


    def __init__(self, datastore=None, url=None, localfile=None, server=None, 
            basedir=u"Extern"):
        MeshAsset.__init__(self)
        self.datastore = datastore
        self.url = url
        self.localfile = localfile
        self.basedir = basedir
        self.parent = server


    def haveAsset(self):
        """
        You need to add a docstring here ! 
        """
        return URLAsset.urlHaveAsset(self)


    def acquireAsset(self):
        """
        You need to add a docstring here ! 
        """
        return URLAsset.urlAcquireAsset(self)



class URLPhysicsAsset(PhysicsAsset, URLAsset):
    """
    You need to add a docstring here ! 
    """
    datastoreAttributes = attributes(PhysicsAsset, URLAsset)
    datastoreHash = ("aid", classHash)
    datastoreVersion = 1


    def __init__(self, datastore=None, url=None, localfile=None, server=None, 
            basedir=u"Extern"):
        PhysicsAsset.__init__(self)
        self.datastore = datastore
        self.url = url
        self.localfile = localfile
        self.basedir = basedir
        self.parent = server 


    def haveAsset(self):
        """
        You need to add a docstring here ! 
        """
        return URLAsset.urlHaveAsset(self)


    def acquireAsset(self):
        """
        You need to add a docstring here ! 
        """
        return URLAsset.urlAcquireAsset(self)



class URLCodeAsset(CodeAsset, URLAsset):
    """
    You need to add a docstring here ! 
    """
    datastoreAttributes = attributes(CodeAsset, URLAsset)
    datastoreHash = ("aid", classHash)
    datastoreVersion = 1


    def __init__(self, datastore=None, url=None, localfile=None, server=None, 
            basedir=u"Extern"):
        CodeAsset.__init__(self)
        self.datastore = datastore
        self.url = url
        self.localfile = localfile
        self.basedir = basedir
        self.parent = server


    def haveAsset(self):
        """
        You need to add a docstring here ! 
        """
        return URLAsset.urlHaveAsset(self)


    def acquireAsset(self):
        """
        You need to add a docstring here ! 
        """
        return URLAsset.urlAcquireAsset(self)



class URLSoundAsset(SoundAsset, URLAsset):
    """
    You need to add a docstring here ! 
    """
    datastoreAttributes = attributes(SoundAsset, URLAsset)
    datastoreHash = ("aid", classHash)
    datastoreVersion = 1


    def __init__(self, datastore=None, url=None, localfile=None, server=None, 
            basedir=u"Extern"):
        SoundAsset.__init__(self)
        self.datastore = datastore
        self.url = url
        self.localfile = localfile
        self.basedir = basedir
        self.parent = server


    def haveAsset(self):
        """
        You need to add a docstring here ! 
        """
        return URLAsset.urlHaveAsset(self)


    def acquireAsset(self):
        """
        You need to add a docstring here ! 
        """
        return URLAsset.urlAcquireAsset(self)



class URLFileAsset(FileAsset, URLAsset):
    """
    You need to add a docstring here ! 
    """
    datastoreAttributes = attributes(FileAsset, URLAsset)
    datastoreHash = ("aid", classHash)
    datastoreVersion = 1


    def __init__(self, datastore=None, url=None, localfile=None, server=None, 
            basedir=u"Extern"):
        FileAsset.__init__(self)
        self.datastore = datastore
        self.url = url
        self.localfile = localfile
        self.basedir = basedir
        self.parent = server


    def haveAsset(self):
        """
        You need to add a docstring here ! 
        """
        return URLAsset.urlHaveAsset(self)


    def acquireAsset(self):
        """
        You need to add a docstring here ! 
        """
        return URLAsset.urlAcquireAsset(self)



def newURLFileAsset(assetgroup, url, localfile=u"", name=u""):
    """
    You need to add a docstring here ! 
    """
    d = assetgroup.newAsset(ClassGenerator(classname = u"URLFileAsset", 
        modulename = u"mv3d.resource.url"))


    def finishUp(tm, url, localfile, name):
        """
        You need to add a docstring here ! 
        """
        tm.setURL(url)
        if localfile != "":
            tm.setLocalFile(localfile)
        if name != "":
            tm.setName(name)
        return tm
    d.addCallback(finishUp, url, localfile, name)
    return d


def newURLImageAsset(assetgroup, url, localfile=u"", name=u""):
    """
    You need to add a docstring here ! 
    """
    d = assetgroup.newAsset(ClassGenerator(classname = u"URLImageAsset", 
        modulename = u"mv3d.resource.url"))


    def finishUp(tm, url, localfile, name):
        """
        You need to add a docstring here ! 
        """
        tm.setURL(url)
        if localfile != "":
            tm.setLocalFile(localfile)
        if name != "":
            tm.setName(name)
        return tm
    d.addCallback(finishUp, url, localfile, name)
    return d



def newURLClassGeneratorAsset(assetgroup, url, classn, localfile="", name="", depends=[]):
    """
    You need to add a docstring here ! 
    """
    tm = assetgroup.newAsset(ClassGenerator(classname = "URLClassGeneratorAsset", 
        modulename = "mv3d.resource.url"))
    def finishUp(tm, url, classn, localfile, name, depends):
        """
        You need to add a docstring here ! 
        """
        tm.setURL(url)
        tm.classname = classn
        if localfile != "":
            tm.setLocalFile(localfile)
        if name != "":
            tm.setName(name)
        for d in depends:
            if isinstance(d, tuple):
                tm.addDependency(d)
            else:
                tm.addDependency(d.getID())
        return tm
    if isinstance(tm, defer.Deferred):
        tm.addCallback(finishUp, url, classn, localfile, name, depends)
        return tm
    return finishUp(tm, url, classn, localfile, name, depends)



class URLClassGeneratorAsset(URLCodeAsset):
    """
    A class generator asset where the code
    can be downloaded via url
    """
    datastoreAttributes = attributes(URLCodeAsset, classname=text)
    datastoreHash = ("aid", classHash)
    datastoreVersion = 1


    def __init__(self, datastore=None, server=None, url=None, localfile=None, 
            basedir=u"Extern", classname=None):
        URLCodeAsset.__init__(self, datastore, url, localfile, basedir)
        self.classname = classname
        self.server = server


    def getModule(self):
        """
        You need to add a docstring here ! 
        """
        f = self.getLocalFile()[:-3]
        f = "Group%d" % self.getID()[0] + "." + f.replace(os.path.sep, ".")
        return f


    def getClass(self):
        """
        You need to add a docstring here ! 
        """
        return self.classname


    @autoStore
    @withClientUpdate
    def setClass(self, f):
        """
        You need to add a docstring here ! 
        """
        self.classname = f


    def checkBaseDir(self):
        """
        You need to add a docstring here ! 
        """
        if not self.getBaseDir() in sys.path:
            sys.path.append(self.getBaseDir())
        mdfile = self.getBaseDir() + os.sep + "Group%d" % self.getID()[0] + os.sep + "__init__.py"
        if not os.path.exists(mdfile):
            # just create it empty
            f = open(mdfile, "w")
            f.write("\"\"\" Init file for Resource Group %d \"\"\"" % self.getID()[0])
            f.close()


    def getCG(self):
        """
        You need to add a docstring here ! 
        """
        self.checkBaseDir()
        return ClassGenerator(modulename=self.getModule(),
            classname=self.classname, assetid=self.getID(),
            assetService=self.parent)


    def acquireAsset_gotAsset(self, r):
        """
        You need to add a docstring here ! 
        """
        URLCodeAsset.acquireAsset_gotAsset(self, r)
        self.checkBaseDir()
        # now check to see if we have already loaded the module....
        if sys.modules.has_key(self.getModule()):
            # ok, we need to reload to upgrade to the
            # newest version.
            self.server.RefreshModule(self.getModule())
        return 1


    def getStateToCacheAndObserveFor(self, perspective, observer):
        """
        You need to add a docstring here ! 
        """
        return Cacheable.getStateToCacheAndObserveFor(self, perspective, observer)


    def setCopyableState(self, state):
        """
        You need to add a docstring here ! 
        """
        return Cacheable.setCopyableState(self, state)
