# -*- test-case-name: mv3d.test.test_conductor -*-
# Copyright (C) 2006-2008 Mortal Coil Games
# See LICENSE for details.
"""
    The Server Class.
    The purpose of this class is to be a hub for all the activities
    of a single server. By itself, it takes care of connections
    using a connectionManager.
    The server publishes Interfaces which can be used to get to specific
    game functionality.
"""


import sys
import socket
import random
import logging
from time import time

from twisted.internet import task, defer
from twisted.python import rebuild
from twisted.application.service import MultiService
from twisted.application import internet
from twisted.conch.manhole_ssh import ConchFactory, TerminalRealm
from twisted.conch.manhole import ColoredManhole
from twisted.cred import checkers, portal
from twisted.conch.insults import insults
from twisted.python.log import msg

from zope.interface import Interface, implements

from mv3d.util.date import Date
from mv3d.net.security import Securable
from mv3d.reporter.event import Event
from mv3d.reporter.event import EventService
from mv3d.net.client import ServiceLoc
from mv3d.util.classgen import ClassGenerator


Date()
ClassGenerator() # get rid of pyflakes warnings


class ServerError(Exception):
    """
    Class for various errors related to servers
    """


class ConnectError(Exception):
    """
    Raised when trying to connect and failing
    """


class IService(Interface):
    """
    The basic interface for a MV3D Service
    """
    def configure(cf):
        """
        Configure this service using the config object cf
        """


    def getInterface(client, protocol):
        """
        Returns a valid interface for the given protocol assuming
        the client specified is allowed to get one
        """


    def getName():
        """
        Just return the name of the service (from Service)
        """


    def isLocal():
        """
        Returns true since this is a local service (feel free to override)
        """



class IIterableService(IService):
    """
    This service needs to be iterated every server tick.
    """
    def iterate(tm):
        """
        Process one server tick. tm specifies the time that has passed
        in seconds.
        """



class IAssetClient(IService):
    """
    Provides asset retrieval services
    """
    def getAsset(aid, local=False):
        """
        Get an asset. If the group is local and the asset
        exists, return it. If the group isn't local, connect
        to a server that has it and get it from there before
        returning it.
        """


    def acquireAsset(aid, **kw):
        """
        Download this asset.
        """



class IDirectoryClient(IService):
    """
    Provides a means of getting information on directory entries
    """
    def getMasterServersFor(nm, i):
        """
        Looks up i (id) in the directory for nm (str)
        return 0 if not found or other error
        return i's master servers otherwise
        it is possible for this to return a Deferred
        """


    def getSlaveServersFor(nm, i):
        """
        Looks up i (id) in the directory for nm (str)
        return 0 if not found or other error
        return i's slave servers otherwise
        it is possible for this to return a Deferred
        """


    def getServersFor(nm, i):
        """
        Looks up i (id) in the directory for nm (str)
        return 0 if not found or other error
        return i's servers otherwise
        it is possible for this to return a Deferred
        """


    def getDirectoryItem(nm, i, lo=False):
        """
        Get a directory item i from directory nm.
        if lo is True, only look at directories we have. Otherwise,
        Query our directory servers for i in directory nm 
        """



class IRealmClient(IIterableService):
    """
    Provides a means to get information about realms
    """
    def getDirectoryServers(): 
        """
        return all of our directory servers 
        """


    def addDirectoryServer(s):
        """
        Add a directory server (s = ServiceLoc) 
        """


    def remDirectoryServer(s):
        """
        Remove a directory server (s=ServiceLoc) 
            if we don't have it, return 0, else 1
        """



    def getRealmServer(realmid):
        """
        Find the realm server for a given realm
        """


    def getServiceLocsFor(realmid):
        """
        Get service locations for the given realm
        """


    def getServiceLocFor(realmid):
        """
        Get master service location for the given realm
        """


    def locateObject(obj, realmid):
        """
        Find which server an object is on
        """


    def getRealm(realmid, local=False):
        """
        Try to get a realm locally if we don't have it, cache
        it from the appropriate server and then return that.
        """




class ISimulationClient(IIterableService):
    """
    Provides client access to the simulated objects
    """
    def getItem(self, iid, local=False, recursive=False):
        """
        Get an item.
        First try to get the item locally. If it doesn't
        exist on our server, then locate it through the 
        realm server of the realm that it belongs to.
        Once located, cache the object from the server
        and return the local copy.
        
        Always returns a deferred
        """



class IPlayerClient(IService):
    """
    Handles player interaction from the client side
    """
    def getPCs(self):
        """
        This function will return a deferred, that will
        eventually give you a list of oids that your account
        can use as Players.
        """


    def connectPC(self, oid):
        """
        This function will return a deferred that will
        return 0 if the client doesn't have access to use
        oid as a PC. Otherwise, it'll return a PlayerManipulator
        object for that PC which we will store as self.pc
        """


    def do(self, what, *a, **kw):
        """
        Execute an action
        """


    def createCharacter(self):
        """
        Tell the server that we would like to create a new player
        character
        """


    def addAction(self, b):
        """
        Add an action
        """


    def remAction(self, nm):
        """
        Removes an action
        """


    def getActions(self):
        """
        Returns all registered actions
        """


    def getAction(self, nm):
        """
        Returns an action with the given name
        """


    def addBinding(self, n, b):
        """
        Add an action binding for n
        """


    def getBindingsFor(self, n):
        """
        Gets all the bindings for n
        """


    def remBindingsFor(self, n):
        """
        Remove all bindings called n
        """


    def remBinding(self, n, b):
        """
        Remove an action binding
        """


    def updateLoginServers(self):
        """ 
        This will get a server connection
        and ask it for suggestions of other login
        servers that we can use.. Update the list
        kept by the client with this new info.
        """


    def connectAndSelectPC(self):
        """
        Connect to a server and select and log in as a PC
        """



class IConductor(Interface):
    """
    Interface for a conductor
    """


def findParent(thing, iface):
    """
    Search parents until we find something that implements iface
    """
    while hasattr(thing, "parent") and thing.parent is not None:
        if iface.implementedBy(thing.parent):
            return thing.parent
        thing = thing.parent
    raise ValueError("No parent of %s implements %s." % (thing, iface))


def getHostIp():
    """
    Cheap trick to get the local IP of host
    that it is run on.
    """
    # http://mail.python.org/pipermail/python-list/2005-January/258981.html
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('1.2.3.4', 56))
    ip = s.getsockname()[0]
    return ip



class Conductor(Securable, MultiService):
    """
    The conductor is in charge of keeping everything in sync
    and working together.
    """
    implements(IConductor)

    tick = 0
    ips = 0
    lips = 0
    lastittime = 0
    runloop = None
    desiredIps = 20
    logLevel = logging.DEBUG

    def __init__(self):
        """
        Initialize the master conductor service.
        """
        MultiService.__init__(self)
        self.host = unicode(getHostIp())
        self.startdate = time()
        self.pips = []
        self.eventservice = EventService(self)
        self.eventlogs = []
        self.iterators = []
        self.iterServices = []
        self.logs = None
        self.connectionFactories = {}
        self.addPermission("login")
        self.addPermission("startstop")
        self.addPermission("kick")
        self.addPermission("ban")
        self.addPermission("run command")


    def configure(self, nm, cf):
        """
        Configure this server from the config options
        """
        if cf.has_option(nm, "name"):
            self.setName(cf.get(nm, "name"))
        if cf.has_option(nm, "logs"):
            self.logs = eval(cf.get(nm, "logs"))
        if cf.has_option(nm, "desiredIps"):
            self.desiredIps = cf.getint(nm, "desiredIps")
        if cf.has_option(nm, "logLevel"):
            self.logLevel = cf.getint(nm, "logLevel")
        svcs = cf.get(nm, "serviceNames").split(",")
        svcs = [s.strip() for s in svcs]
        for svcn in svcs:
            stype = cf.get(svcn, "type")
            smod = ".".join(stype.split(".")[:-1])
            scls = stype.split(".")[-1]
            svc = getattr(__import__(smod, globals(), locals(), [scls]), scls)()
            svc.setName(svcn)
            svc.setServiceParent(self)
            svc.configure(svcn, cf)
            if IIterableService.providedBy(svc):
                self.iterServices.append(svcn)
        cfacts = cf.get(nm, "connectionFactories").split(",")
        cfacts = [s.strip() for s in cfacts]
        for cfact in cfacts:
            ctype = cf.get(cfact, "type")
            cmod = ".".join(ctype.split(".")[:-1])
            ccls = ctype.split(".")[-1]
            fact = getattr(__import__(cmod, globals(), locals(), [ccls]), ccls)(self)
            fact.configure(cfact, cf)
            self.addConnectionFactory(fact)
        if cf.has_option(nm, "consolePort"):
            # make a console service
            port = int(cf.get(nm, "consolePort"))
            u = cf.get(nm, "consoleUsers").split(".")
            users = dict([tuple(uu.split(":")) for uu in u])
            checker = checkers.InMemoryUsernamePasswordDatabaseDontUse(**users)
            realm = TerminalRealm()
            consoleNamespace = dict(conductor=self)
            for svcn in svcs:
                consoleNamespace[svcn] = self.getServiceNamed(svcn)
            realm.chainedProtocolFactory = lambda: insults.ServerProtocol(
               ColoredManhole, consoleNamespace)
            portl = portal.Portal(realm, [checker])
            fact = ConchFactory(portl)
            service = internet.TCPServer(port, fact)
            service.setServiceParent(self)


    def iterate(self):
        """
        This function should be called at most at 20Hz.
        While calling it more often is not harmful, it can
        cause more frequent updates to clients.
        Calling it less often will cause things to slow down
        on the server.
        """
        try:
            self.tick += 1
            if self.lastittime != 0:
                tm = time()
                dt = tm - self.lastittime
                self.pips.append(self.lastittime)
                self.lastittime = tm
                if len(self.pips) > 40:
                    self.pips = self.pips[1:]
                ips = (tm - self.pips[0]) / float(len(self.pips))
                if ips != 0:
                    self.lips = self.ips
                    self.ips = 1 / ips
            else:
                self.lastittime = time()
                dt = 0.05
            for n in self.iterServices:
                self.getServiceNamed(n).iterate(dt)
            for i in self.iterators:
                i(dt)
        except:
            import traceback
            e = sys.exc_info()
            exc = traceback.format_exception(*e)
            for ee in exc:
                print ee


    def startService(self):
        """
        Start the server!
        """
        self.runloop = task.LoopingCall(self.iterate)
        self.runloop.start(0.05)
        if not "Extern" in sys.path:
            sys.path.append("Extern")
        return MultiService.startService(self)


    def stopService(self):
        """
        Stop the server!
        """
        if self.runloop.running:
            self.runloop.stop()
        for el in self.eventlogs:
            el.finalize()
        return MultiService.stopService(self)


    def getName(self):
        """
        Return the name of this server 
        """
        return self.name


    def getServiceAlias(self, protocol, service, type=None):
        """
        Returns a ServiceLoc that is an alias for the local
        service on th given protocol.
        """
        assert protocol != "local", "That doesn't make sense"
        fact = self.connectionFactories.get(protocol)
        if fact is not None:
            sl = fact.getLocalAlias(type)
            sl.path = "/" + service
            return sl
        raise ValueError("Unknown protocol %s" % protocol)


    def getService(self, loc):
        """
        Get a service based on a ServiceLoc
        """
        loc = ServiceLoc(loc)
        if loc.protocol == "local":
            return defer.succeed(
                self.getServiceNamed(loc.path.split("/")[1]))
        fact = self.connectionFactories.get(loc.protocol)
        if fact is not None:
            return fact.getService(loc)
        raise ValueError("Unknown protocol %s." % loc.protocol)


    def getOneService(self, locations):
        """
        Takes a list of locations, and gets the quickest service
        """
        locs = [ServiceLoc(l) for l in locations]
        bad = []
        for l in locs:
            if l.protocol == "local":
                try:
                    return self.getService(l)
                except KeyError:
                    bad.append(l)
        for b in bad:
            locs.remove(b)
        def tryAgain(ignored=None, remove=None):
            if remove is not None:
                locs.remove(remove)
            if not len(locs):
                raise ConnectError("Out of services in getOneService!")
            for l in locs:
                if self.isServiceConnected(l):
                    d = self.getService(l)
                    d.addErrback(tryAgain, l)
                    return d
            pick = random.choice(locs)
            try:
                d = self.getService(pick)
            except ValueError:
                return tryAgain(remove=pick)
            d.addErrback(tryAgain, pick)
            return d
        return tryAgain()


    def isServiceConnected(self, loc):
        """
        See if a service is connected based on the ServiceLoc
        """
        loc = ServiceLoc(loc)
        if loc.protocol == "local":
            return True
        fact = self.connectionFactories.get(loc.protocol)
        if fact is not None:
            return fact.isServiceConnected(loc)
        return False


    def getLocalServices(self, interface):
        """
        Get all local services that provide the given interface
        """
        res = []
        for s in self.services:
            if interface.providedBy(s):
                res.append(s)
        return res


    def getLocalService(self, interface):
        """
        Get the first local service that provides the given interface
        """
        svcs = self.getLocalServices(interface)
        if not len(svcs):
            raise KeyError(
                "No service provides %s." % interface.__class__.__name__)
        return svcs[0]


    def addConnectionFactory(self, factory):
        """
        Add a connection factory to this conductor
        """
        assert not self.connectionFactories.has_key(factory.getProtocol())
        self.connectionFactories[factory.getProtocol()] = factory


    def getEventService(self):
        """
        Simply return out event service 
        """
        return self.eventservice


    def broadcastEvent(self, e=0, **kw):
        """
        Send out an event to the eventservice.
        Either e is an event, or use keywords to create
        one.
        """
        if e == 0:
            e = Event(**kw)
        # e.setServer(ServiceLoc().fromServer(self))
        print " ".join(str(e).split(" ")[2:])
        self.getEventService().pushEvent(e)


    def receiveEvent(self, e):
        """
        This is called when an event comes in through a connection
        """
        self.GetEventService().pushEvent(e)


    def uptime(self):
        """
        Returns the amount of time since the server was started 
        """
        return time() - self.startdate


    def getIPS(self):
        """
        returns the average iterations per second (over the last few 
        seconds)
        """
        return self.ips


    def getDesiredIPS(self):
        """
        returns the desired iteration rate. 
        """
        return 20


    def getLoad(self):
        """
        returns the server load 0 - 1, with 1 being no load. 
        """
        return self.ips / self.getDesiredIPS()


    def getLastLoad(self):
        """
        Returns the load at the last iteration. values are
        in the same format as getLoad
        """
        return self.lips / self.GetDesiredIPS()


    def getTick(self):
        """
        returns the current tick (number of iterations 
        """
        return self.tick


    def addIterator(self, i):
        """
        Add an Iterator to the list. i is a function that accepts
        one argument. The argument will be the amount of time that
        has passed since the last iteration (in theory). It will be
        called back every iteration with that arg.
        """
        self.iterators.append(i)


    def remIterator(self, i):
        """
        Remove an Iterator from the list 
        """
        self.iterators.remove(i)


    def refreshModule(self, m):
        """
        This notifies Python that module m
        has changed on disk and should be reloaded.
        Using Twisted's Rebuild module, we will reload
        it.
        """
        self.log("Refreshing module %s" % m.__name__, logging.INFO)
        if isinstance(m, str):
            if not sys.modules.has_key(m):
                return 0
            m = sys.modules[m]
        if (isinstance(m, tuple) or isinstance(m, int) or isinstance(m, 
            float) or isinstance(m, list) or isinstance(m, str) or 
            m == None):
            return 0
        return rebuild.rebuild(m)


    def refreshModules(self, ms):
        """
        Same as RefreshModule, but ms is a list
        of modules (or names of modules)
        """
        for m in ms:
            self.RefreshModule(m)
        return 1


    def refreshAll(self):
        """
        This will go through all the modules in sys.modules
        and update them all. IT IS PROBABLY NOT WHAT YOU WANT
        TO DO. This should probably not be used on a live server.
        """
        return  self.RefreshModules(sys.modules.keys())


    def log(self, message, logLevel=50, **kw):
        """
        Emid a log message
        """
        if logLevel > self.logLevel:
            msg(message, **kw)