# -*- test-case-name: mv3d.test.test_axiomstore -*-
# Copyright (C) 2008 Mortal Coil Games
# See LICENSE for details.
"""
"""

from time import time

try:
    import cPickle as pickle
except ImportError:
    import pickle

from axiom import store, item, attributes, errors

from twisted.internet.task import LoopingCall

from mv3d.util.datastore.store import (
    PickleBasedStore, StoreError, TranslatingAttribute,
    VersionError
    )
from mv3d.util.datastore.localfile import ObjectContainer
from mv3d.util.datastore.query import QueryError



class AxiomVector(item.Item):
    """
    Provides a way to DirectStore a vector
    """
    x = attributes.ieee754_double()
    y = attributes.ieee754_double()
    z = attributes.ieee754_double()
    did = attributes.text()
    attribute = attributes.text()
    getter = attributes.text()
    setter = attributes.text()
    object = attributes.inmemory()


    def toStore(self):
        """
        Send object.attribute to the store
        """
        if self.getter is not None:
            value = getattr(self.object, self.getter)()
        else:
            value = getattr(self.object, self.attribute)
        value = tuple(value)
        if (self.x, self.y, self.z) != value:
            self.x, self.y, self.z = tuple(value)


    def fromStore(self):
        """
        Take our data and send it to object.attribute
        """
        return (self.x, self.y, self.z)



class AxiomQuaternion(item.Item):
    """
    Provides a way to DirectStore a vector
    """
    w = attributes.ieee754_double()
    x = attributes.ieee754_double()
    y = attributes.ieee754_double()
    z = attributes.ieee754_double()
    did = attributes.text()
    attribute = attributes.text()
    getter = attributes.text()
    setter = attributes.text()
    object = attributes.inmemory()


    def toStore(self):
        """
        Send object.attribute to the store
        """
        if self.getter is not None:
            value = getattr(self.object, self.getter)()
        else:
            value = getattr(self.object, self.attribute)
        value = tuple(value)
        if (self.w, self.x, self.y, self.z) != value:
            self.w, self.x, self.y, self.z = value


    def fromStore(self):
        """
        Take our data and send it to object.attribute
        """
        return (self.w, self.x, self.y, self.z)



class AxiomPhysics(item.Item):
    """
    This represents the standard physics values for
    stuff (pos, rot, lv, av)
    """
    did = attributes.text()
    attribute = attributes.text()
    getter = attributes.text()
    setter = attributes.text()
    object = attributes.inmemory()

    px = attributes.ieee754_double()
    py = attributes.ieee754_double()
    pz = attributes.ieee754_double()
    rw = attributes.ieee754_double()
    rx = attributes.ieee754_double()
    ry = attributes.ieee754_double()
    rz = attributes.ieee754_double()
    lvx = attributes.ieee754_double()
    lvy = attributes.ieee754_double()
    lvz = attributes.ieee754_double()
    avx = attributes.ieee754_double()
    avy = attributes.ieee754_double()
    avz = attributes.ieee754_double()


    def getValue(self):
        """
        Get our value as a set of tuples
        """
        return ((self.px, self.py, self.pz),
            (self.rw, self.rx, self.ry, self.rz),
            (self.lvx, self.lvy, self.lvz),
            (self.avx, self.avy, self.avz))


    def toStore(self):
        """
        Send object.attribute to the store
        """
        if self.getter is not None:
            value = getattr(self.object, self.getter)()
        else:
            value = getattr(self.object, self.attribute)
        if self.getValue() != value:
            self.px, self.py, self.pz = value[0]
            self.rw, self.rx, self.ry, self.rz = value[1]
            self.lvx, self.lvy, self.lvz = value[2]
            self.avx, self.avy, self.avz = value[2]


    def fromStore(self):
        """
        Take our data and send it to object.attribute
        """
        return (self.w, self.x, self.y, self.z)



class DirectStore(TranslatingAttribute):
    """
    This attribute defines something that is stored
    directly in axiom
    """
    def __init__(self, storeClass, autoStore=False, getter=None, setter=None):
        self.storeClass = storeClass
        self.autoStore = autoStore
        self.getter = getter
        self.setter = setter


    def toStore(self, store, value, object, attribute):
        """
        Save value in the store

        This could probably be made slightly faster
        by changing the "did" parameter to be the storeID
        of the AxiomContainer for object.
        """
        sc = store.store.findOrCreate(self.storeClass,
            did=object._datastore_id,
            attribute=unicode(attribute),
            getter=self.getter,
            setter=self.setter)
        sc.object = object
        sc.toStore()
        if self.autoStore:
            store.addAutoStore(sc)
        return sc.storeID


    def fromStore(self, store, value, object):
        """
        Retrieve the value from the store
        """
        sc = store.store.getItemByID(value)
        sc.object = object
        r = sc.fromStore()
        if self.setter is not None:
            getattr(object, self.setter)(r)
        if self.autoStore:
            store.addAutoStore(sc)
        return r



class AxiomContainer(item.Item, ObjectContainer):
    """
    This serves as the Axiom object that gets stored in the
    datastore and houses other objects that we want to
    persist
    """
    object = attributes.inmemory()
    obdict = attributes.inmemory()
    datastore = attributes.inmemory()
    clazz = attributes.text()
    module = attributes.text()
    did = attributes.text()
    data = attributes.text()
    revision = attributes.integer()

    def __init__(self, *a, **kw):
        """
        Convert the object to pickle at this time
        """
        item.Item.__init__(self, *a, **kw)
        if kw.has_key("object"):
            self.freezeObject(kw["object"])


    def freezeObject(self, object):
        """
        Store an object
        """
        self.clazz = unicode(object.__class__.__name__)
        self.module = unicode(
            object.__class__.__module__)
        obdict = dict()
        for k, v in self.object.datastoreAttributes.items():
            obdict[k] = getattr(self.object, k)
            if isinstance(v, TranslatingAttribute):
                obdict[k] = v.toStore(self.datastore,
                    obdict[k], self.object, k)
        obdict["datastoreVersion"] = getattr(self.object,
            "datastoreVersion", 1)
        self.data = unicode(pickle.dumps(obdict))


    def activate(self):
        """
        Just do some basic stuff to get ready
        """
        if self.object is None:
            class dummy:
                pass
            self.object = dummy()


    def buildObject(self, store):
        """
        Called after being unpickled
        (copied from localfile because we can't do
        self.store=store)
        """
        clazz = getattr(__import__(self.module, globals(),
            locals(), self.clazz), self.clazz)
        # if isinstance(clazz, object):
            # raise StoreError("%s is a new style class" % self.clazz)
        self.object.__class__ = clazz
        version = getattr(self.object, "datastoreVersion", 1)
        self.object.__dict__.update(self.obdict)
        if version != self.object.__class__.datastoreVersion:
            if not hasattr(self.object, "datastoreUpgrade"):
                raise VersionError(
                    "No method to update %s from version %d to %d." % (
                    clazz.__name__, version, 
                    self.object.__class__.datastoreVersion))
            self.object.datastoreUpgrade(store, version)
        for k, v in self.object.datastoreAttributes.items():
            if isinstance(v, TranslatingAttribute):
                setattr(self.object, k, v.fromStore(store, 
                getattr(self.object, k), self.object))
        self.object._datastore_id = self.did
        self.object.datastore = store
        self.datastore = store
        return self.object



class AxiomIndex(item.Item):
    """
    This defines an entry in an index
    """
    clazz = attributes.text()
    attribute = attributes.text()
    item = attributes.text()
    revision = attributes.integer()
    valueBool = attributes.boolean()
    valueInt = attributes.integer()
    valueFloat = attributes.ieee754_double()
    valueStr = attributes.text()
    valueRef = attributes.text()
    valueId = attributes.text()


    @classmethod
    def prepQueryAttribute(cls, other):
        """
        Returns information about what attributes to query
        based on the type of other, but also formats other
        properly.
        """
        if hasattr(other, "_datastore_id"):
            return AxiomIndex.valueRef, unicode(other._datastore_id)
        if isinstance(other, bool):
            return AxiomIndex.valueBool, other
        if isinstance(other, float):
            return AxiomIndex.valueFloat, other
        if isinstance(other, int):
            return AxiomIndex.valueInt, other
        if isinstance(other, (str, unicode)):
            return AxiomIndex.valueStr, unicode(other)
        if isinstance(other, tuple):
            assert isinstance(other[0], int), "Invalid id tuple %s" % repr(other)
            return AxiomIndex.valueId, unicode(repr(other))
        raise ValueError("Invalid query type %r" % other)


    def setValue(self, value):
        """
        Sets the value attribute warranted by what type value is
        """
        if hasattr(value, "_datastore_id"):
            self.valueRef = value._datastore_id
        elif isinstance(value, bool):
            self.valueBool = value
        elif isinstance(value, float):
            self.valueFloat = value
        elif isinstance(value, int):
            self.valueInt = value
        elif isinstance(value, (str, unicode)):
            self.valueStr = unicode(value)
        elif isinstance(value, tuple):
            assert isinstance(value[0], int), "Invalid id tuple %s" % repr(value)
            self.valueId = unicode(repr(value))
        else:
            raise ValueError("Invalid value for index %r" % value)



def getIndexes(clazz):
    """
    Get the datastoreIndexes from a class
    Somewhat recursive.
    """
    idxes = []
    if not hasattr(clazz, "datastoreIndexes"):
        return []
    for i in clazz.datastoreIndexes:
        if hasattr(i, "datastoreIndexes"):
            idxes += getIndexes(i)
        else:
            idxes.append((clazz, i))
    return idxes


class AxiomCombination:
    """
    Defines and or or
    """
    def __init__(self, q1, op, q2):
        self.q1 = q1
        self.op = op
        self.q2 = q2


    def doQuery(self, store, cname, revision=None):
        left = self.q1.doQuery(store, cname, revision)
        leftids = [x.item for x in left]
        right = self.q2.doQuery(store, cname, revision)
        rightids = [x.item for x in right]
        res = []
        if self.op == "&":
            for x, l in zip(leftids, left):
                if x in rightids:
                    res.append(l)
        if self.op == "|":
            for x, l in zip(leftids, left):
                if not x in rightids:
                    res.append(l)
            for x, r in zip(rightids, right):
                if not x in leftids:
                    res.append(r)
        return res



class AxiomQueryCompare:
    """
    Defines a comparison between axiom attributes
    """
    def __init__(self, clazz1, name1, compare, value=None, clazz2=None, name2=None):
        self.clazz1 = clazz1
        self.name1 = name1
        self.compare = compare
        self.value = value
        self.clazz2 = clazz2
        self.name2 = name2


    def makeQuery(self):
        if self.clazz2 is None:
            q1 = AxiomIndex.attribute==self.name1
            qa, value = AxiomIndex.prepQueryAttribute(self.value)
            if self.compare == ">":
                q2 = qa > value
            elif self.compare == "<":
                q2 = qa < value
            elif self.compare == "==":
                q2 = qa == value
            elif self.compare == "!=":
                q2 = qa != value
            return attributes.AND(q1, q2)


    def doQuery(self, store, cname, revision=None):
        q1 = AxiomIndex.clazz == unicode(cname)
        if revision is not None:
            q1 = attributes.AND(q1, AxiomIndex.revision <= revision)
        return store.query(AxiomIndex, attributes.AND(
            q1,
            self.makeQuery()))


    def __and__(self, other):
        return AxiomCombination(self, "&", other)


    def __or__(self, other):
        return AxiomCombination(self, "|", other)



class AxiomQueryParam:
    """
    Allows operators to be used on it in order to generate
    a query
    """
    def __init__(self, clazz, name):
        self.clazz = clazz
        self.name = unicode(name)


    def __gt__(self, other):
        if not isinstance(other, AxiomQueryParam):
            return AxiomQueryCompare(self.clazz, self.name, ">", other)


    def __lt__(self, other):
        if not isinstance(other, AxiomQueryParam):
            return AxiomQueryCompare(self.clazz, self.name, "<", other)


    def __eq__(self, other):
        if not isinstance(other, AxiomQueryParam):
            return AxiomQueryCompare(self.clazz, self.name, "==", other)


    def __ne__(self, other):
        if not isinstance(other, AxiomQueryParam):
            return AxiomQueryCompare(self.clazz, self.name, "!=", other)



class AxiomQuery:
    """
    This is the query object used by the axiom store
    it can fetch query results based on indexes generated
    by the store in the axiom db
    """
    def __init__(self, store, clazz, revision=None):
        self.store = store
        self.clazz = clazz
        self.revision = revision
        idxes = getIndexes(clazz)
        for c, idx in idxes:
            setattr(self, idx, AxiomQueryParam(c, idx))


    def xfetchResults(self, query=None, offset=None, limit=None,
            sort=None):
        """
        Actually perform the query
        
        Some example queries
            q.fetchResults(q.name=="Mike")
            q.fetchResults(q.age>=20)
            q.fetchResults(q.name=="Mike" & q.age<100) 
        """
        cname = self.clazz.__name__
        if not "." in cname:
            cname = ".".join([self.clazz.__module__, self.clazz.__name__])
        if query is not None:
            results = query.doQuery(self.store.store, cname, self.revision)
        else:
            attrib = getIndexes(self.clazz)[-1][1]
            q1 = AxiomIndex.clazz == unicode(cname)
            if self.revision is not None:
                q1 = attributes.AND(q1, AxiomIndex.revision <= self.revision)
            results = self.store.store.query(AxiomIndex,
                attributes.AND(q1,
                AxiomIndex.attribute == unicode(attrib)),
                offset=offset, limit=limit, sort=sort)
        for result in results:
            if self.revision is not None:
                r = self.store.store.findFirst(AxiomIndex, attributes.AND(
                    AxiomIndex.item == result.item,
                    AxiomIndex.clazz == result.clazz,
                    AxiomIndex.attribute == result.attribute,
                    AxiomIndex.revision <= self.revision), 
                    sort=AxiomIndex.revision.descending)
                if result.revision == r.revision:
                    yield self.store.retrieveObject(result.item)
            else:
                yield self.store.retrieveObject(result.item)


    def fetchResults(self, query=None, offset=None, limit=None,
            sort=None):
        """
        Actually perform the query
        
        Some example queries
            q.fetchResults(q.name=="Mike")
            q.fetchResults(q.age>=20)
            q.fetchResults(q.name=="Mike" & q.age<100) 
        """
        def inTransaction():
            return list(self.xfetchResults(query, offset, limit, sort))
        return self.store.transact(inTransaction)


    def findUnique(self, query=None, offset=None, limit=None, sort=None):
        """
        Run a query, and raise an error if there is not
        one single result
        """
        r = self.xfetchResults(query, offset, limit, sort)
        result = None
        for res in r:
            if result is None:
                result = res
            else:
                raise QueryError("Query did not return unique results")
        if result is None:
            raise QueryError("Query did not return any results")
        return result


    def findFirst(self, query=None, offset=None, limit=None, sort=None):
        """
        Run a query, and return the first result (or error if none)
        """
        r = self.xfetchResults(query, offset, limit, sort)
        for res in r:
            return res
        raise QueryError("Query did not return any results")



class AxiomStore(PickleBasedStore):
    """
    This store works like other pickle based stores except
    in this case, the data is not saved directly to disk
    but instead goes to Axiom (which may then save it to disk
    through SQLite).
    """
    store = None
    lock = False
    autoStore = None
    autoStoreFrequency = 15


    def __init__(self):
        self.objectcache = {}
        self.retrieving = {}
        self.autoStoreObjects = []


    def create(self, path=None):
        """
        Create the datastore.
        """
        self.store = store.Store(path)
        self.lock = True


    def Log(self, message):
        """
        """
        # print message


    def open(self, path):
        """
        Open a pre-existing datastore
        """
        self.store = store.Store(path)
        self.lock = True


    def openOrCreate(self, path, force=True):
        """
        Open a pre-existing store or create a new one.
        
        The force argument is to satisfy the api.
        """
        return self.open(path)


    def storeObject(self, object, async=False):
        """
        Store the object synchronously
        """
        if async:
            raise StoreError("This is not an async store!")
        return self.store.transact(
            PickleBasedStore.storeObject, self, object, async=False)


    def storeObjects(self, objects):
        """
        Store a bunch of objects in a transaction
        """
        def inTransaction():
            oids = []
            for o in objects:
                oids.append(self.storeObject(o))
            return oids
        return self.store.transact(inTransaction)


    def storeObjectSync(self, object):
        """
        Synchronously store an object via an AxiomContainer
        """
        self.Log("Storing %s" % str(object.__class__.__name__))
        if hasattr(object, "_datastore_id"):
            oid = object._datastore_id
        else:
            oid = self.createObjectId(object)
            # do a sanity check to make sure that nothing is
            # already using this oid
            r = self.store.query(AxiomContainer, AxiomContainer.did == oid)
            if r.count() != 0:
                raise StoreError("Duplicate oid detected for class %s" %
                    object.__class__.__name__)
            object._datastore_id = oid
        oc = self.objectcache.get(oid)
        if oc:
            oc.freezeObject(object)
        else:
            oc = AxiomContainer(store=self.store, object=object,
                did=oid, datastore=self)
        assert oc.did == oid
        assert oc.did is not None
        object.__dict__["_datastore_id"] = oid
        for clazz, itm in getIndexes(object.__class__):
            i = self.store.findOrCreate(AxiomIndex,
                clazz=unicode(clazz),
                attribute=unicode(itm),
                item=oid)
            attr = getattr(object, itm)
            i.setValue(attr)
        self.objectcache[oid] = oc
        if hasattr(object, "_datastore_storing"):
            del object._datastore_storing
        return oc.did


    def createObjectId(self, object):
        """
        Just call super and convert to unicode
        """
        return unicode(
            PickleBasedStore.createObjectId(self, object))


    def retrieveObjectSync(self, oid):
        def retrieve():
            try:
                oc = self.store.findUnique(AxiomContainer,
                    AxiomContainer.did == oid)
            except errors.ItemNotFound:
                raise StoreError("Item not found")
            self.objectcache[oid] = oc
            if oc.object is not None:
                oc.obdict = pickle.loads(str(oc.data))
                oc.buildObject(self)
            return oc.getObject()
        return self.store.transact(retrieve)


    def retrieveObject(self, oid, async=False, useCache=True):
        """
        Retrieve the object synchronously
        """
        if async:
            raise StoreError("This is an async store!")
        return PickleBasedStore.retrieveObject(self, oid, async=False,
            useCache=useCache)


    def retrieveObjects(self, objects):
        """
        Retrieve a bunch of objects in a transaction
        """
        def inTransaction():
            ret = []
            for o in objects:
                ret.append(self.retrieveObject(o))
            return ret
        return self.store.transact(inTransaction)


    def transact(self, f, *a, **kw):
        """
        Run some commands in a tranaction
        """
        return self.store.transact(f, *a, **kw)


    def addAutoStore(self, directStore):
        """
        Add a direct store object to be auto stored
        """
        self.autoStoreObjects.append(directStore)
        self.startAutoStore()


    def remAutoStore(self, directStore):
        """
        Remove an autoStore entry
        """
        self.autoStoreObjects.remove(directStore)
        if not len(self.autoStoreObjects):
            self.stopAutoStore()


    def startAutoStore(self):
        """
        Start auto storing
        """
        if self.autoStore is None:
            self.autoStore = LoopingCall(self.syncAutoStore)
            self.autoStore.start(self.autoStoreFrequency, now=False)


    def stopAutoStore(self):
        """
        Stop auto storing
        """
        if self.autoStore is not None:
            self.syncAutoStore()
            if self.autoStore.running:
                self.autoStore.stop()
            self.autoStore = None


    def syncAutoStore(self):
        """
        Sync all the auto-store objects
        """
        start = time()
        def inTransaction():
            for ds in self.autoStoreObjects:
                ds.toStore()
            # print "Synced %d autoStore objects in %.2f" % (
                # len(self.autoStoreObjects), time() - start)
        self.transact(inTransaction)


    def delete(self, object):
        """
        Remove an object from the store
        """
        def inTransaction():
            oid = object._datastore_id
            oc = self.objectcache[oid]
            oc.deleteFromStore()
            del object._datastore_id
            object.datastore = None
            rem = []
            for ds in self.autoStoreObjects:
                if ds.did == oid:
                    rem.append(ds)
                    ds.deleteFromStore()
            for ds in rem:
                self.autoStoreObjects.remove(ds)
        self.transact(inTransaction)


    def close(self):
        """
        Close this datastore
        """
        self.stopAutoStore()
        self.store.close()
        self.lock = False


    def query(self, clazz, async=False):
        """
        Return a query object
        """
        return AxiomQuery(self, clazz)