# Copyright (C) 2007-2008 Mortal Coil Games
# See LICENSE for details.
""" -*- test-case-name: mv3d.test.test_datastore -*-

"""


import os, sys, time
from threading import RLock
try:
    import cPickle as pickle
except ImportError:
    import pickle

from twisted.internet import defer
from twisted.internet.threads import deferToThread

from mv3d.util.datastore.store import (
    PickleBasedStore, StoreError, ObjectDoesntExist, TranslatingAttribute,
    VersionError)


class ObjectContainer:
    """
    This class wraps all objects that are stored
    """
    object = None
    clazz = None
    module = None
    oid = None
    store = None
    
    
    def __init__(self, store, obj, oid):
        self.store = store
        self.object = obj
        self.oid = oid
        
        
    def getObject(self):
        return self.object


    def __getstate__(self):
        """
        Called when pickled. Get the storable attributes from the 
        """
        obdict = dict()
        for k, v in self.object.datastoreAttributes.items():
            obdict[k] = getattr(self.object, k)
            if isinstance(v, TranslatingAttribute):
                obdict[k] = v.toStore(self.store, obdict[k], self.object, k)
        obdict["datastoreVersion"] = getattr(self.object, "datastoreVersion", 1)
        clazz = ".".join([self.object.__class__.__module__, 
            self.object.__class__.__name__.split(".")[-1]])
        state = dict(obdict=obdict, oid=self.oid, clazz=clazz)
        return state
 
 
    def __setstate__(self, state):
        """
        Given state from the pickler, recreate the object
        """
        self.obdict = state["obdict"]
        self.oid = state["oid"]
        self.clazz = state["clazz"].split(".")[-1]
        self.module = ".".join(state["clazz"].split(".")[:-1])
        class dummy:
            pass
        self.object = dummy()


    def buildObject(self, store):
        """
        Called after being unpickled
        """
        clazz = getattr(__import__(self.module, globals(),
            locals(), self.clazz), self.clazz)
        # if isinstance(clazz, object):
            # raise StoreError("%s is a new style class" % self.clazz)
        # print clazz, type(clazz), self.clazz, self.module
        self.object.__class__ = clazz
        version = getattr(self.object, "datastoreVersion", 1)
        self.object.__dict__.update(self.obdict)
        if version != self.object.__class__.datastoreVersion:
            if not hasattr(self.object, "datastoreUpgrade"):
                raise VersionError(
                    "No method to update %s from version %d to %d." % (
                    clazz.__name__, version, 
                    self.object.__class__.datastoreVersion))
            self.object.datastoreUpgrade(store, version)
        for k, v in self.object.datastoreAttributes.items():
            if isinstance(v, TranslatingAttribute):
                setattr(self.object, k, v.fromStore(store, 
                getattr(self.object, k)))
        self.object._datastore_id = self.oid
        self.object.datastore = store
        self.store = store
        return self.object



class LocalFileDataStore(PickleBasedStore):
    """
    This DataStore saves its data to local files
    """
    def __init__(self):
        self.path = None
        self.objectcache = {}
        self.locations = {}
        self.queryindexes = {}
        self.userdata = {}
        self.objectnames = {}
        self.retrieving = {}
        self.lock = False
        self.storeversion = 1
        self.logfile = None
        self.plock = RLock()
    

    def __del__(self):
        # print "Del called!", self.lock
        if self.lock:
            try:
                self.updateIndex()
            finally:
                self.unlockStore()
                
                
    def __getstate__(self):
        raise StoreError("You can't store the datastore, silly!")
        
        
    def Log(self, message):
        if self.logfile is None:
            return
        self.logfile.write("%f - %s\n"%(time.time(), message))
        self.logfile.flush()

    
    def create(self, path=None):
        """
        Create a new datastore at kw[path]
        """
        d = defer.Deferred()
        try:
            self.path = path
            if os.path.exists(self.path):
                raise StoreError("Path already exists when creating a DataStore!")
            os.makedirs(self.path)
            self.lockStore()
            self.updateIndex()
            d.callback(True)
        except:
            d.errback(sys.exc_info()[1])
        return d
    
    
    def open(self, path=None):
        """
        Open a datastore at kw[path]
        """
        d = defer.Deferred()
        self.path = path
        self.lockStore()
        def openThread():
            """
            Since this could take a long time, we put it in a thread
            """
            self.loadIndex()
            return True
        return deferToThread(openThread)
        
        
    def close(self):
        """
        Close this datastore
        """
        self.Log("Closing the datastore")
        def closeThread():
            self.updateIndex()
            self.unlockStore()
        return deferToThread(closeThread)
        
        
    def sync(self):
        """
        Sync this datastore to persistant media
        """
        def syncThread():
            self.updateIndex()
        return deferToThread(syncThread)
        
        
    def syncDump(self, ob, f):
        """
        A synchronous pickle dumper
        """
        self.plock.acquire()
        try:
            pickle.dump(ob, f)
        finally:
            self.plock.release()
        
        
    def syncLoad(self, f):
        """
        A synchronous pickle loader
        """
        self.plock.acquire()
        try:
            r = pickle.load(f)
        finally:
            self.plock.release()
        return r


    def storeObjectSync(self, object):
        self.Log("Storing %s"%str(object.__class__.__name__))
        if hasattr(object, "_datastore_id"):
            oid = object._datastore_id
        else:
            oid = self.createObjectId(object)
        oc = ObjectContainer(self, object, oid)
        object.__dict__["_datastore_id"] = oid
        filename = self.getFileFor(oid)
        f = open(filename, "w")
        try:
            self.syncDump(oc, f)
        except:
            from traceback import format_exc
            self.Log("Error Storing! %s %s %s %s"% (object, 
                sys.exc_info()[0], sys.exc_info()[1], format_exc()))
            raise #sys.exc_info()[1]
        f.close()
        self.objectcache[oid] = oc
        self.locations[oid] = filename
        if hasattr(object, "_datastore_storing"):
            del object._datastore_storing
        self.updateQueryIndex(object)
        return oid


    def retrieveObjectSync(self, oid):
        if not self.locations.has_key(oid):
            raise ObjectDoesntExist(
                "There is no object with that id in this store")
        filename = self.locations[oid]
        f = open(filename, "r")
        oc = self.syncLoad(f)
        f.close()
        self.objectcache[oid] = oc
        oc.buildObject(self)
        self.resetObjectQueryValues(oc.getObject())
        return oc.getObject()


    def delete(self, object):
        """
        Delete object from the store
        """
        self.removeFromIndexes(object)
        object.datastore = None
        oid = object._datastore_id
        filename = self.locations[oid]
        del self.locations[oid]
        if self.objectcache.has_key(oid):
            del self.objectcache[oid]
        os.remove(filename)
        return defer.succeed(True)


    def updateIndex(self):
        """
        Update the store's index file
        """
        self.Log("Updating store index.")
        indexfile = os.path.join(self.path, "index")
        f = open(indexfile, "w")
        self.syncDump(self.storeversion, f)
        self.syncDump(self.locations, f)
        self.syncDump(self.userdata, f)
        self.syncDump(self.objectnames, f)
        self.syncDump(self.queryindexes.keys(), f)
        f.close()
        self.storeQueryIndexes()
        
        
    def loadIndex(self):
        """
        Load the index of the store into memory
        """
        self.Log("Loading store index.")
        indexfile = os.path.join(self.path, "index")
        f = open(indexfile, "r")
        sv = self.syncLoad(f)
        if sv != self.storeversion:
            raise StoreError("The store is an incompatible version!")
        self.locations = self.syncLoad(f)
        self.userdata = self.syncLoad(f)
        self.objectnames = self.syncLoad(f)
        qis = self.syncLoad(f)
        for qi in qis:
            self.loadQueryIndex(qi)
        f.close()
    
    
    def lockStore(self):
        """
        Lock this store for writing
        
        Ideally, this should include our ip and port so that anyone trying to 
        get the lock can instead connect remotely to us.
        """
        self.Log("Locking the store.")
        lockfile = os.path.join(self.path, "lock")
        if os.path.exists(lockfile):
            raise StoreError("The store is already locked!")
        f = open(lockfile, "w")
        self.syncDump([os.getpid(), self.path], f)
        f.close()
        self.lock = True
        logfile = os.path.join(self.path, "store.log")
        self.logfile = open(logfile, "a")
        
    
    def unlockStore(self):
        """
        Unlock the store if we previously had it locked
        """
        self.Log("Unlocking the store.")
        if not self.lock:
            raise StoreError("Can't unlock the DataStore if it isn't locked!")
        lockfile = os.path.join(self.path, "lock")
        os.remove(lockfile)
        self.logfile.close()
        self.lock = False
        
        
    def getFileFor(self, oid):
        """
        Get a unique filename for this object id.
        """
        if self.locations.has_key(oid):
            return self.locations[oid]
        return os.path.join(self.path, oid + ".pkl")
        
        
    def storeQueryIndex(self, classname):
        """
        Stores all the query indexes for classname
        """
        filename = os.path.join(self.path, "queryindex."+classname)
        f = open(filename, "w")
        self.syncDump(self.queryindexes[classname], f)
        f.close()
        
        
    def storeQueryIndexes(self):
        """
        Store all of our query indexes
        """
        for k in self.queryindexes.keys():
            self.storeQueryIndex(k)
            
            
    def loadQueryIndex(self, classname):
        """
        Loads all the query indexes for a classname
        """
        filename = os.path.join(self.path, "queryindex."+classname)
        f = open(filename, "r")
        self.queryindexes[classname] = self.syncLoad(f)
        f.close()
