# -*- test-case-name: mv3d.test.test_memaxiom -*-
# Copyright (C) 2008 Mortal Coil Games
# See LICENSE for details.
"""
"""

from time import time
try:
    import cPickle as pickle
except ImportError:
    import pickle

from axiom import item, attributes

from twisted.internet.task import coiterate
from twisted.internet.defer import inlineCallbacks

from mv3d.util.datastore.store import StoreError, PickleBasedStore

from mv3d.util.datastore.axiomstore import (
    AxiomStore, AxiomContainer, AxiomIndex,
    AxiomQuery, getIndexes
    )


class StoreRevision(item.Item):
    """
    Holds a revision number for a store
    """
    revision = attributes.integer()
    timestamp = attributes.integer()



class SyncIterator:
    """
    Syncs up a few items at a time
    """
    def __init__(self, store, changes, perIter):
        self.store = store
        self.changes = changes
        self.perIter = perIter
        self.cleanup = []


    def __iter__(self):
        return self


    def next(self):
        """
        Sync up to perIter items
        """
        def inTransaction():
            for x in xrange(self.perIter):
                try:
                    change = self.changes.pop(0)
                except IndexError:
                    return False
                oc, idxes = change
                oc.store = self.store.store
                self.cleanup.append(oc.did)
                for i in idxes:
                    i.store = self.store.store
            return True
        if not self.store.transact(inTransaction):
            raise StopIteration


class ItemCleaner:
    """
    Cleans up old items
    """
    def __init__(self, store, items, perIter):
        self.store = store
        self.items = items
        self.perIter = perIter
        self.revision = self.store.getRevision()


    def __iter__(self):
        return self


    def next(self):
        """
        Clean up the next perIter items
        """
        def inTransaction():
            for x in xrange(self.perIter):
                try:
                    item = self.items.pop(0)
                except IndexError:
                    return False
                copies = self.store.store.query(AxiomContainer,
                    attributes.AND(
                    AxiomContainer.revision != self.revision,
                    AxiomContainer.did == item))
                for copy in copies:
                    copy.deleteFromStore()
                indexes = self.store.store.query(AxiomIndex,
                    attributes.AND(
                    AxiomIndex.item == item,
                    AxiomIndex.revision != self.revision))
                for index in indexes:
                    index.deleteFromStore()
            return True
        if not self.store.transact(inTransaction):
            raise StopIteration



class MemoryAxiomStore(AxiomStore):
    """
    This store keeps a queue of changes that need to be
    persisted when sync is called
    """
    syncing = False
    revision = None

    def __init__(self):
        self.objectcache = {}
        self.retrieving = {}
        self.changes = []


    @inlineCallbacks
    def sync(self, perIter=50):
        """
        Asynchronously sync the store in the background
        """
        if self.syncing:
            raise StoreError("Can't Sync-- already sync'ing!")
        self.syncing = True
        syncIter = SyncIterator(self, self.changes, perIter)
        yield coiterate(syncIter)
        self.updateRevision(self.revision + 1)
        yield self.cleanUp(syncIter.cleanup)
        self.syncing = False
        # nitems = self.store.count(AxiomContainer)
        # nindexes = self.store.count(AxiomIndex)
        # print "Items", nitems, "indexes", nindexes


    def cleanUp(self, items, perIter=10):
        """
        Given a list of items, iterate through them
        and delete old versions
        """
        return coiterate(ItemCleaner(self, items, perIter))


    def getRevision(self):
        """
        Returns the current revision number
        """
        if self.revision is not None:
            return self.revision
        r = self.store.findFirst(StoreRevision,
            sort=StoreRevision.revision.descending)
        if r is None:
            self.revision = 0
            return 0
        self.revision = r.revision
        return r.revision


    def updateRevision(self, revision):
        """
        Change the revision of the store
        """
        r = StoreRevision(store=self.store, revision=revision,
            timestamp=int(time()))
        self.revision = revision


    def storeObject(self, object, async=False):
        """
        Store the object synchronously
        """
        if async:
            raise StoreError("This is not an async store!")
        return PickleBasedStore.storeObject(self, object, async=False)


    def storeObjectSync(self, object):
        """
        Synchronously store an object via an AxiomContainer
        """
        self.getRevision()
        self.Log("Storing %s" % str(object.__class__.__name__))
        if hasattr(object, "_datastore_id"):
            oid = object._datastore_id
        else:
            oid = self.createObjectId(object)
            # do a sanity check to make sure that nothing is
            # already using this oid
            # r = self.store.query(AxiomContainer, attributes.AND(
                # AxiomContainer.did == oid,
                # AxiomContainer.revision <= self.revision))
            # if r.count() != 0:
                # raise StoreError("Duplicate oid detected for class %s" %
                    # object.__class__.__name__)
            object._datastore_id = oid
        oc = AxiomContainer(object=object,
            did=oid, datastore=self, revision=self.revision + 1)
        assert oc.did == oid
        assert oc.did is not None
        object.__dict__["_datastore_id"] = oid
        idxes = []
        for clazz, itm in getIndexes(object.__class__):
            i = AxiomIndex(
                clazz=unicode(clazz),
                attribute=unicode(itm),
                item=oid,
                revision=self.revision + 1)
            attr = getattr(object, itm)
            i.setValue(attr)
            idxes.append(i)
        self.changes.append((oc, idxes))
        self.objectcache[oid] = oc
        if hasattr(object, "_datastore_storing"):
            del object._datastore_storing
        return oc.did


    def retrieveObjectSync(self, oid):
        def retrieve():
            revision = self.getRevision()
            oc = self.store.findFirst(AxiomContainer,
                attributes.AND(AxiomContainer.did == oid,
                AxiomContainer.revision <= revision),
                sort=AxiomContainer.revision.descending)
            self.objectcache[oid] = oc
            if oc.object is not None:
                oc.obdict = pickle.loads(str(oc.data))
                oc.buildObject(self)
            return oc.getObject()
        return self.store.transact(retrieve)


    def query(self, clazz, async=False):
        """
        Return a query object
        """
        return AxiomQuery(self, clazz, self.revision)


    def close(self):
        """
        Close this datastore
        """
        def done(_):
            self.store.close()
            self.lock = False
        return self.sync().addCallback(done)
