# Copyright (C) 2007-2008 Mortal Coil Games
# See LICENSE for details.
# -*- test-case-name: mv3d.test.test_memorystore -*-
"""
A datastore that is in memory, but can be dumped to disk if desired
"""

import os, sys, time
from threading import RLock

try:
    import cPickle as pickle
except ImportError:
    import pickle

from twisted.internet import defer
from twisted.internet.threads import deferToThread

from mv3d.util.datastore.store import (PickleBasedStore, StoreError, 
    ObjectDoesntExist)
from mv3d.util.datastore.localfile import ObjectContainer


class MemoryStore(PickleBasedStore):
    """
    This DataStore saves its data to local files
    """
    def __init__(self):
        self.path = None
        self.objectcache = {}
        self.queryindexes = {}
        self.userdata = {}
        self.objectnames = {}
        self.retrieving = {}
        self.lock = True
        self.storeversion = 1
        self.logfile = None
        self.database = {}
        self.plock = RLock()
    

    def __getstate__(self):
        raise StoreError("You can't store the datastore, silly!")
        
        
    def Log(self, message):
        if self.logfile is None:
            return
        self.logfile.write("%f - %s\n"%(time.time(), message))
        self.logfile.flush()

    
    def create(self, path=None):
        """
        Create a new datastore
        """
        self.path = path
        return True
    
    
    def open(self, path=None, async=False):
        """
        Open a datastore. If path is None, works just like create.
        Otherwise, reads our databse from path.
        """
        if path is None:
            return defer.succeed(True)
        self.path = path
        if not async:
            self.load()
            return True
        def openThread():
            """
            Since this could take a long time, we put it in a thread
            """
            self.load()
            return True
        return deferToThread(openThread)
        
        
    def openOrCreate(self, path=None):
        """
        Try opening the store, but if that fails, create a new one
        """
        try:
            return self.open(path)
        except IOError:
            return self.create(path)


    def close(self):
        """
        Close this datastore
        """
        self.Log("Closing the datastore")
        if self.path is None:
            return defer.succeed(True)
        def closeThread():
            self.save()
        return deferToThread(closeThread)
        
        
    def sync(self):
        """
        Sync this datastore to persistant media
        """
        if self.path is None:
            return defer.succeed(True)
        def syncThread():
            self.save()
        return deferToThread(syncThread)
        
        
    def storeObjectSync(self, object):
        self.Log("Storing %s"%str(object.__class__.__name__))
        if hasattr(object, "_datastore_id"):
            oid = object._datastore_id
        else:
            oid = self.createObjectId(object)
        oc = ObjectContainer(self, object, oid)
        object.__dict__["_datastore_id"] = oid
        try:
            self.database[oid] = pickle.dumps(oc)
        except:
            from traceback import format_exc
            self.Log("Error Storing! %s %s %s %s" % (
                object, sys.exc_info()[0], sys.exc_info()[1], format_exc()))
            raise #sys.exc_info()[1]
        self.objectcache[oid] = oc
        if hasattr(object, "_datastore_storing"):
            del object._datastore_storing
        self.updateQueryIndex(object)
        return oid


    def storeObject(self, object, async=False):
        """
        Store the object synchronously
        """
        if async:
            raise StoreError("This is an async store!")
        return PickleBasedStore.storeObject(self, object, async=False)


    def retrieveObjectSync(self, oid):
        if not self.database.has_key(oid):
            raise ObjectDoesntExist(
                "There is no object with that id in this store (%s)" % oid)
        oc = pickle.loads(self.database[oid])
        self.objectcache[oid] = oc
        oc.buildObject(self)
        self.resetObjectQueryValues(oc.getObject())
        return oc.getObject()


    def retrieveObject(self, oid, async=False, useCache=True):
        """
        Retrieve the object synchronously
        """
        if async:
            raise StoreError("This is an async store!")
        return PickleBasedStore.retrieveObject(self, oid, async=False,
            useCache=useCache)


    def delete(self, object):
        if not hasattr(object, "_datastore_id"):
            # not in a store
            object.datastore = None
            return defer.succeed(True)
        oid = object._datastore_id
        del self.database[oid]
        object.datastore = None
        self.removeFromIndexes(object)
        if self.objectcache.has_key(oid):
            del self.objectcache[oid]
        return defer.succeed(True)


    def save(self):
        """
        Save the store to disk
        """
        self.Log("Saving store.")
        try:
            if os.path.exists(self.path + ".bak"):
                os.remove(self.path + ".bak")
            if os.path.exists(self.path):
                os.rename(self.path, self.path + ".bak")
            out = (self.storeversion, self.userdata, self.objectnames, 
                self.queryindexes, self.database)
            mydir = os.path.split(self.path)[0]
            if not os.path.exists(mydir):
                os.makedirs(mydir)
            f = open(self.path, "w")
            pickle.dump(out, f)
            f.close()
        except:
            if os.path.exists(self.path):
                os.rename(self.path, self.path + ".error")
            if os.path.exists(self.path + ".bak"):
                os.rename(self.path + ".bak", self.path)
            raise
        
        
    def load(self):
        """
        Load the store into memory
        """
        self.Log("Loading store.")
        f = open(self.path, "r")
        data = pickle.load(f)
        f.close()
        self.storeversion, self.userdata, self.objectnames = data[:3]
        self.queryindexes, self.database = data[3:]


    def storeNamedObject(self, name, object):
        """
        Store an object by name
        """
        oid = self.storeObject(object)
        self.nameObject(name, oid)
        return oid


    def createQueryIndex(self, classname, property):
        """
        Makes it so queries can be run against classname.property
        """
        if not self.queryindexes.has_key(classname):
            self.queryindexes[classname] = {}
        self.queryindexes[classname][property] = {}


    def transact(self, f, *a, **kw):
        """
        In other stores, this runs f in a transaction.
        That isn't too useful here
        """
        return f(*a, **kw)