# -*- test-case-name: mv3d.test.test_datastore -*-
# Copyright (C) 2007-2008 Mortal Coil Games
# See LICENSE for details.
"""

"""


class QueryError(Exception):
    """
    An exception for problems with queries
    """
    

class QueryCompare:
    """
    QueryParam creates these. They are required to support and/or
    """
    def __init__(self, name, cmp, value):
        self.name = name
        self.cmp = cmp
        self.value = value
        if hasattr(value, "_datastore_id"):
            self.value = value._datastore_id
        
        
    def __getitem__(self, x):
        if x == 0 or x == "name":
            return self.name
        if x == 1 or x == "cmp":
            return self.cmp
        if x == 2 or x == "value":
            return self.value
        raise QueryError("Unknown x in __getitem__")
        
        
    def __and__(self, other):
        return QueryCompare(self, "&", other)
        
        
    def __or__(self, other):
        return QueryCompare(self, "|", other)


    def __str__(self):
        op = self.cmp
        if op == "&":
            op = "and"
        if op == "|":
            op = "or"
        return " ".join([str(self.name), op, str(self.value)])


    def check(self, left):
        """
        Check if the query succeeds against left
        """
        if hasattr(left, "_datastore_id"):
            # print "comparing",left,"to", self.value
            left = left._datastore_id
        if self.cmp == "==":
            return left == self.value
        if self.cmp == "<":
            return left < self.value
        if self.cmp == ">":
            return left > self.value
        if self.cmp == "<=":
            return left <= self.value
        if self.cmp == ">=":
            return left >= self.value
        if self.cmp == "!=":
            return left != self.value
        if self.cmp == "&":
            return left and self.value
        if self.cmp == "|":
            return left and self.value
        raise QueryError("Unknown comparison %s" % self.cmp)



class QueryParam:
    """
    Used to specify parameters in a query
    """
    def __init__(self, name):
        self.name = name
        
        
    def __eq__(self, value):
        return QueryCompare(self.name, "==", value)


    def __ne__(self, value):
        return QueryCompare(self.name, "!=", value)


    def __lt__(self, value):
        return QueryCompare(self.name, "<", value)


    def __gt__(self, value):
        return QueryCompare(self.name, ">", value)


    def __le__(self, value):
        return QueryCompare(self.name, "<=", value)


    def __ge__(self, value):
        return QueryCompare(self.name, ">=", value)
        
        
    def __and__(self, value):
        return QueryCompare(self.name, "&", value)


    def __or__(self, value):
        return QueryCompare(self.name, "|", value)


class QueryObject(object):
    """
    This object can be used to run a query on a store
    """
    def __init__(self, store, clazz):
        self.store = store
        self.clazz = clazz
        if store is not None:
            self.queryindex = self.store.queryindexes.get(clazz.__name__, {})


    def __getattr__(self, attrib):
        a = self.__dict__.get(attrib, getattr(self.__class__, attrib, None))
        if a is None:
            return QueryParam(attrib)
        return a


    def fetchResults(self, query=None, offset=None, limit=None, sort=None):
        """
        Actually perform the query
        
        Some example queries
            q.fetchResults(q.name=="Mike")
            q.fetchResults(q.age>=20)
            q.fetchResults(q.name=="Mike" & q.age<100) 
        """
        if not len(self.queryindex):
            return QueryResults(self.store, [])
        ret = []
        if query is None:
            qi = self.queryindex.values()[0]
            for v in qi.values():
                ret.extend(v)
            ret = ret[offset:]
            if limit is not None:
                return QueryResults(self.store, ret[:limit])
            return QueryResults(self.store, ret)
        if isinstance(query[0], str):
            qi = self.queryindex.get(query[0])
            if qi is None:
                return QueryResults(self.store, [])
            qik = qi.keys()
            matches = filter(query.check, qik)
            for m in matches:
                ret += qi[m]
            ret = ret[offset:]
            if limit is not None:
                ret = ret[:limit]
            return QueryResults(self.store, ret)
        if query[1] == "&":
            left = self.fetchResults(query[0], sort=sort).results
            right = self.fetchResults(query[2], sort=sort).results
            # @@TODO switch filter based on whether left has more entries 
            # than right
            ret = filter(lambda x: x in right, left)[offset:]
            if limit is not None:
                ret = ret[:limit]
            return QueryResults(self.store, ret)
        if query[1] == "|":
            left = self.fetchResults(query[0], sort=sort).results
            right = self.fetchResults(query[2], sort=sort).results
            ret = (filter(lambda x: x not in right, left) + 
                filter(lambda x: x not in left, right))[offset:]
            if limit is not None:
                ret = ret[:limit]
            return QueryResults(self.store, ret)
        raise QueryError("Combined queries are not yet supported")


    def findUnique(self, query=None, offset=None, limit=None, sort=None):
        """
        Run a query that you expect to get one result from
        eventually returns that one result
        """
        results = self.fetchResults(query, offset, limit, sort)
        if len(results) != 1:
            if len(results):
                msg = "Query was not unique (%d results)." % len(results)
            else:
                msg = "No results for findUnique. (q: %s) " % str(query)
            raise QueryError(msg)
        return results[0]


    def findFirst(self, query=None, offset=None, limit=None, sort=None):
        """
        Run a query return the first result
        """
        results = self.fetchResults(query, offset, limit, sort)
        if not len(results):
            raise QueryError("No results for findFirst (q: %s" % str(query))
        return results[0]


queryObject = QueryObject(None, None)


class QueryResults:
    """
    Stores results from a query
    """
    def __init__(self, store, results):
        self.store = store
        self.results = results


    def __getitem__(self, offset):
        """
        Fetch the item from the store and return that
        """
        # print "get item", offset
        if isinstance(offset, slice):
            return QueryResults(self.store, self.results[offset])
        oid = self.results[offset]
        r = self.store.retrieveObject(oid)
        # print "got", r
        return r


    def __len__(self):
        return len(self.results)


    def __add__(self, other):
        assert isinstance(other, QueryResults), "You can only add QueryResults"
        assert other.store == self.store, "Stores must match"
        return QueryResults(self.store, self.results + other.results)


    def count(self):
        return len(self.results)