# -*- test-case-name: mv3d.test.test_datastore -*-
# Copyright (C) 2007-2008 Mortal Coil Games
# See LICENSE for details.
"""

"""

import sha
from time import time
from random import random

from twisted.internet import defer
from twisted.internet.threads import deferToThread
from twisted.python.util import mergeFunctionMetadata

from mv3d.util.datastore.query import QueryObject


class VersionError(Exception):
    """
    Raised when there is a problem upgrading from one version to another
    """


def autoStore(func):
    """
    Automatically store this object after the completion of the wrapped function
    if the wrapped function returns a deferred, wait for it to complete before
    saving.
    If the object didn't come from or hasn't been saved in a store, it won't
    be saved now.
    """
    def wrapper(self, *args, **kw):
        d = func(self, *args, **kw)
        def done(r):
            if hasattr(self, "datastore") and self.datastore is not None:
                d = self.datastore.storeObject(self)
                if isinstance(d, defer.Deferred):
                    return d.addCallback(lambda x: r)
                return r
            return r
        if isinstance(d, defer.Deferred):
            return d.addCallback(done)
        return done(d)
    mergeFunctionMetadata(func, wrapper)
    return wrapper



class StoreError(Exception):
    """
    An exception for major store issues (can't open, can't create, etc)
    """



class ObjectDoesntExist(Exception):
    """
    An exception to specify that the object doesn't exist
    """
    


class ObjectReference:
    """
    This class is stored with other objects and denotes a reference to a 
    seperate object
    """
    oid = None
    
    def __init__(self, oid):
        self.oid = oid
        
        
    def __setstate__(self, state):
        """ 
        Called when unpickled. We get magically changed to the new object
        """
        self.__dict__.update(state)



class DataStoreAutoSave:
    """
    If you inherit from this class, your object will automatically save itself
    """
    def __setattr__(self, k, v):
        if hasattr(self, "_datastore_from"):
            self.__dict__["_datastore_saving"] = self._datastore_from.storeObject(self)
        self.__dict__[k] = v



def attributes(*classes, **attrs):
    """
    Takes a list of classes and a list of attributes from this class and generates a
    dict including the attributes from the classes and args
    """
    for c in classes:
        attrs.update(c.datastoreAttributes)
    return attrs



class DataStoreMethod(object):
    """
    Defines a way of storing a particular class in the datastore
    """


asObject = DataStoreMethod()



class DataStoreAttribute(object):
    """
    Defines a type of attribute that can be stored in a datastore
    """


class TranslatingAttribute(DataStoreAttribute):
    """
    An attribute that requires translation between memory and store
    """
    def toStore(self, store, value, object=None, attribute=None):
        """
        Convert value to the value to be saved in the store
        """
        return value


    def fromStore(self, store, value, object=None, attribute=None):
        """
        Convert value from the stored state to in memory state
        """
        return value



class Reference(TranslatingAttribute):
    """
    Defines a reference to another item in the data store
    """
    def toStore(self, store, value, object=None, attribute=None):
        """
        Convert the value into a placeholder and store the value
        """
        if value is None:
            return value
        oid = store.storeObject(value, async=False)
        return ObjectReference(oid)


    def fromStore(self, store, value, object=None, attribute=None):
        """
        Load the value's oid from the store and return that
        """
        if value is None:
            return value
        return store.retrieveObject(value.oid, async=False)



class ReferenceDict(TranslatingAttribute):
    """
    Defines a dict that can contain values that are references
    """
    def toStore(self, store, value, object=None, attribute=None):
        if value is None:
            return None
        newdict = value.copy()
        for k, v in newdict.items():
            if hasattr(v, "datastoreAttributes"):
                oid = store.storeObject(v, async=False)
                newdict[k] = ObjectReference(oid)
            if isinstance(v, dict):
                newdict[k] = self.toStore(store, v)
            if isinstance(v, list) or isinstance(v, tuple):
                newdict[k] = referenceList.toStore(store, v)
        return newdict


    def fromStore(self, store, value, object=None, attribute=None):
        if value is None:
            return None
        for k, v in value.items():
            if isinstance(v, ObjectReference):
                value[k] = store.retrieveObject(v.oid, async=False)
            if isinstance(v, dict):
                value[k] = self.fromStore(store, v)
            if isinstance(v, list) or isinstance(v, tuple):
                value[k] = referenceList.fromStore(store, v)
        return value



class ReferenceList(TranslatingAttribute):
    """
    Defines a list that can contain values that are references
    """
    def toStore(self, store, value, object=None, attribute=None):
        if value is None:
            return None
        newlist = list(value)
        for x in range(len(newlist)):
            i = newlist[x]
            if hasattr(i, "datastoreAttributes"):
                oid = store.storeObject(i, async=False)
                newlist[x] = ObjectReference(oid)
            if isinstance(i, dict):
                newlist[x] = referenceDict.toStore(store, i)
            if isinstance(i, list) or isinstance(i, tuple):
                newlist[x] = self.toStore(store, i)
        if isinstance(value, tuple):
            return tuple(newlist)
        return newlist


    def fromStore(self, store, value, object=None, attribute=None):
        if value is None:
            return value
        tupl = isinstance(value, tuple)
        value = list(value)
        for x in range(len(value)):
            i = value[x]
            if isinstance(i, ObjectReference):
                value[x] = store.retrieveObject(i.oid, async=False)
            if isinstance(i, dict):
                value[x] = referenceDict.fromStore(store, i)
            if isinstance(i, list) or isinstance(i, tuple):
                value[x] = self.fromStore(store, i)
        if tupl:
            return tuple(value)
        return value


text = DataStoreAttribute()
simpleList = DataStoreAttribute()
simpleDict = DataStoreAttribute()
dictionary = DataStoreAttribute()
number = DataStoreAttribute()
boolean = DataStoreAttribute()
inline = DataStoreAttribute()
reference = Reference()
referenceDict = ReferenceDict()
referenceList = ReferenceList()



class SpecialHash:
    """
    A special hash item
    """
    def getHash(self, object):
        """
        Get the special hash for the object
        """


class TimeHash(SpecialHash):
    """
    A time based hash
    """
    def getHash(self, object):
        """
        Get the current time
        """
        return str(time())


class RandomHash(SpecialHash):
    """
    A time based hash
    """
    def getHash(self, object):
        """
        Get a random number
        """
        return str(random())


class ClassHash(SpecialHash):
    """
    Class name based hash
    """
    def getHash(self, object):
        """
        Returns the classes name
        """
        return object.__class__.__name__


class ReferenceHash(SpecialHash):
    """
    Hash based on an object id
    """
    def __init__(self, attrib):
        self.attrib = attrib


    def getHash(self, object):
        """
        Returns the object id of the attrib
        """
        return str(getattr(object, self.attrib)._datastore_id)



specialHash = SpecialHash()
# timeHash = TimeHash()
classHash = ClassHash()
randomHash = RandomHash()



def setAttrs(self, **kw):
    """
    Check the keywords to make sure self's class has an attribute for
    it, and then set it if so.
    """
    for k, v in kw.iteritems():
        if not hasattr(self.__class__, k):
            raise ValueError("Inalid keyword argument %s to %s" % (k,
                self.__class__.__name__))
    self.__dict__.update(kw)



def getState(self):
    """
    For inline objects, you can use this for __getstate__. It adheres
    to datastoreAttributes. Note that references won't work here.
    """
    state = dict()
    for k, v in self.datastoreAttributes.items():
        state[k] = getattr(self, k)
        if v.__class__.__name__.startswith("reference"):
            raise StoreError("Inline %s defines attribute %s as a"
                " reference!" % (self.__class__.__name, k))
        if isinstance(v, TranslatingAttribute):
            state[k] = v.toStore(self.store, state[k])
    state["datastoreVersion"] = getattr(self, "datastoreVersion", 1)
    return state



def buildObject(self, store):
    """
    For inline objects, calling this will
    allow you to version your objects in the database and adheres
    to datastoreAttributes.
    """
    version = getattr(self, "datastoreVersion", 1)
    if version != self.__class__.datastoreVersion:
        if not hasattr(self, "datastoreUpgrade"):
            raise VersionError(
                "No method to update %s from version %d to %d." % (
                self.__class__.__name__, version, 
                self.__class__.datastoreVersion))
        self.datastoreUpgrade(store, version)
    for k, v in self.datastoreAttributes.items():
        if isinstance(v, TranslatingAttribute):
            setattr(self, k, v.fromStore(store, 
            getattr(self, k)))



class DataStore(object):
    """
    This is the interface to all datastores
    """
    def create(self, **kw):
        """
        Create a new datastore specified by kw
        Returns a deferred
        """
        
    
    def open(self, **kw):
        """
        Open a pre-existing datastore specified by kw
        Returns a deferred
        """
               

    def close(self):
        """
        Close this datastore
        Returns a deferred
        """
        
        
    def sync(self):
        """
        Sync this datastore to persistant media
        """
        
        
    def storeObject(self, object, async=True):
        """
        Store object in the datastore. Return a reference to its id
        Returns a deferred if async is True, otherwise blocks until object
        it stored.
        """
        
        
    def retrieveObject(self, oid, async=True):
        """
        Retrieve object with id of oid from the store
        Returns a deferred if async is True, otherwise blocks until object
        is retrieved (and returns object)
        """


    def delete(self, object):
        """
        Delete object from the store
        """
    

    def createQueryIndex(self, classname, property):
        """
        Makes it so queries can be run against classname.property
        """
        
        
    def query(self, classname):
        """
        Return a query object
        """
        
        
    def nameObject(self, name, oid):
        """
        Allow users to store name:oid mappings
        """
        
        
    def retrieveNamedObject(self, name):
        """
        Retrieve an object by name
        """
        
        
    def storeNamedObject(self, name, object):
        """
        Store an object by name
        """
        
        
    def retrieveNamedObjectTo(self, name, dest):
        """
        Retrieve a named object into dest
        """
        
        
    def setUserData(self, k, v):
        """
        Sets data in the user dict for this store
        """
        
        
    def getUserData(self, k, default=None):
        """
        Get value from user dict or return default
        """



class PickleBasedStore(DataStore):
    """
    A DataStore type that is based on pickling objects
    """
    def storeObject(self, object, async=True):
        """
        Store an object in the datastore
        """
        if not self.lock:
            raise StoreError("The store isn't locked!")
        if hasattr(object, "_datastore_storing"):
            if async:
                d = defer.Deferred()
                d.callback(object._datastore_id)
                return d
            return object._datastore_id
        object.__dict__["_datastore_storing"] = True
        if hasattr(object, "prepareForStore"):
            object.prepareForStore(self)
        def done(r):
            object.datastore = self
            if hasattr(object, "savedToStore"):
                object.savedToStore()
            return r
        if async:
            d = deferToThread(self.storeObjectSync, object)
            def error(e):
                self.Log("Error storing %s: %s"%(str(object), str(e)))
                e.raiseException()
            d.addCallback(done)
            d.addErrback(error)
            return d
        return done(self.storeObjectSync(object))


    def storeObjects(self, objectList):
        """
        Store a bunch of objects. Handles async and sync stores
        """
        deferred = False
        results = []
        for obj in objectList:
            results.append(self.storeObject(obj))
            if isinstance(results[-1], defer.Deferred):
                deferred = True
        if deferred:
            return defer.gatherResults(results)
        return results


    def retrieveObject(self, oid, async=True, useCache=True):
        """
        Retrieve an object from the datastore
        """
        if not self.lock:
            raise StoreError("The store is not locked!")
        if useCache and self.objectcache.has_key(oid):
            if async:
                return defer.succeed(self.objectcache[oid].getObject())
            return self.objectcache[oid].getObject()
        if self.retrieving.has_key(oid):
            raise StoreError("Thou shalt not retrieve an object that is already"
                "being retrieved")
            d = defer.Deferred()
            self.retrieving[oid].append(d)
            return d
        def finished(r):
            r.datastore = self
            for d in self.retrieving[oid]:
                d.callback(r)
            del self.retrieving[oid]
            if hasattr(r, "loadedFromStore"):
                r.loadedFromStore()
            return r
        self.retrieving[oid] = []
        self.Log("Retrieving %s"%oid)
        if async:
            d = deferToThread(self.retrieveObjectSync, oid)
            def error(e):
                self.Log("Error retrieving %s: %s"%(str(oid), str(e)))
                e.raiseException()
            d.addCallback(finished)
            d.addErrback(error)
        else:
            d = finished(self.retrieveObjectSync(oid))
        return d


    def retrieveObjects(self, oidList):
        """
        Retrieve a bunch of objects. Handles async and sync stores
        """
        deferred = False
        results = []
        for oid in oidList:
            results.append(self.retrieveObject(oid))
            if isinstance(results[-1], defer.Deferred):
                deferred = True
        if deferred:
            return defer.gatherResults(results)
        return results


    def updateQueryIndex(self, object):
        """
        Update query indexes for an object
        """
        indexes = getattr(object, "datastoreIndexes", None)
        oid = object._datastore_id
        if indexes is not None and len(indexes):
            dsv = getattr(object, "datastoreQueryValue", dict())
            def update(clazz, idxes):
                if not self.queryindexes.has_key(clazz):
                    self.queryindexes[clazz] = dict()
                qi = self.queryindexes[clazz]
                for idx in idxes:
                    if hasattr(idx, "datastoreIndexes"):
                        # @@ WARNING: Recursive
                        update(idx.__name__, idx.datastoreIndexes)
                        continue
                    prop = getattr(object, idx)
                    if hasattr(prop, "_datastore_id"):
                        prop = prop._datastore_id
                    if dsv.has_key(idx) and dsv[idx] != prop:
                        if qi.has_key(idx) and qi[idx].has_key(dsv[idx]):
                            # flush the old value
                            qi[idx][dsv[idx]].remove(oid)
                            # print "flushing", prop
                        # else:
                            # print "not flushing", prop, qi.has_key(idx), qi[idx].has_key(dsv[idx])
                    if not qi.has_key(idx):
                        qi[idx] = dict()
                    v = qi[idx]
                    if not v.has_key(prop):
                        v[prop] = []
                    if oid not in v[prop]:
                        v[prop].append(oid)
            update(object.__class__.__name__, indexes)
        self.resetObjectQueryValues(object)


    def resetObjectQueryValues(self, object):
        """
        Set all the query values for an object
        """
        indexes = getattr(object, "datastoreIndexes", None)
        if indexes is not None and len(indexes):
            object.datastoreQueryValue = dict()
            def reset(idxes):
                for idx in idxes:
                    if hasattr(idx, "datastoreIndexes"):
                        # @@ WARNING: Recursive
                        reset(idx.datastoreIndexes)
                        continue
                    object.datastoreQueryValue[idx] = getattr(object, idx)
            reset(indexes)


    def removeFromIndexes(self, object):
        """
        Remove this obejct from all indexes
        """
        oid = object._datastore_id
        indexes = getattr(object, "datastoreIndexes", None)
        if indexes is not None and len(indexes):
            def remove(clazz, idxes):
                qi = self.queryindexes[clazz]
                for idx in idxes:
                    if hasattr(idx, "datastoreIndexes"):
                        # @@ WARNING: Recursive
                        remove(idx.__name__, idx.datastoreIndexes)
                        continue
                    prop = getattr(object, idx)
                    if hasattr(prop, "_datastore_id"):
                        prop = prop._datastore_id
                    qi[idx][prop].remove(oid)
            remove(object.__class__.__name__, indexes)


    def createObjectId(self, object):
        """
        Create a unique object id for this object
        """
        h = sha.new()
        for k in object.datastoreHash:
            if isinstance(k, SpecialHash):
                h.update(k.getHash(object))
            else:
                h.update(repr(getattr(object, k)))
        return object.__class__.__name__ + "_" + h.hexdigest()


    def createQueryIndex(self, classname, property):
        """
        Makes it so queries can be run against classname.property
        """
        if not self.queryindexes.has_key(classname):
            self.queryindexes[classname] = {}
        self.queryindexes[classname][property] = {}
        return defer.succeed(True)


    def query(self, clazz, async=True):
        """
        Return a query object
        """
        return QueryObject(self, clazz)


    def nameObject(self, name, oid):
        """
        Allow users to store name:oid mappings
        """
        self.objectnames[name] = oid
        
        
    def retrieveNamedObject(self, name):
        """
        Retrieve an object by name
        """
        return self.retrieveObject(self.objectnames[name])
        
        
    def storeNamedObject(self, name, object):
        """
        Store an object by name
        """
        d = self.storeObject(object)
        def stored(r):
            self.nameObject(name, r)
            return r
        d.addCallback(stored)
        return d
        
        
    def retrieveNamedObjectTo(self, name, dest):
        """
        Retrieve a named object into dest
        """
        return self.retrieveObject(self.objectnames[name], dest)
        
        
    def setUserData(self, k, v):
        """
        Sets data in the user dict for this store
        """
        self.userdata[k] = v
        
        
    def getUserData(self, k, default=None):
        """
        Get value from user dict or return default
        """
        return self.userdata.get(k, default)


    def vacuum(self):
        """
        Go through the store and delete any files that aren't in the index
        """