# Copyright (C) 2007-2008 Mortal Coil Games
# See LICENSE for details.
""" -*- test-case-name: mv3d.test.test_syncfilestore -*-

"""


import os, sys, time
try:
    import cPickle as pickle
except ImportError:
    import pickle

from mv3d.util.datastore.store import (
    PickleBasedStore, StoreError, ObjectDoesntExist,)
from mv3d.util.datastore.localfile import ObjectContainer



class SyncFileStore(PickleBasedStore):
    """
    This DataStore saves its data to local files synchronously
    """
    path = None
    lock = False
    storeversion = 1
    logfile = None
    lastsave = None
    saveinterval = 10


    def __init__(self):
        self.objectcache = {}
        self.locations = {}
        self.queryindexes = {}
        self.userdata = {}
        self.objectnames = {}
        self.retrieving = {}
    

    def __del__(self):
        if self.lock:
            try:
                self.updateIndex()
            finally:
                self.unlockStore()
                
                
    def __getstate__(self):
        raise StoreError("You can't store the datastore, silly!")
        
        
    def Log(self, message):
        if self.logfile is None:
            return
        self.logfile.write("%f - %s\n"%(time.time(), message))
        self.logfile.flush()

    
    def create(self, path=None):
        """
        Create a new datastore at kw[path]
        """
        self.path = path
        if os.path.exists(self.path):
            raise StoreError("Path already exists when creating a DataStore!")
        os.makedirs(self.path)
        self.lockStore()
        self.updateIndex()
        return True
    
    
    def open(self, path=None, force=False):
        """
        Open a datastore at kw[path]
        """
        self.path = path
        self.lockStore(force)
        self.loadIndex()
        return True


    def openOrCreate(self, path=None, force=False):
        """
        Try opening the store, but if that fails, create a new one
        """
        try:
            return self.open(path, force)
        except IOError:
            return self.create(path)


    def close(self):
        """
        Close this datastore
        """
        self.Log("Closing the datastore")
        self.updateIndex()
        self.unlockStore()
        return True


    def sync(self):
        """
        Sync this datastore to persistant media
        """
        self.updateIndex()


    def storeObjectSync(self, object):
        self.Log("Storing %s"%str(object.__class__.__name__))
        if hasattr(object, "_datastore_id"):
            oid = object._datastore_id
        else:
            oid = self.createObjectId(object)
        oc = ObjectContainer(self, object, oid)
        object.__dict__["_datastore_id"] = oid
        filename = self.getFileFor(oid)
        f = open(filename, "w")
        try:
            pickle.dump(oc, f)
        except:
            from traceback import format_exc
            self.Log("Error Storing! %s %s %s %s"% (object, 
                sys.exc_info()[0], sys.exc_info()[1], format_exc()))
            raise #sys.exc_info()[1]
        f.close()
        self.objectcache[oid] = oc
        self.locations[oid] = filename
        if hasattr(object, "_datastore_storing"):
            del object._datastore_storing
        self.updateQueryIndex(object)
        return oid


    def storeObject(self, object, async=False):
        """
        Store the object synchronously
        """
        if async:
            raise StoreError("This is an async store!")
        return PickleBasedStore.storeObject(self, object, async=False)


    def retrieveObjectSync(self, oid):
        if not self.locations.has_key(oid):
            raise ObjectDoesntExist(
                "There is no object with that id in this store")
        filename = self.locations[oid]
        f = open(filename, "r")
        oc = pickle.load(f)
        f.close()
        self.objectcache[oid] = oc
        oc.buildObject(self)
        self.resetObjectQueryValues(oc.getObject())
        return oc.getObject()


    def retrieveObject(self, oid, async=False, useCache=True):
        """
        Retrieve the object synchronously
        """
        if async:
            raise StoreError("This is an async store!")
        return PickleBasedStore.retrieveObject(self, oid, async=False,
            useCache=useCache)


    def delete(self, object):
        """
        Delete object from the store
        """
        self.removeFromIndexes(object)
        object.datastore = None
        oid = object._datastore_id
        filename = self.locations[oid]
        del self.locations[oid]
        if self.objectcache.has_key(oid):
            del self.objectcache[oid]
        os.remove(filename)


    def updateIndex(self):
        """
        Update the store's index file
        """
        self.Log("Updating store index.")
        indexfile = os.path.join(self.path, "index")
        f = open(indexfile, "w")
        pickle.dump(self.storeversion, f)
        pickle.dump(self.locations, f)
        pickle.dump(self.userdata, f)
        pickle.dump(self.objectnames, f)
        pickle.dump(self.queryindexes.keys(), f)
        f.close()
        self.storeQueryIndexes()
        
        
    def loadIndex(self):
        """
        Load the index of the store into memory
        """
        self.Log("Loading store index.")
        indexfile = os.path.join(self.path, "index")
        f = open(indexfile, "r")
        sv = pickle.load(f)
        if sv != self.storeversion:
            raise StoreError("The store is an incompatible version!")
        self.locations = pickle.load(f)
        self.userdata = pickle.load(f)
        self.objectnames = pickle.load(f)
        qis = pickle.load(f)
        for qi in qis:
            self.loadQueryIndex(qi)
        f.close()
    
    
    def lockStore(self, force=False):
        """
        Lock this store for writing
        
        Ideally, this should include our ip and port so that anyone trying to 
        get the lock can instead connect remotely to us.
        """
        self.Log("Locking the store.")
        lockfile = os.path.join(self.path, "lock")
        if os.path.exists(lockfile) and not force:
            raise StoreError("The store is already locked!")
        f = open(lockfile, "w")
        pickle.dump([os.getpid(), self.path], f)
        f.close()
        self.lock = True
        logfile = os.path.join(self.path, "store.log")
        self.logfile = open(logfile, "a")
        
    
    def unlockStore(self):
        """
        Unlock the store if we previously had it locked
        """
        self.Log("Unlocking the store.")
        if not self.lock:
            raise StoreError("Can't unlock the DataStore if it isn't locked!")
        lockfile = os.path.join(self.path, "lock")
        os.remove(lockfile)
        self.logfile.close()
        self.lock = False
        
        
    def getFileFor(self, oid):
        """
        Get a unique filename for this object id.
        """
        if self.locations.has_key(oid):
            return self.locations[oid]
        return os.path.join(self.path, oid + ".pkl")
        
        
    def storeQueryIndex(self, classname):
        """
        Stores all the query indexes for classname
        """
        filename = os.path.join(self.path, "queryindex."+classname)
        f = open(filename, "w")
        pickle.dump(self.queryindexes[classname], f)
        f.close()
        
        
    def storeQueryIndexes(self):
        """
        Store all of our query indexes
        """
        for k in self.queryindexes.keys():
            self.storeQueryIndex(k)
            
            
    def loadQueryIndex(self, classname):
        """
        Loads all the query indexes for a classname
        """
        filename = os.path.join(self.path, "queryindex."+classname)
        f = open(filename, "r")
        self.queryindexes[classname] = pickle.load(f)
        f.close()


    def storeNamedObject(self, name, object):
        """
        Store an object by name
        """
        oid = self.storeObject(object)
        self.nameObject(name, oid)
        return oid


    def createQueryIndex(self, classname, property):
        """
        Makes it so queries can be run against classname.property
        """
        if not self.queryindexes.has_key(classname):
            self.queryindexes[classname] = {}
        self.queryindexes[classname][property] = {}


