# Copyright (C) 2006-2008 Mortal Coil Games
# See LICENSE for details.
""" -*- test-case-name: mv3d.test.test_noise -*-
Perlin noise functions.
    Thanks to Ken Perlin and 
    http://freespace.virgin.net/hugo.elias/models/m_perlin.htm
"""


import math
import random

from mv3d.util.datastore.store import (attributes, 
    randomHash, number, simpleDict)
from mv3d.net.pb import Cacheable
from mv3d.net.security import Securable


class PerlinNoise(Cacheable, Securable):
    """
    This class is a base noise class. It contains default
    implimentations of Perlin noise functions
    it works in any amount of dimensions.
    """
    datastoreAttributes = attributes(Securable, Cacheable,
        seed=number, persistence=number, octaves=number, map=simpleDict)
    datastoreHash = ("seed", "persistence", randomHash)
    datastoreVersion = 1
    datastore = None

    seed = 234
    persistence = 0.5
    octaves = 4
    map = None
    cache = None
    
    def __init__(self, seed=234, persistence=0.5, octaves=4, map=None):
        # self.primes = [15731, 789221, 1376312589, 1073741824.0]
        self.map = map or {}
        self.octaves = octaves
        self.persistence = persistence
        self.seed = seed
        self.cache = {}
        self.addExclude("cache")

    
    def activate(self):
        """
        Called when loaded
        """
        self.cache = {}


    def updateMap(self, pos, value):
        """
        You need to add a docstring here ! 
        """
        self.map[pos] = value
        
        
    def makeSeed(self, pos):
        """
        Generate a seed value for pos
        """
        if not isinstance(pos, list) and not isinstance(pos, tuple):
            pos = (pos,)
        x = 1
        for n, r in zip(pos, range(len(pos))):
            x += n * (r + 1) + (n * r * 1000)
        assert self.seed != 0
        x *= self.seed
        return x


    def noise(self, pos):
        """
        Pos can be a single number, or a tuple
        this function must generate randomish numbers, but
        always give the same number for the same value of pos
        """
        #print pos
        opos = pos
        if self.map.has_key(pos):
            return self.map[pos]
        if self.cache.has_key(pos):
            return self.cache[pos]
        x = self.makeSeed(pos)
        random.seed(x)
        res = random.random() * 2 - 1
        self.cache[opos] = res
        return res
        #x=pow((x<<3),x)
        #return (1.0 - (int( x * (x * x * int(self.primes[0]) + int(self.primes[1])) + self.primes[2]) & 0x7fffffff) / float(self.primes[3]))


    def smoothNoise(self, pos):
        """
        You need to add a docstring here ! 
        """
        if not isinstance(pos, list) and not isinstance(pos, tuple):
            pos = [pos]
        dims = len(pos)
        offs = [ - 1, 0, 1]
        mpow = dims * dims
        acums = [0] * dims
        posn = []
        if dims == 1:
            return self.noise(pos) / 2.0 + self.noise(pos[0] - 1) / 4.0 + self.noise(pos[0] + 1) / 4.0
        if dims == 2:
            x = pos[0]
            y = pos[1]
            corners = ( self.noise((x - 1, y - 1)) + self.noise((x + 1, y - 1)) + self.noise((x - 1, y + 1)) + self.noise((x + 1, y + 1)) ) / 16
            sides   = ( self.noise((x - 1, y))  +self.noise((x + 1, y)) + self.noise((x, y - 1))  +self.noise((x, y + 1))   ) /  8
            center  =  self.noise((x, y)) / 4
            return corners + sides + center
        if dims == 3:
            x = pos[0]
            y = pos[1]
            z = pos[2]
            f = self.noise
            ocn = f((x - 1, y - 1, z - 1)) + f((x + 1, y - 1, z - 1)) + f((x - 1, y + 1, z - 1)) + f((x + 1, y + 1, z - 1))
            ocn += f((x - 1, y - 1, z + 1)) + f((x + 1, y - 1, z + 1)) + f((x - 1, y + 1, z + 1)) + f((x + 1, y + 1, z + 1))
            ocn /= 64
            corners = f((x - 1, y - 1, z)) + f((x + 1, y - 1, z)) + f((x + 1, y + 1, z)) + f((x - 1, y + 1, z))
            corners += f((x - 1, y, z - 1)) + f((x + 1, y, z - 1)) + f((x + 1, y, z + 1)) + f((x - 1, y, z + 1))
            corners += f((x, y - 1, z - 1)) + f((x, y + 1, z - 1)) + f((x, y + 1, z + 1)) + f((x, y - 1, z + 1))
            corners /= 32
            sides = f((x - 1, y, z)) + f((x + 1, y, z)) + f((x, y - 1, z)) + f((x, y + 1, z))
            sides += f((x, y, z + 1)) + f((x, y, z - 1))
            sides /= 16
            center = f((x, y, z)) / 8
            return center + sides + corners + ocn
        if dims == 4:
            x = pos[0]
            y = pos[1]
            z = pos[2]
            a = pos[3]
            f = self.noise
            cn4 = f((x - 1, y - 1, z - 1, a - 1)) + f((x + 1, y - 1, z - 1, a - 1)) + f((x - 1, y + 1, z - 1, a - 1))
            cn4 += f((x + 1, y + 1, z - 1, a - 1)) + f((x - 1, y - 1, z + 1, a - 1)) + f((x + 1, y - 1, z + 1, a - 1))
            cn4 += f((x - 1, y + 1, z + 1, a - 1)) + f((x + 1, y + 1, z + 1, a - 1)) + f((x - 1, y - 1, z - 1, a + 1))
            cn4 += f((x + 1, y - 1, z - 1, a + 1)) + f((x - 1, y + 1, z - 1, a + 1)) + f((x + 1, y + 1, z - 1, a + 1))
            cn4 += f((x - 1, y - 1, z + 1, a + 1)) + f((x + 1, y - 1, z + 1, a + 1)) + f((x - 1, y + 1, z + 1, a + 1))
            cn4 += f((x + 1, y + 1, z + 1, a + 1))
            cn4 /= 256
            ocn = f((x - 1, y - 1, z - 1, a)) + f((x + 1, y - 1, z - 1, a)) + f((x - 1, y + 1, z - 1, a)) + f((x + 1, y + 1, z - 1, a))
            ocn += f((x - 1, y - 1, z + 1, a)) + f((x + 1, y - 1, z + 1, a)) + f((x - 1, y + 1, z + 1, a)) + f((x + 1, y + 1, z + 1, a))
            ocn += f((x, y - 1, z - 1, a - 1)) + f((x, y - 1, z - 1, a + 1)) + f((x, y + 1, z - 1, a - 1)) + f((x, y + 1, z - 1, a + 1))
            ocn += f((x, y - 1, z + 1, a - 1)) + f((x, y - 1, z + 1, a + 1)) + f((x, y + 1, z + 1, a - 1)) + f((x, y + 1, z + 1, a + 1))
            ocn += f((x - 1, y, z - 1, a - 1)) + f((x - 1, y, z - 1, a + 1)) + f((x + 1, y, z - 1, a - 1)) + f((x + 1, y, z - 1, a + 1))
            ocn += f((x - 1, y, z + 1, a - 1)) + f((x - 1, y, z + 1, a + 1)) + f((x + 1, y, z + 1, a - 1)) + f((x + 1, y, z + 1, a + 1))
            ocn += f((x - 1, y - 1, z, a - 1)) + f((x - 1, y - 1, z, a + 1)) + f((x + 1, y - 1, z, a - 1)) + f((x + 1, y - 1, z, a + 1))
            ocn += f((x - 1, y + 1, z, a - 1)) + f((x - 1, y + 1, z, a + 1)) + f((x + 1, y + 1, z, a - 1)) + f((x + 1, y + 1, z, a + 1))
            ocn /= 128
            corners = f((x - 1, y - 1, z, a)) + f((x + 1, y - 1, z, a)) + f((x + 1, y + 1, z, a)) + f((x - 1, y + 1, z, a))
            corners += f((x - 1, y, z - 1, a)) + f((x + 1, y, z - 1, a)) + f((x + 1, y, z + 1, a)) + f((x - 1, y, z + 1, a))
            corners += f((x, y - 1, z - 1, a)) + f((x, y + 1, z - 1, a)) + f((x, y + 1, z + 1, a)) + f((x, y - 1, z + 1, a))
            corners += f((x - 1, y, z, a - 1)) + f((x + 1, y, z, a - 1)) + f((x + 1, y, z, a + 1)) + f((x - 1, y, z, a + 1))
            corners += f((x, y - 1, z, a - 1)) + f((x, y + 1, z, a - 1)) + f((x, y + 1, z, a + 1)) + f((x, y - 1, z, a + 1))
            corners += f((x, y, z - 1, a - 1)) + f((x, y, z - 1, a + 1)) + f((x, y, z + 1, a + 1)) + f((x, y, z + 1, a - 1))
            corners /= 64
            sides = f((x - 1, y, z, a)) + f((z + 1, y, z, a)) + f((x, y - 1, z, a)) + f((x, y + 1, z, a))
            sides += f((x, y, z + 1, a)) + f((x, y, z - 1, a)) + f((x, y, z, a + 1)) + f((x, y, z, a - 1))
            sides /= 32
            center = f((x, y, z)) / 16
            return center + sides + corners + ocn + cn4


    def linterp(self, a, b, p):
        """
        You need to add a docstring here ! 
        """
        return a * (1.0 - p) + b * p


    def cosinterp(self, a, b, p):
        """
        You need to add a docstring here ! 
        """
        ft = p * 3.1415927
        f = (1 - math.cos(ft)) * 0.5
        return a * (1 - f) + b * f


    def interpolatedNoise(self, pos):
        """
        You need to add a docstring here ! 
        """
        if not isinstance(pos, list) and not isinstance(pos, tuple):
            pos = [pos]
        if len(pos) == 1:
            x = pos[0]
            ix = int(x)
            fx = x - ix
            v1 = self.smoothNoise(ix)
            v2 = self.smoothNoise(ix + 1)
            return self.cosinterp(v1, v2, fx)
        if len(pos) == 2:
            x = pos[0]
            y = pos[1]
            ix = int(x)
            iy = int(y)
            fx = x - ix
            fy = y - iy
            v1 = self.smoothNoise([ix, iy])
            v2 = self.smoothNoise([ix + 1, iy])
            v3 = self.smoothNoise([ix, iy + 1])
            v4 = self.smoothNoise([ix + 1, iy + 1])
            i1 = self.cosinterp(v1, v2, fx)
            i2 = self.cosinterp(v3, v4, fx)
            return self.cosinterp(i1, i2, fy)
        if len(pos) == 3:
            x = pos[0]
            y = pos[1]
            z = pos[2]
            ix = int(x)
            iy = int(y)
            iz = int(z)
            fx = x - ix
            fy = y - iy
            fz = z - iz
            f = self.smoothNoise
            v = [f([ix, iy, iz]), f([ix + 1, iy, iz]), f([ix, iy + 1, iz]), f([ix + 1, iy + 1, iz]), 
                f([ix, iy, iz + 1]), f([ix + 1, iy, iz + 1]), f([ix, iy + 1, iz + 1]), f([ix + 1, iy + 1, iz + 1])]
            ix1 = self.cosinterp(v[0], v[1], fx)
            iy1 = self.cosinterp(v[2], v[3], fx)
            ixy = self.cosinterp(ix1, iy1, fy)
            ix2 = self.cosinterp(v[4], v[5], fx)
            iy2 = self.cosinterp(v[6], v[7], fx)
            ixz = self.cosinterp(ix2, iy2, fy)
            return self.cosinterp(ixy, ixz, fz)
        if len(pos) == 4:
            x = pos[0]
            y = pos[1]
            z = pos[2]
            a = pos[3]
            ix = int(x)
            iy = int(y)
            iz = int(z)
            ia = int(a)
            fx = x - ix
            fy = y - iy
            fz = z - iz
            fa = a - ia
            f = self.smoothNoise
            v = [f([ix, iy, iz, ia]), f([ix + 1, iy, iz, ia]), f([ix, iy + 1, iz, ia]), 
                f([ix + 1, iy + 1, iz, ia]), f([ix, iy, iz + 1, ia]), f([ix + 1, iy, iz + 1, ia]), 
                f([ix, iy + 1, iz + 1, ia]), f([ix + 1, iy + 1, iz + 1, ia]), f([ix, iy, iz, ia + 1]), 
                f([ix + 1, iy, iz, ia + 1]), f([ix, iy + 1, iz, ia + 1]), f([ix + 1, iy + 1, iz, ia + 1]), 
                f([ix, iy, iz + 1, ia + 1]), f([ix + 1, iy, iz + 1, ia + 1]), f([ix, iy + 1, iz + 1, ia + 1]), 
                f([ix + 1, iy + 1, iz + 1, ia + 1])]
            ix1 = self.cosinterp(v[0], v[1], fx)
            iy1 = self.cosinterp(v[2], v[3], fx)
            ixy1 = self.cosinterp(ix1, iy1, fy)
            ix2 = self.cosinterp(v[4], v[5], fx)
            iy2 = self.cosinterp(v[6], v[7], fx)
            ixy2 = self.cosinterp(ix2, iy2, fy)
            ixyz1 = self.cosinterp(ixy1, ixy2, fz)
            ix3 = self.cosinterp(v[8], v[9], fx)
            iy3 = self.cosinterp(v[10], v[11], fx)
            ixy3 = self.cosinterp(ix3, iy3, fy)
            ix4 = self.cosinterp(v[12], v[13], fx)
            iy4 = self.cosinterp(v[14], v[15], fx)
            ixy4 = self.cosinterp(ix4, iy4, fy)
            ixyz2 = self.cosinterp(ixy3, ixy4, fz)
            return self.cosinterp(ixyz1, ixyz2, fa)


    def perlinNoise(self, pos):
        """
        You need to add a docstring here ! 
        """
        if not isinstance(pos, list) and not isinstance(pos, tuple):
            pos = [pos]
        total = 0
        for i in range(self.octaves):
            freq = math.pow(2, i)
            amp = math.pow(self.persistence, i)
            p = []
            for pp in pos:
                p.append(pp * freq)
            total += self.interpolatedNoise(p) * amp
        return total
